Require Export Metatheory.
Require Export FSets.

Set Undo 100000.

Inductive exp : Type :=
  | exp_bvar : nat -> exp
  | exp_fvar : atom -> exp
  | exp_base : atom -> exp
  | exp_lam  : typ -> exp -> exp
  | exp_app  : exp -> exp -> exp
  | exp_tab  : kind -> exp -> exp
  | exp_tap  : exp -> typ -> exp
(*   | exp_cap  : exp -> exp -> exp *)
(*   | exp_case : exp -> (pat * exp) -> exp -> exp *)
  | exp_new  : exp -> exp
  | exp_bot  : exp
  | exp_cert : exp -> (list exp_or_typ) -> exp
with typ : Type :=
  | typ_btvar   : nat -> typ
  | typ_ftvar   : atom -> typ
  | typ_const   : atom -> typ
  | typ_fun     : typ -> typ -> typ
  | typ_univ    : kind -> effect -> typ -> typ
(*   | typ_constr : typ -> typ -> typ *)
  | typ_qtype   : typ -> typ
  | typ_app     : typ -> typ -> typ
  | typ_dep     : typ -> exp -> typ
  | typ_sing    : typ -> typ -> typ
  | typ_dropped : typ -> typ   
with kind : Type :=
  | kind_A    : kind 
  | kind_U    : kind
  | kind_N    : kind
  | kind_tcon : kind -> kind -> kind
  | kind_dcon : typ -> kind -> kind
with exp_or_typ : Type :=
  | eort_exp : exp -> exp_or_typ
  | eort_typ : typ -> exp_or_typ
with effect : Type :=
  | FX : effect
  | NO_FX : effect. 

Notation "[ x ]" := (x :: nil).

Fixpoint open_exp (k : nat) (f : exp) (e : exp) {struct e} : exp :=
  match e with
    | exp_bvar i => if k === i then f else (exp_bvar i)
    | exp_fvar x => exp_fvar x
    | exp_base x => exp_base x
    | exp_lam t e1 => exp_lam (open_typ k f t) (open_exp (S k) f e1)
    | exp_app e1 e2 => exp_app (open_exp k f e1) (open_exp k f e2)
    | exp_tab kk e1 => exp_tab kk (open_exp k f e1)
    | exp_tap e1 t => exp_tap (open_exp k f e1) (open_typ k f t)
    | exp_new e1 => exp_new (open_exp k f e1)
    | exp_bot => exp_bot
    | exp_cert e1 et_l => exp_cert e1 et_l (* Fix this *)
  end
with open_typ (k : nat) (f : exp) (t : typ) {struct t} : typ :=
  match t with 
    | typ_btvar i => typ_btvar i
    | typ_ftvar x => typ_ftvar x
    | typ_const x => typ_const x
    | typ_fun t1 t2 => typ_fun (open_typ k f t1) (open_typ (S k) f t2)
    | typ_univ kk ef t1 => typ_univ kk ef (open_typ k f t1)
    | typ_qtype t1 => typ_qtype (open_typ k f t1)
    | typ_app t1 t2 => typ_app (open_typ k f t1) (open_typ k f t2)
    | typ_dep t1 e2 => typ_dep (open_typ k f t1) (open_exp k f e2)
    | typ_sing t1 t2 => typ_sing (open_typ k f t1) (open_typ k f t2)
    | typ_dropped t1 => typ_dropped (open_typ k f t1)
  end.


Fixpoint open_exp_typ (k : nat) (t : typ) (e : exp) {struct e} : exp :=
  match e with
    | exp_bvar i => exp_bvar i
    | exp_fvar x => exp_fvar x
    | exp_base x => exp_base x
    | exp_lam t1 e1 => exp_lam (open_typ_typ k t t1) (open_exp_typ k t e1)
    | exp_app e1 e2 => exp_app (open_exp_typ k t e1) (open_exp_typ k t e2)
    | exp_tab kk e1 => exp_tab kk (open_exp_typ (S k) t e1)
    | exp_tap e1 t2 => exp_tap (open_exp_typ k t e1) (open_typ_typ k t t2)
    | exp_new e1 => exp_new (open_exp_typ k t e1)
    | exp_bot => exp_bot
    | exp_cert e1 et_l => exp_cert e1 et_l (* Fix this *)
  end
with open_typ_typ (k : nat) (t : typ) (t' : typ) {struct t'} : typ :=
  match t' with 
    | typ_btvar i => if k === i then t else typ_btvar i
    | typ_ftvar x => typ_ftvar x
    | typ_const x => typ_const x
    | typ_fun t1 t2 => typ_fun (open_typ_typ k t t1) (open_typ_typ k t t2)
    | typ_univ kk ef t1 => typ_univ kk ef (open_typ_typ (S k) t t1)
    | typ_qtype t1 => typ_qtype (open_typ_typ k t t1)
    | typ_app t1 t2 => typ_app (open_typ_typ k t t1) (open_typ_typ k t t2)
    | typ_dep t1 e2 => typ_dep (open_typ_typ k t t1) (open_exp_typ k t e2)
    | typ_sing t1 t2 => typ_sing (open_typ_typ k t t1) (open_typ_typ k t t2)
    | typ_dropped t1 => typ_dropped (open_typ_typ k t t1)
  end.

Coercion typ_btvar : nat >-> typ.
Coercion typ_ftvar : atom >-> typ.
Coercion exp_bvar : nat >-> exp.
Coercion exp_fvar : atom >-> exp.
Notation domain := (list exp_or_typ).

(* Local closure for types and terms *)
Inductive lc_typ : typ -> Prop :=
| lc_typ_fvar  : forall a, lc_typ (typ_ftvar a)
| lc_typ_const : forall a, lc_typ (typ_const a)
| lc_typ_fun   : 
  forall L t1 t2 a, a `notin` L ->
    (lc_typ t1) -> 
    (lc_typ (open_typ O a t2)) -> 
    (lc_typ (typ_fun t1 t2))
| lc_typ_univ : 
  forall L t1 ef a k, a `notin` L ->
    (lc_typ (open_typ_typ 0 a t1)) -> 
    (lc_typ (typ_univ k ef t1))
| lc_typ_qtype :
  forall t1, (lc_typ t1) -> (lc_typ (typ_qtype t1))
| lc_typ_app : 
  forall t1 t2, (lc_typ t1) -> (lc_typ t2) -> (lc_typ (typ_app t1 t2))
| lc_typ_dep :
  forall t1 e1, (lc_typ t1) -> (lc_exp e1) -> (lc_typ (typ_dep t1 e1))
| lc_typ_sing :
  forall t1 t2, (lc_typ t1) -> (lc_typ t2) -> (lc_typ (typ_sing t1 t2))
| lc_typ_dropped: 
  forall t1, (lc_typ t1) -> (lc_typ (typ_dropped t1))
with lc_exp : exp -> Prop :=
| lc_exp_fvar : forall x:atom, lc_exp (exp_fvar x)
| lc_exp_base : forall x:atom, lc_exp (exp_base x)
| lc_exp_lam  : 
(*
  forall L t e a, a `notin` L ->
    (lc_typ t) -> (lc_exp (open_exp 0 a e)) -> (lc_exp (exp_lam t e))
*)
  forall L t e, (lc_typ t) ->
    (forall a, a `notin` L -> lc_exp (open_exp 0 a e)) -> lc_exp (exp_lam t e)
(*  forall t e x, lc_typ t -> lc_exp (open_exp 0 x e) -> lc_exp (exp_lam t e) *)
| lc_exp_app  : 
  forall e1 e2, 
    (lc_exp e1) -> (lc_exp e2) -> (lc_exp (exp_app e1 e2))
| lc_exp_tab  :
  forall L k e a, a `notin` L -> 
    (lc_exp (open_exp_typ 0 a e)) -> (lc_exp (exp_tab k e))
| lc_exp_tap  : 
  forall e t, (lc_exp e) -> (lc_typ t) -> (lc_exp (exp_tap e t))
| lc_exp_new  : 
  forall e, (lc_exp e) -> (lc_exp (exp_new e))
| lc_exp_bot  : (lc_exp exp_bot)
| lc_exp_cert :
  forall e el,
    (lc_exp e) -> (lc_domain el) -> (lc_exp (exp_cert e el))
with lc_domain : domain -> Prop :=
| lc_domain_nil : lc_domain nil
| lc_domain_cons_exp : 
  forall e tl, (lc_exp e) -> (lc_domain tl) -> 
    (lc_domain ((eort_exp e)::tl))
| lc_domain_cons_typ : 
  forall t tl, (lc_typ t) -> (lc_domain tl) -> 
    (lc_domain ((eort_typ t)::tl)).
  
Inductive value : exp -> Prop :=
| value_abs : forall (t:typ) (e:exp),
  (lc_exp (exp_lam t e)) -> value (exp_lam t e)
| value_tab : forall (k:kind) (e:exp),
  (lc_exp (exp_tab k e)) -> value (exp_tab k e)
| value_new : forall (e:exp), 
  (value e) -> (value (exp_new e))
| value_cert : forall (a:atom) (d:domain), 
  (lc_exp (exp_base a)) -> (val_domain d) -> (value (exp_cert (exp_base a) d))
with val_domain : domain -> Prop :=
| val_domain_nil : (val_domain nil)
| val_domain_t : forall (t:typ) (tl:domain), 
  (lc_typ t) -> (val_domain tl) -> val_domain ((eort_typ t)::tl)
| val_domain_e : forall (e:exp) (tl:domain), 
  (lc_exp e) -> (val_domain tl) -> val_domain ((eort_exp e)::tl).

(* Dynamic semantics *)

Inductive binding : Type :=
| binding_typ : typ -> binding
| binding_kind : kind -> binding.

(* Coercion binding_typ : typ >-> binding. *)
(* Coercion binding_kind : kind >-> binding.  *)

Notation gamma := (list (atom * binding)).
Notation env   := (prod gamma atoms).

Notation dom_eqn := (prod domain exp). 

Inductive in_eqns  : domain -> list dom_eqn -> exp -> Prop := .
Inductive not_in_eqns : domain -> list dom_eqn -> Prop := .

Inductive simple_kind : kind -> Prop := 
| sk_A : simple_kind kind_A
| sk_U : simple_kind kind_U
| sk_N : simple_kind kind_N.

(* Relative closure for types and terms *)
Inductive rc_typ : typ -> gamma -> Prop :=
| rc_typ_bvar : forall n g, rc_typ (typ_btvar n) g

| rc_typ_fvar : forall a g k, 
  (binds a (binding_kind k) g) -> rc_typ (typ_ftvar a) g

| rc_typ_const : forall T g, 
  rc_typ (typ_const T) g
  
| rc_typ_fun : forall t1 t2 g, 
  (rc_typ t1 g) -> (rc_typ t2 g) -> rc_typ (typ_fun t1 t2) g
  
| rc_typ_univ : forall k t ef g, 
  (rc_typ t g) -> rc_typ (typ_univ k ef t) g

| rc_typ_qtype : forall t g, 
  rc_typ t g -> rc_typ (typ_qtype t) g

| rc_typ_app : forall t1 t2 g, 
  rc_typ t1 g -> rc_typ t2 g -> rc_typ (typ_app t1 t2) g

| rc_typ_dep : forall t e g, 
  rc_typ t g -> rc_exp e g -> rc_typ (typ_dep t e) g

| rc_typ_sing : forall t a g, 
  rc_typ t g -> rc_typ a g -> rc_typ (typ_sing t a) g

| rc_typ_dropped : forall t g, 
  rc_typ t g -> rc_typ (typ_dropped t) g

with rc_exp : exp -> gamma -> Prop :=
| rc_exp_bvar : forall n g, rc_exp (exp_bvar n) g

| rc_exp_fvar : forall x g t,
  (binds x (binding_typ t) g) -> rc_exp (exp_fvar x) g
  
| rc_exp_base : forall b g,
  rc_exp (exp_base b) g

| rc_exp_lam : forall t e g,
  (rc_typ t g) -> (rc_exp e g) -> rc_exp (exp_lam t e) g

| rc_exp_app : forall e1 e2 g, 
  (rc_exp e1 g) -> (rc_exp e2 g) -> rc_exp (exp_app e1 e2) g
  
| rc_exp_tab : forall k e g,
  (simple_kind k) -> (rc_exp e g) -> rc_exp (exp_tab k e) g

| rc_exp_tap  :  forall e t g, 
  (rc_exp e g) -> (rc_typ t g) -> (rc_exp (exp_tap e t) g)

| rc_exp_new  : forall e g, 
  (rc_exp e g) -> (rc_exp (exp_new e) g)

| rc_exp_bot  : forall g, 
  rc_exp exp_bot g

| rc_exp_cert : forall e el g,
    (rc_exp e g) -> (rc_domain el g) -> rc_exp (exp_cert e el) g

with rc_domain : domain -> gamma -> Prop :=
| rc_domain_nil : forall g, 
  rc_domain nil g

| rc_domain_cons_exp :  forall e tl g, 
  (rc_exp e g) -> (rc_domain tl g) -> rc_domain ((eort_exp e)::tl) g

| rc_domain_cons_typ :  forall t tl g, 
  (rc_typ t g) -> (rc_domain tl g) -> 
  (rc_domain ((eort_typ t)::tl) g).

Definition wf_typ (t:typ) (g:gamma) := lc_typ t /\  rc_typ t g.
Definition wf_exp (e:exp) (g:gamma) := lc_exp e /\ rc_exp e g.

Inductive splitA : atoms -> atoms -> atoms -> Prop := 
| split_A : forall A1 A2, (A1 `inter` A2 = {}) -> splitA A1 A2 (A1 `union` A2).

(* Auxiliary functions a(x,A) and p(A) *)
Definition aqual (x:atom) (k:kind) (A:atoms) : atoms := 
  match k with 
    | kind_A => (singleton x) `union` A
    | _      => A
  end.

Definition ptype (t:typ) (A:atoms) (fx:atoms) : typ :=
  match AtomSet.F.is_empty (A `union` fx) with 
    | true => t
    | false => (typ_qtype t)
  end.

Definition affines (G:gamma) : atoms := {}. (* expand this *)

Definition effect_ok fx G :=  forall a, a `in` fx -> (binds a (binding_kind kind_N) G).

Inductive only_names : gamma -> atoms -> Prop :=
| only_names_empty :  
   only_names nil {}

| only_names_extra_in_G : forall G X, 
  only_names G {} -> 
  only_names (([(X, binding_kind kind_N)])++G) {} 
   
| only_names_cons :
  forall G A A' (X:atom), 
    only_names G A -> 
    splitA (singleton X) A A' ->
    only_names ([(X, binding_kind kind_N)] ++ G) A'.

Module Type LAIR_PARAMS.

Parameter Model : (list (atom * (list dom_eqn))).
Parameter Signature : gamma.

Inductive eval : exp -> exp -> Prop :=

| eval_bot_bot : eval exp_bot exp_bot

| eval_app1 : forall e1 e1' e2, 
  eval e1 e1' -> eval (exp_app e1 e2) (exp_app e1' e2)
  
| eval_app1_bot : forall e1 e2, 
  eval e1 exp_bot -> eval (exp_app e1 e2) exp_bot
  
| eval_app2 : forall e1 e2 e2', 
  eval e2 e2' -> eval (exp_app e1 e2) (exp_app e1 e2')

| eval_app2_bot : forall e1 e2, 
  eval e2 exp_bot -> eval (exp_app e1 e2) exp_bot

| eval_beta : forall t e v, 
  (value v) -> (eval (exp_app (exp_lam t e) v) (open_exp 0 v e))

| eval_tap1 : forall e1 e1' t, 
  (eval e1 e1') -> (eval (exp_tap e1 t) (exp_tap e1' t))

| eval_tap1_bot : forall e1 t, 
  (eval e1 exp_bot) -> (eval (exp_tap e1 t) exp_bot)

| eval_tap : forall e1 k t, 
  (eval (exp_tap (exp_tab k e1) t) (open_exp_typ 0 t e1))

| eval_new1 : forall e1 e1', 
  (eval e1 e1') ->  (eval (exp_new e1) (exp_new e1'))

| eval_new1_bot : forall e1, 
  (eval e1 exp_bot) ->  (eval (exp_new e1) exp_bot)

| eval_delta : forall a eqns, 
  (binds a eqns Model) -> (eval (exp_base a) (exp_cert (exp_base a) nil))

| eval_b1 : forall b d v eqns v',
  (value (exp_cert (exp_base b) d)) -> (value v) -> 
  (binds b eqns Model) -> (in_eqns ((eort_exp v)::d) eqns v') -> 
  (eval (exp_app (exp_cert (exp_base b) d) v) v')

| eval_b2 : forall b d v eqns,
  (value (exp_cert (exp_base b) d)) -> (value v) -> 
  (binds b eqns Model) -> (not_in_eqns ((eort_exp v)::d) eqns) -> 
  (eval (exp_app (exp_cert (exp_base b) d) v)  (exp_cert (exp_base b) ((eort_exp v)::d)))

| eval_b3 : forall b d t eqns v',
  (value (exp_cert (exp_base b) d)) -> 
  (binds b eqns Model) -> (in_eqns ((eort_typ t)::d) eqns v') -> 
  (eval (exp_tap (exp_cert (exp_base b) d) t)  v')

| eval_b4 : forall b d t eqns,
  (value (exp_cert (exp_base b) d)) -> 
  (binds b eqns Model) -> (not_in_eqns ((eort_typ t)::d) eqns) -> 
  (eval (exp_tap (exp_cert (exp_base b) d) t)  (exp_cert (exp_base b) ((eort_typ t)::d))).

(* Static semantics *)

Inductive typing : gamma -> atoms -> exp -> (typ * atoms) -> Prop:=
| type_X : 
  forall G x t, 
    (binds x (binding_typ t) G) -> 
    (kinding G t kind_U) ->
    (typing G {} x (t, {}))

| type_XA :
  forall G x t,
    (binds x (binding_typ t) G) ->
    (kinding G t kind_A) ->
    (typing G (singleton x) x (t, {}))
    
| type_NEW :
  forall G A e t a fx a_fx,
    (typing G A e (t, fx)) ->
    (kinding G t kind_U) ->
    (binds a (binding_kind kind_N) G) ->
    (splitA (singleton a) fx a_fx) ->
    (typing G A (exp_new e) ((typ_sing (typ_qtype t) a), a_fx))

| type_ABS :
  forall L G A t_x k e t_e fx, 
    (forall x, x `notin` L ->
      (typing ([(x,binding_typ t_x)] ++ G) (aqual x k A) (open_exp 0 x e) ((open_typ 0 x t_e), fx))) ->
    (kinding G t_x k) ->
    (typing G A (exp_lam t_x e) ((ptype (typ_fun t_x t_e) A fx), fx))
    
| type_TAB :
  forall L G A k e t_e fx, 
    (forall a, a `notin` L -> a `notin` fx ->
      (typing ([(a,binding_kind k)] ++ G) A (open_exp_typ 0 a e) ((open_typ_typ 0 a t_e), fx))) ->
    (simple_kind k) ->
    (typing G A (exp_tab k e) ((ptype (typ_univ k NO_FX t_e) A fx), fx))
    
| type_TAB_FX :
  forall L G A k e t_e fx,
    (AtomSet.F.Subset fx L) ->
    (forall a, a `notin` L -> 
      (typing ([(a,binding_kind k)] ++ G) A (open_exp_typ 0 a e) ((open_typ_typ 0 a t_e), fx `union` singleton a))) ->
    (simple_kind k) ->
    (typing G A (exp_tab k e) ((ptype (typ_univ k FX t_e) A fx), fx))
    
| type_APP :
  forall G A A1 A2 e1 e2 t2 t fx1 fx2 fx t_e1, 
    (splitA A1 A2 A) ->
    (t_e1 = (typ_fun t2 t) \/ t_e1 = (typ_qtype (typ_fun t2 t))) ->
    (typing G A1 e1 (t_e1, fx1)) ->
    (typing G A2 e2 (t2, fx2)) -> 
    (splitA fx1 fx2 fx) ->
    (typing G A (exp_app e1 e2) ((open_typ 0 e2 t), fx))

(* | type_QAPP : *)
(*   forall G A A1 A2 e1 e2 t2 t fx1 fx2 fx,  *)
(*     (splitA A1 A2 A) -> *)
(*     (typing G A1 e1 (typ_qtype (typ_fun t2 t), fx1)) -> *)
(*     (typing G A2 e2 (t2, fx2)) ->  *)
(*     (splitA fx1 fx2 fx) -> *)
(*     (typing G A (exp_app e1 e2) ((open_typ 0 e2 t), fx)) *)

| type_TAP_NOFX :
  forall G A e t k t' fx te,
    (te = (typ_univ k NO_FX t') \/ te = typ_qtype (typ_univ k NO_FX t')) ->
    (typing G A e (te, fx)) ->
    (kinding G t k) ->
    (typing G A (exp_tap e t) ((open_typ_typ 0 t t'), fx))

| type_TAP_FX :
  forall G A e a t' fx' fx te,
    (te = (typ_univ kind_N FX t') \/ te = typ_qtype (typ_univ kind_N FX t')) ->
    (typing G A e (te, fx)) ->
    (kinding G (typ_ftvar a) kind_N) ->
    (splitA (singleton a) fx fx') -> 
    (typing G A (exp_tap e (typ_ftvar a)) ((open_typ_typ 0 (typ_ftvar a) t'), fx'))

| type_BOT :
  forall G t fx,
    (typing G {} exp_bot (t, fx))
    
| type_DROP :
  forall G A e t a fx,
    (typing G A e ((typ_sing t a), fx)) ->
    (typing G A e ((typ_dropped t), fx))

| type_AFN :
  forall G A A1 A2 e t fx fx1 fx2, 
    (splitA A1 A2 A) ->
    (typing G A1 e (t, fx1)) -> 
    (splitA fx1 fx2 fx) -> 
    (effect_ok fx G) ->
    (typing G A e (t, fx))
    
| type_B :
  forall G B t,
    (binds B (binding_typ t) Signature) ->
    (typing G {} (exp_base B) (t, {}))

| type_B1 :
  forall G1 G2 A B domain v t fx,
    (only_names G1 {}) -> 
    (wf_exp v G1) -> 
    (typing (G2++G1) A (exp_app (exp_cert (exp_base B) domain) v) (t,fx)) -> 
    (typing (G2++G1) A (exp_cert (exp_base B) ((eort_exp v)::domain)) (t,fx))
    
| type_B2 :
  forall G1 G2 A B domain t' t fx,
    (only_names G1 {}) -> 
    (wf_typ t' G1) ->
    (typing (G2++G1) A (exp_tap (exp_cert (exp_base B) domain) t') (t,fx)) -> 
    (typing (G2++G1) A (exp_cert (exp_base B) ((eort_typ t')::domain)) (t,fx))
    
| type_B3 :
  forall G A B t fx,
    (typing G A (exp_base B) (t, fx)) -> 
    (typing G A (exp_cert (exp_base B) nil) (t,fx))

| type_CONV :
  forall G A e t t' fx,
    (typing G A e (t,fx)) -> 
    (tequiv t t') ->
    (typing G A e (t',fx))
    
with kinding : gamma -> typ -> kind -> Prop :=

| kinding_A :
  forall G a k,
    (binds a (binding_kind k) G) ->
    (kinding G a k)
    
| kinding_N :
  forall G t a,
    (kinding G t kind_A) ->
    (kinding G a kind_N) ->
    (kinding G (typ_sing t a) kind_A)

| kinding_TC :
  forall G tc K,
    (binds tc (binding_kind K) Signature) ->
    (kinding G (typ_const tc) K)

| kinding_AFN :
  forall G t,
    (kinding G t kind_U) ->
    (kinding G (typ_qtype t) kind_A)

| kinding_ALL :
  forall L G k t k' fx,  
    (forall a, a `notin` L ->
      (kinding ([(a,binding_kind k)] ++ G) (open_typ_typ 0 a t) k')) ->
    (kinding G (typ_univ k fx t) kind_U)
    
| kinding_FUN :
  forall L G t_x t k_x k,
    (kinding G t_x k_x) ->
    (forall x, x `notin` L ->
      (kinding ([(x,binding_typ t_x)] ++ G) (open_typ 0 x t) k)) -> 
    (kinding G (typ_fun t_x t) kind_U)

| kinding_TAP :
  forall G t t' k K,
    (kinding G t (kind_tcon k K)) ->
    (kinding G t' k) ->
    (kinding G (typ_app t t') K)

| kinding_DEP :
  forall G t e t' K,
    (kinding G t (kind_dcon t' K)) ->
    (typing G (affines G) e (t',{})) -> (* fix this to include a separate type-level phase *)
    (kinding G (typ_dep t e) K)

| kinding_DROPPED :
  forall G t k,
    (kinding G t k) ->
    (kinding G (typ_dropped t) k)

with tequiv : typ -> typ -> Prop := 
| te_ID : forall t, tequiv  t t
  
| te_SYM : forall  t t', tequiv  t' t -> tequiv  t t'
  
| te_CTX_FUN_1 : 
  forall  tx tx' t, 
    tequiv  tx tx' -> 
    tequiv  (typ_fun tx t) (typ_fun tx' t)
    
| te_CTX_FUN_2 : 
  forall  tx t t', 
    tequiv  t t' -> 
    tequiv  (typ_fun tx t) (typ_fun tx t')

| te_CTX_UNIV :
  forall  k t t' ef, 
    tequiv  t t' -> 
    tequiv  (typ_univ k ef t) (typ_univ k ef t')

| te_CTX_Q :
  forall  t t', 
    tequiv  t t' ->
    tequiv  (typ_qtype t) (typ_qtype t')
    
| te_CTX_APP1 :
  forall  t1 t1' t2, 
    tequiv  t1 t1' -> 
    tequiv  (typ_app t1 t2) (typ_app t1' t2)

| te_CTX_APP2 :
  forall  t1 t2 t2', 
    tequiv  t2 t2' -> 
    tequiv  (typ_app t1 t2) (typ_app t1 t2')

| te_CTX_SING :
  forall  t t' n, 
    tequiv  t t' -> 
    tequiv  (typ_sing t n) (typ_sing t' n)

| te_CTX_DEP_1:
  forall  t t' e, 
    tequiv  t t' -> 
    tequiv  (typ_dep t e) (typ_dep t' e)

| te_RED:
  forall  t_e fx t e e', 
    typing nil {} e (t_e, fx) -> 
    typing nil {} e' (t_e, fx) -> 
    eval e e' ->
    tequiv  (typ_dep t e) (typ_dep t e')

| te_CTX_DROPPED:
  forall  t t', 
    tequiv  t t' ->
    tequiv  (typ_dropped t) (typ_dropped t').

Scheme typing_ind2 :=
  Induction for typing Sort Prop
with kinding_ind2 :=
  Induction for kinding Sort Prop
with tequiv_ind2 :=
  Induction for tequiv Sort Prop.

Inductive wf_kind : kind -> Prop :=
| wf_kind_simple : forall k, simple_kind k -> wf_kind k

| wf_kind_tcon : forall k k', 
  simple_kind k -> wf_kind k' -> wf_kind (kind_tcon k k')

| wf_kind_dcon : forall t tk k, 
  kinding nil t tk -> wf_kind k -> wf_kind (kind_dcon t k).

End LAIR_PARAMS.

(* The [subst] functions assume that their second argument is locally
   closed. *)
Fixpoint subst_exp (x:atom) (ex:exp) (e:exp) {struct e} : exp :=
  match e with
    | exp_bvar i => exp_bvar i
    | exp_fvar y => if x == y then ex else exp_fvar y
    | exp_base b => exp_base b
    | exp_lam t e1 => exp_lam (subst_typ x ex t) (subst_exp x ex e1)
    | exp_app e1 e2 => exp_app (subst_exp x ex e1) (subst_exp x ex e2)
    | exp_tab kk e1 => exp_tab kk (subst_exp x ex e1)
    | exp_tap e1 t => exp_tap (subst_exp x ex e1) (subst_typ x ex t)
    | exp_new e1 => exp_new (subst_exp x ex e1)
    | exp_bot => exp_bot
    | exp_cert e1 et_l => exp_cert e1 et_l (* (subst_exp x ex e1) (subst_domain x ex et_l) *) (* These are values ... ie. closed exps *)
  end
with subst_typ (x : atom) (ex : exp) (t : typ) {struct t} : typ :=
  match t with 
    | typ_btvar i => typ_btvar i
    | typ_ftvar tv => typ_ftvar tv
    | typ_const c => typ_const c
    | typ_fun t1 t2 => typ_fun (subst_typ x ex t1)  (subst_typ x ex t2) (* fx is only names *)
    | typ_univ kk ef t1 => typ_univ kk ef (subst_typ x ex t1)
    | typ_qtype t1 => typ_qtype (subst_typ x ex t1)
    | typ_app t1 t2 => typ_app (subst_typ x ex t1) (subst_typ x ex t2)
    | typ_dep t1 e2 => typ_dep (subst_typ x ex t1) (subst_exp x ex e2)
    | typ_sing t1 t2 => typ_sing (subst_typ x ex t1) (subst_typ x ex t2)
    | typ_dropped t1 => typ_dropped (subst_typ x ex t1)
  end.

Fixpoint subst_typ_exp (a:atom) (t':typ) (e:exp) {struct e} : exp :=
  match e with
    | exp_bvar i => exp_bvar i
    | exp_fvar y => exp_fvar y
    | exp_base b => exp_base b
    | exp_lam t e1 => exp_lam (subst_typ_typ a t' t) (subst_typ_exp a t' e1)
    | exp_app e1 e2 => exp_app (subst_typ_exp a t' e1) (subst_typ_exp a t' e2)
    | exp_tab kk e1 => exp_tab kk (subst_typ_exp a t' e1)
    | exp_tap e1 t => exp_tap (subst_typ_exp a t' e1) (subst_typ_typ a t' t)
    | exp_new e1 => exp_new (subst_typ_exp a t' e1)
    | exp_bot => exp_bot
    | exp_cert e1 et_l => exp_cert (subst_typ_exp a t' e1) (List.map (subst_typ_eort a t') et_l)
  end
with subst_typ_typ (a : atom) (t' : typ) (t : typ) {struct t} : typ :=
  match t with 
    | typ_btvar i => typ_btvar i
    | typ_ftvar tv => if a == tv then t' else typ_ftvar tv
    | typ_const c => typ_const c
    | typ_fun t1 t2 => typ_fun (subst_typ_typ a t' t1)  (subst_typ_typ a t' t2) (* fx is only names *)
    | typ_univ kk ef t1 => typ_univ kk ef (subst_typ_typ a t' t1)
    | typ_qtype t1 => typ_qtype (subst_typ_typ a t' t1)
    | typ_app t1 t2 => typ_app (subst_typ_typ a t' t1) (subst_typ_typ a t' t2)
    | typ_dep t1 e2 => typ_dep (subst_typ_typ a t' t1) (subst_typ_exp a t' e2)
    | typ_sing t1 t2 => typ_sing (subst_typ_typ a t' t1) (subst_typ_typ a t' t2)
    | typ_dropped t1 => typ_dropped (subst_typ_typ a t' t1)
  end
with subst_typ_eort (a:atom) (t':typ) (eort:exp_or_typ) {struct eort} : exp_or_typ :=
  match eort with 
    | eort_exp e => eort_exp (subst_typ_exp a t' e)
    | eort_typ t => eort_typ (subst_typ_typ a t' t)
  end.

Definition subst_binding (x:atom) (ex:exp) (ab: atom * binding) : (atom * binding) :=
  match ab with 
    | (a, binding_typ t) => (a, binding_typ (subst_typ x ex t))
    | (a, binding_kind k) =>  (a, binding_kind k)
  end.

Definition subst_typ_binding (a:atom) (t:typ) (ab: atom * binding) : (atom * binding) :=
  match ab with 
    | (b, binding_typ t') => (b, binding_typ (subst_typ_typ a t t'))
    | (a, binding_kind k) =>  (a, binding_kind k)
  end.


Definition subst_typ_binding_env (a:atom) (t:typ) (b: binding) : (binding) :=
  match b with 
    | binding_typ t' => binding_typ (subst_typ_typ a t t')
    | binding_kind k =>  binding_kind k
  end.

Definition subst_gamma (x:atom) (ex:exp) (G:gamma) : gamma :=
  List.map (subst_binding x ex) G.

Definition subst_typ_gamma (a:atom) (t:typ) (G:gamma) : gamma :=
  Environment.map (subst_typ_binding_env a t) G.

Scheme exp_ind2 :=
  Induction for exp Sort Prop
with typ_ind2 :=
  Induction for typ Sort Prop
with kind_ind2 :=
  Induction for kind Sort Prop
with exp_or_typ_ind2 :=
  Induction for exp_or_typ Sort Prop.

Combined Scheme exp_typ_mutual_ind from exp_ind2, typ_ind2, kind_ind2, exp_or_typ_ind2.

(* ********************************************************************** *)
(* Some more special tactics *)
(* ********************************************************************** *)
Ltac gather_atoms :=
  let A := gather_atoms_with (fun x : atoms => x) in
  let B := gather_atoms_with (fun x : atom => singleton x) in
  let C := gather_atoms_with (fun x : list (atom * binding) => dom x) in
  constr:(A `union` B `union` C ).

Tactic Notation "pick" "fresh" ident(x) :=
  let L := gather_atoms in
  (pick fresh x for L).

Tactic Notation
      "pick" "fresh" ident(atom_name) "and" "apply" constr(lemma) :=
  let L := gather_atoms in
  pick fresh atom_name excluding L and apply lemma.
