Require Import Lair_Definitions.

(* Hand constructed induction principles to take account of the lists
   (aka domains) within [exp_cert]s. *)
(* This wasn't quite as painful to construct as it looks. I modeled it
   after ltree_ind2 from
   http://www.labri.fr/perso/casteran/CoqArt/induc-fond/SRC/chap14.v,
   just tweaking exp_typ_mutual_ind to get it to work. *)

Definition exp_ind3 :=
fun (P : exp -> Prop) (P0 : typ -> Prop) (P1 : kind -> Prop)
  (P2 : exp_or_typ -> Prop) (Q : domain -> Prop) (f : forall n : nat, P n)
  (f0 : forall a : atom, P a) (f1 : forall a : atom, P (exp_base a))
  (f2 : forall t : typ, P0 t -> forall e : exp, P e -> P (exp_lam t e))
  (f3 : forall e : exp, P e -> forall e0 : exp, P e0 -> P (exp_app e e0))
  (f4 : forall k : kind, P1 k -> forall e : exp, P e -> P (exp_tab k e))
  (f5 : forall e : exp, P e -> forall t : typ, P0 t -> P (exp_tap e t))
  (f6 : forall e : exp, P e -> P (exp_new e)) (f7 : P exp_bot)
  (f8 : forall e : exp, P e -> forall l : domain, Q l -> P (exp_cert e l))
  (f9 : forall n : nat, P0 n) (f10 : forall a : atom, P0 a)
  (f11 : forall a : atom, P0 (typ_const a))
  (f12 : forall t : typ, P0 t -> forall t0 : typ, P0 t0 -> P0 (typ_fun t t0))
  (f13 : forall k : kind,
         P1 k -> forall (e : effect) (t : typ), P0 t -> P0 (typ_univ k e t))
  (f14 : forall t : typ, P0 t -> P0 (typ_qtype t))
  (f15 : forall t : typ, P0 t -> forall t0 : typ, P0 t0 -> P0 (typ_app t t0))
  (f16 : forall t : typ, P0 t -> forall e : exp, P e -> P0 (typ_dep t e))
  (f17 : forall t : typ, P0 t -> forall t0 : typ, P0 t0 -> P0 (typ_sing t t0))
  (f18 : forall t : typ, P0 t -> P0 (typ_dropped t)) 
  (f19 : P1 kind_A) (f20 : P1 kind_U) (f21 : P1 kind_N)
  (f22 : forall k : kind,
         P1 k -> forall k0 : kind, P1 k0 -> P1 (kind_tcon k k0))
  (f23 : forall t : typ, P0 t -> forall k : kind, P1 k -> P1 (kind_dcon t k))
  (f24 : forall e : exp, P e -> P2 (eort_exp e))
  (f25 : forall t : typ, P0 t -> P2 (eort_typ t))
  (f26 : Q nil)
  (f27 : forall eort : exp_or_typ, P2 eort -> forall l : domain, Q l -> Q (eort :: l))
 =>
fix F (e : exp) : P e :=
  match e as e0 return (P e0) with
  | exp_bvar n => f n
  | exp_fvar a => f0 a
  | exp_base a => f1 a
  | exp_lam t e0 => f2 t (F0 t) e0 (F e0)
  | exp_app e0 e1 => f3 e0 (F e0) e1 (F e1)
  | exp_tab k e0 => f4 k (F1 k) e0 (F e0)
  | exp_tap e0 t => f5 e0 (F e0) t (F0 t)
  | exp_new e0 => f6 e0 (F e0)
  | exp_bot => f7
  | exp_cert e0 l =>
      f8 e0 (F e0) l
        ((fix l_ind (d:domain) : Q d :=
           match d return Q d with
             | nil => f26
             | eort :: tl => f27 eort (F2 eort) tl (l_ind tl)
             end) l)
  end
with F0 (t : typ) : P0 t :=
  match t as t0 return (P0 t0) with
  | typ_btvar n => f9 n
  | typ_ftvar a => f10 a
  | typ_const a => f11 a
  | typ_fun t0 t1 => f12 t0 (F0 t0) t1 (F0 t1)
  | typ_univ k e t0 => f13 k (F1 k) e t0 (F0 t0)
  | typ_qtype t0 => f14 t0 (F0 t0)
  | typ_app t0 t1 => f15 t0 (F0 t0) t1 (F0 t1)
  | typ_dep t0 e => f16 t0 (F0 t0) e (F e)
  | typ_sing t0 t1 => f17 t0 (F0 t0) t1 (F0 t1)
  | typ_dropped t0 => f18 t0 (F0 t0)
  end
with F1 (k : kind) : P1 k :=
  match k as k0 return (P1 k0) with
  | kind_A => f19
  | kind_U => f20
  | kind_N => f21
  | kind_tcon k0 k1 => f22 k0 (F1 k0) k1 (F1 k1)
  | kind_dcon t k0 => f23 t (F0 t) k0 (F1 k0)
  end
with F2 (e : exp_or_typ) : P2 e :=
  match e as e0 return (P2 e0) with
  | eort_exp e0 => f24 e0 (F e0)
  | eort_typ t => f25 t (F0 t)
  end
for F.

Definition typ_ind3 :=
fun (P : exp -> Prop) (P0 : typ -> Prop) (P1 : kind -> Prop)
  (P2 : exp_or_typ -> Prop) (Q : domain -> Prop) (f : forall n : nat, P n)
  (f0 : forall a : atom, P a) (f1 : forall a : atom, P (exp_base a))
  (f2 : forall t : typ, P0 t -> forall e : exp, P e -> P (exp_lam t e))
  (f3 : forall e : exp, P e -> forall e0 : exp, P e0 -> P (exp_app e e0))
  (f4 : forall k : kind, P1 k -> forall e : exp, P e -> P (exp_tab k e))
  (f5 : forall e : exp, P e -> forall t : typ, P0 t -> P (exp_tap e t))
  (f6 : forall e : exp, P e -> P (exp_new e)) (f7 : P exp_bot)
  (f8 : forall e : exp, P e -> forall l : domain, Q l -> P (exp_cert e l))
  (f9 : forall n : nat, P0 n) (f10 : forall a : atom, P0 a)
  (f11 : forall a : atom, P0 (typ_const a))
  (f12 : forall t : typ, P0 t -> forall t0 : typ, P0 t0 -> P0 (typ_fun t t0))
  (f13 : forall k : kind,
         P1 k -> forall (e : effect) (t : typ), P0 t -> P0 (typ_univ k e t))
  (f14 : forall t : typ, P0 t -> P0 (typ_qtype t))
  (f15 : forall t : typ, P0 t -> forall t0 : typ, P0 t0 -> P0 (typ_app t t0))
  (f16 : forall t : typ, P0 t -> forall e : exp, P e -> P0 (typ_dep t e))
  (f17 : forall t : typ, P0 t -> forall t0 : typ, P0 t0 -> P0 (typ_sing t t0))
  (f18 : forall t : typ, P0 t -> P0 (typ_dropped t)) 
  (f19 : P1 kind_A) (f20 : P1 kind_U) (f21 : P1 kind_N)
  (f22 : forall k : kind,
         P1 k -> forall k0 : kind, P1 k0 -> P1 (kind_tcon k k0))
  (f23 : forall t : typ, P0 t -> forall k : kind, P1 k -> P1 (kind_dcon t k))
  (f24 : forall e : exp, P e -> P2 (eort_exp e))
  (f25 : forall t : typ, P0 t -> P2 (eort_typ t))
  (f26 : Q nil)
  (f27 : forall eort : exp_or_typ, P2 eort -> forall l : domain, Q l -> Q (eort :: l))
 =>
fix F (e : exp) : P e :=
  match e as e0 return (P e0) with
  | exp_bvar n => f n
  | exp_fvar a => f0 a
  | exp_base a => f1 a
  | exp_lam t e0 => f2 t (F0 t) e0 (F e0)
  | exp_app e0 e1 => f3 e0 (F e0) e1 (F e1)
  | exp_tab k e0 => f4 k (F1 k) e0 (F e0)
  | exp_tap e0 t => f5 e0 (F e0) t (F0 t)
  | exp_new e0 => f6 e0 (F e0)
  | exp_bot => f7
  | exp_cert e0 l =>
      f8 e0 (F e0) l
        ((fix l_ind (d:domain) : Q d :=
           match d return Q d with
             | nil => f26
             | eort :: tl => f27 eort (F2 eort) tl (l_ind tl)
             end) l)
  end
with F0 (t : typ) : P0 t :=
  match t as t0 return (P0 t0) with
  | typ_btvar n => f9 n
  | typ_ftvar a => f10 a
  | typ_const a => f11 a
  | typ_fun t0 t1 => f12 t0 (F0 t0) t1 (F0 t1)
  | typ_univ k e t0 => f13 k (F1 k) e t0 (F0 t0)
  | typ_qtype t0 => f14 t0 (F0 t0)
  | typ_app t0 t1 => f15 t0 (F0 t0) t1 (F0 t1)
  | typ_dep t0 e => f16 t0 (F0 t0) e (F e)
  | typ_sing t0 t1 => f17 t0 (F0 t0) t1 (F0 t1)
  | typ_dropped t0 => f18 t0 (F0 t0)
  end
with F1 (k : kind) : P1 k :=
  match k as k0 return (P1 k0) with
  | kind_A => f19
  | kind_U => f20
  | kind_N => f21
  | kind_tcon k0 k1 => f22 k0 (F1 k0) k1 (F1 k1)
  | kind_dcon t k0 => f23 t (F0 t) k0 (F1 k0)
  end
with F2 (e : exp_or_typ) : P2 e :=
  match e as e0 return (P2 e0) with
  | eort_exp e0 => f24 e0 (F e0)
  | eort_typ t => f25 t (F0 t)
  end
for F0.

Definition kind_ind3 :=
fun (P : exp -> Prop) (P0 : typ -> Prop) (P1 : kind -> Prop)
  (P2 : exp_or_typ -> Prop) (Q : domain -> Prop) (f : forall n : nat, P n)
  (f0 : forall a : atom, P a) (f1 : forall a : atom, P (exp_base a))
  (f2 : forall t : typ, P0 t -> forall e : exp, P e -> P (exp_lam t e))
  (f3 : forall e : exp, P e -> forall e0 : exp, P e0 -> P (exp_app e e0))
  (f4 : forall k : kind, P1 k -> forall e : exp, P e -> P (exp_tab k e))
  (f5 : forall e : exp, P e -> forall t : typ, P0 t -> P (exp_tap e t))
  (f6 : forall e : exp, P e -> P (exp_new e)) (f7 : P exp_bot)
  (f8 : forall e : exp, P e -> forall l : domain, Q l -> P (exp_cert e l))
  (f9 : forall n : nat, P0 n) (f10 : forall a : atom, P0 a)
  (f11 : forall a : atom, P0 (typ_const a))
  (f12 : forall t : typ, P0 t -> forall t0 : typ, P0 t0 -> P0 (typ_fun t t0))
  (f13 : forall k : kind,
         P1 k -> forall (e : effect) (t : typ), P0 t -> P0 (typ_univ k e t))
  (f14 : forall t : typ, P0 t -> P0 (typ_qtype t))
  (f15 : forall t : typ, P0 t -> forall t0 : typ, P0 t0 -> P0 (typ_app t t0))
  (f16 : forall t : typ, P0 t -> forall e : exp, P e -> P0 (typ_dep t e))
  (f17 : forall t : typ, P0 t -> forall t0 : typ, P0 t0 -> P0 (typ_sing t t0))
  (f18 : forall t : typ, P0 t -> P0 (typ_dropped t)) 
  (f19 : P1 kind_A) (f20 : P1 kind_U) (f21 : P1 kind_N)
  (f22 : forall k : kind,
         P1 k -> forall k0 : kind, P1 k0 -> P1 (kind_tcon k k0))
  (f23 : forall t : typ, P0 t -> forall k : kind, P1 k -> P1 (kind_dcon t k))
  (f24 : forall e : exp, P e -> P2 (eort_exp e))
  (f25 : forall t : typ, P0 t -> P2 (eort_typ t))
  (f26 : Q nil)
  (f27 : forall eort : exp_or_typ, P2 eort -> forall l : domain, Q l -> Q (eort :: l))
 =>
fix F (e : exp) : P e :=
  match e as e0 return (P e0) with
  | exp_bvar n => f n
  | exp_fvar a => f0 a
  | exp_base a => f1 a
  | exp_lam t e0 => f2 t (F0 t) e0 (F e0)
  | exp_app e0 e1 => f3 e0 (F e0) e1 (F e1)
  | exp_tab k e0 => f4 k (F1 k) e0 (F e0)
  | exp_tap e0 t => f5 e0 (F e0) t (F0 t)
  | exp_new e0 => f6 e0 (F e0)
  | exp_bot => f7
  | exp_cert e0 l =>
      f8 e0 (F e0) l
        ((fix l_ind (d:domain) : Q d :=
           match d return Q d with
             | nil => f26
             | eort :: tl => f27 eort (F2 eort) tl (l_ind tl)
             end) l)
  end
with F0 (t : typ) : P0 t :=
  match t as t0 return (P0 t0) with
  | typ_btvar n => f9 n
  | typ_ftvar a => f10 a
  | typ_const a => f11 a
  | typ_fun t0 t1 => f12 t0 (F0 t0) t1 (F0 t1)
  | typ_univ k e t0 => f13 k (F1 k) e t0 (F0 t0)
  | typ_qtype t0 => f14 t0 (F0 t0)
  | typ_app t0 t1 => f15 t0 (F0 t0) t1 (F0 t1)
  | typ_dep t0 e => f16 t0 (F0 t0) e (F e)
  | typ_sing t0 t1 => f17 t0 (F0 t0) t1 (F0 t1)
  | typ_dropped t0 => f18 t0 (F0 t0)
  end
with F1 (k : kind) : P1 k :=
  match k as k0 return (P1 k0) with
  | kind_A => f19
  | kind_U => f20
  | kind_N => f21
  | kind_tcon k0 k1 => f22 k0 (F1 k0) k1 (F1 k1)
  | kind_dcon t k0 => f23 t (F0 t) k0 (F1 k0)
  end
with F2 (e : exp_or_typ) : P2 e :=
  match e as e0 return (P2 e0) with
  | eort_exp e0 => f24 e0 (F e0)
  | eort_typ t => f25 t (F0 t)
  end
for F1.

Definition exp_or_typ_ind3 :=
fun (P : exp -> Prop) (P0 : typ -> Prop) (P1 : kind -> Prop)
  (P2 : exp_or_typ -> Prop) (Q : domain -> Prop) (f : forall n : nat, P n)
  (f0 : forall a : atom, P a) (f1 : forall a : atom, P (exp_base a))
  (f2 : forall t : typ, P0 t -> forall e : exp, P e -> P (exp_lam t e))
  (f3 : forall e : exp, P e -> forall e0 : exp, P e0 -> P (exp_app e e0))
  (f4 : forall k : kind, P1 k -> forall e : exp, P e -> P (exp_tab k e))
  (f5 : forall e : exp, P e -> forall t : typ, P0 t -> P (exp_tap e t))
  (f6 : forall e : exp, P e -> P (exp_new e)) (f7 : P exp_bot)
  (f8 : forall e : exp, P e -> forall l : domain, Q l -> P (exp_cert e l))
  (f9 : forall n : nat, P0 n) (f10 : forall a : atom, P0 a)
  (f11 : forall a : atom, P0 (typ_const a))
  (f12 : forall t : typ, P0 t -> forall t0 : typ, P0 t0 -> P0 (typ_fun t t0))
  (f13 : forall k : kind,
         P1 k -> forall (e : effect) (t : typ), P0 t -> P0 (typ_univ k e t))
  (f14 : forall t : typ, P0 t -> P0 (typ_qtype t))
  (f15 : forall t : typ, P0 t -> forall t0 : typ, P0 t0 -> P0 (typ_app t t0))
  (f16 : forall t : typ, P0 t -> forall e : exp, P e -> P0 (typ_dep t e))
  (f17 : forall t : typ, P0 t -> forall t0 : typ, P0 t0 -> P0 (typ_sing t t0))
  (f18 : forall t : typ, P0 t -> P0 (typ_dropped t)) 
  (f19 : P1 kind_A) (f20 : P1 kind_U) (f21 : P1 kind_N)
  (f22 : forall k : kind,
         P1 k -> forall k0 : kind, P1 k0 -> P1 (kind_tcon k k0))
  (f23 : forall t : typ, P0 t -> forall k : kind, P1 k -> P1 (kind_dcon t k))
  (f24 : forall e : exp, P e -> P2 (eort_exp e))
  (f25 : forall t : typ, P0 t -> P2 (eort_typ t))
  (f26 : Q nil)
  (f27 : forall eort : exp_or_typ, P2 eort -> forall l : domain, Q l -> Q (eort :: l))
 =>
fix F (e : exp) : P e :=
  match e as e0 return (P e0) with
  | exp_bvar n => f n
  | exp_fvar a => f0 a
  | exp_base a => f1 a
  | exp_lam t e0 => f2 t (F0 t) e0 (F e0)
  | exp_app e0 e1 => f3 e0 (F e0) e1 (F e1)
  | exp_tab k e0 => f4 k (F1 k) e0 (F e0)
  | exp_tap e0 t => f5 e0 (F e0) t (F0 t)
  | exp_new e0 => f6 e0 (F e0)
  | exp_bot => f7
  | exp_cert e0 l =>
      f8 e0 (F e0) l
        ((fix l_ind (d:domain) : Q d :=
           match d return Q d with
             | nil => f26
             | eort :: tl => f27 eort (F2 eort) tl (l_ind tl)
             end) l)
  end
with F0 (t : typ) : P0 t :=
  match t as t0 return (P0 t0) with
  | typ_btvar n => f9 n
  | typ_ftvar a => f10 a
  | typ_const a => f11 a
  | typ_fun t0 t1 => f12 t0 (F0 t0) t1 (F0 t1)
  | typ_univ k e t0 => f13 k (F1 k) e t0 (F0 t0)
  | typ_qtype t0 => f14 t0 (F0 t0)
  | typ_app t0 t1 => f15 t0 (F0 t0) t1 (F0 t1)
  | typ_dep t0 e => f16 t0 (F0 t0) e (F e)
  | typ_sing t0 t1 => f17 t0 (F0 t0) t1 (F0 t1)
  | typ_dropped t0 => f18 t0 (F0 t0)
  end
with F1 (k : kind) : P1 k :=
  match k as k0 return (P1 k0) with
  | kind_A => f19
  | kind_U => f20
  | kind_N => f21
  | kind_tcon k0 k1 => f22 k0 (F1 k0) k1 (F1 k1)
  | kind_dcon t k0 => f23 t (F0 t) k0 (F1 k0)
  end
with F2 (e : exp_or_typ) : P2 e :=
  match e as e0 return (P2 e0) with
  | eort_exp e0 => f24 e0 (F e0)
  | eort_typ t => f25 t (F0 t)
  end
for F2.

Combined Scheme exp_typ_mutual_ind2 from exp_ind3, typ_ind3, kind_ind3, exp_or_typ_ind3.
