Require Export Lair_Model.

Module LAIR_Infrastructure (Import X : LAIR_PARAMS).

  Lemma weakening : forall G1 G2 A1 A2 A e t, 
    wf_env G1 A1 -> splitA A1 A2 A -> wf_env (G2 ++ G1) A -> 
    typing G1 A1 e t -> typing (G2 ++ G1) A e t.
  Admitted.

  Lemma typing_wf_typ : forall G A e t fx, 
    typing G A e (t, fx) -> wf_typ t G.
  Admitted.

  Lemma typing_kinded_typ : forall G A e t fx, 
    typing G A e (t, fx) -> exists k, kinding G t k.
  Admitted.

  Lemma kinding_wf_typ : forall G t k, 
    kinding G t k -> wf_typ t G. 
  Admitted.

  Lemma kinding_wf_kind : forall G t k, 
    kinding G t k -> wf_kind k.
  Admitted.

  Lemma kinding_uniqueness : forall G t k k', 
    kinding G t k -> kinding G t k' -> k = k'.
  Admitted.

  Lemma typing_effect_ok : forall G A e t fx, 
    typing G A e (t, fx) -> effect_ok fx G.
  Admitted.

  Lemma typing_lc : forall (G:gamma) (A:atoms) (e:exp) (t:typ) (fx:atoms),
    wf_env G A -> typing G A e (t,fx) -> lc_exp e.
  Admitted.

  Lemma typing_wf_exp : forall G A e t fx, 
    typing G A e (t,fx) -> wf_exp e G.
  Admitted.

  Lemma effect_ok_subst_typ_gamma : forall G1 G2 A a (b:atom) fx0 fx_e, 
      wf_env G1 {} -> kinding G1 b kind_N -> b `notin` fx0 -> 
      splitA (singleton a) fx0 fx_e -> wf_env (G2 ++ [(a,binding_kind kind_N)] ++ G1) A ->
      effect_ok fx_e (G2 ++ [(a,binding_kind kind_N)] ++ G1) ->  b <> a -> 
      effect_ok (singleton b `union` fx0)(subst_typ_gamma a b G2 ++ G1).
  Proof.
     intros until 1. rename H into wf_G1.
     intros k_b b_notin_fx split_fx wf_G e0 b_neq_a.
   unfold effect_ok in e0.
   unfold effect_ok.
   intros.
   assert (a0 <> a).
     unfold not. intro.
     subst a0.
     assert (a `notin` fx0). 
       inversion split_fx. eapply inter_notin; eauto.
     destruct (AtomSet.F.union_1 H).
       destruct b_neq_a. fsetdec.
       fsetdec.
   assert ((binds a0 (binding_kind kind_N) (G2 ++ G1)) -> 
           (binds a0 (binding_kind kind_N) (subst_typ_gamma a b G2 ++ G1))) as binds_imp.
     intro b_g2_g1.
     destruct (binds_concat_inv  _ _ _ _ _ b_g2_g1) as [[not_in_g2 binds_g1] | binds_g2].
       rewrite (subst_typ_gamma_dom G2 a b) in not_in_g2.
       eapply binds_tail; eauto.
       eapply binds_head; eauto using Gamma_subst_typ_binding_kind.
   eapply binds_imp. clear binds_imp.
     destruct (AtomSet.F.union_1 H).
       SSSCase "a0 = b".
       assert (a0 = b). fsetdec. subst a0.
       eapply binds_tail.
         inversion k_b. eauto.
         inversion k_b. rename H4 into binds_b_G1.
          destruct (binds_exists_singleton _ _ _ binds_b_G1) as [G1_1 [G2_1 G1_eq]].
          assert (b `notin` (dom (G2 ++ [(a, binding_kind kind_N)] ++ G1_1))).
           pose proof (wf_is_ok _ _ wf_G) as ok.
           assert (binds b (binding_kind kind_N) ((G2 ++ [(a, binding_kind kind_N)]) ++ G1)).
             eapply binds_weaken_at_head; simpl_env; eauto.
           rewrite G1_eq in H4, ok.
           rewrite_env ((G2 ++ [(a, binding_kind kind_N)] ++ G1_1) ++ [(b, binding_kind kind_N)] ++ G2_1) in H4.
           rewrite_env ((G2 ++ [(a, binding_kind kind_N)] ++ G1_1) ++ [(b, binding_kind kind_N)] ++ G2_1) in ok.
           eapply fresh_mid_head; eauto.
           simpl_env in H4. notin_solve.
       SSCase "a0 in fx0".
        assert (a0 `in` fx_e) as a0_in_fx_e.
          inversion split_fx. fsetdec.
        eapply binds_remove_mid; eauto.
     Qed.

End LAIR_Infrastructure.

Lemma effect_ok_weakening : forall G1 G2 fx A, 
  effect_ok fx G1 -> wf_env (G2 ++ G1) A -> effect_ok fx (G2 ++ G1).
Admitted.

Lemma effect_ok_subst : forall G1 x tx G2 fx1 fx2 A fx v, 
  wf_env (G2 ++ [(x, binding_typ tx)] ++ G1) A ->
  effect_ok fx1 (G2 ++ [(x, binding_typ tx)] ++ G1) ->
  effect_ok fx2 G1 ->
  splitA fx1 fx2 fx -> 
  effect_ok fx (subst_gamma x v G2 ++ G1).
Admitted.
