Require Export Lair_Environment.

Notation model := (list (atom * (list dom_eqn))).

Module LAIR_Model (Import X : LAIR_PARAMS).

Inductive dom_eq_consistent : atom -> list dom_eqn -> Prop :=
| DE_nil : forall b, dom_eq_consistent b nil
| DE_cons : forall (b:atom) (d:domain) (e:exp) (tl:list dom_eqn) tfx, 
  typing nil {} (exp_cert b d) tfx ->
  typing nil {} e tfx -> 
  dom_eq_consistent b tl -> 
  dom_eq_consistent b ((d,e)::tl).
  
Inductive MS_consistent : model -> gamma -> Prop :=
| MS_base : forall (M:model), MS_consistent M nil
| MS_cons : 
  forall (M:model) (S:gamma) (b:atom) de_l (t:typ),
    MS_consistent M S -> (binds b de_l M) -> 
    dom_eq_consistent b de_l ->
    MS_consistent M ([(b, binding_typ t)] ++ S).

Definition wf_signature s := 
  forall b t K, ((binds b (binding_typ t) s) -> (wf_typ t nil)) /\ 
                ((binds b (binding_kind K) s) -> (wf_kind K)).


Lemma MS_binding : forall (m:model) (s:gamma) (a:atom) (t:typ),
  MS_consistent m s -> (binds a (binding_typ t) s) -> (exists de, binds a de m).
Admitted.

Lemma cert_is_value : forall (G:gamma) (A:atoms) (B:atom) (d:domain) (t:typ) fx, 
  wf_env G A -> only_names G A -> typing G A (exp_cert (exp_base B) d) (t,fx) -> 
  value (exp_cert (exp_base B) d).
Admitted.

Lemma cert_is_wf : forall G A b d t fx, 
  typing G A (exp_cert b d) (t, fx) -> (exists G', wf_exp (exp_cert b d) G' /\ only_names G' {}) .
Admitted. 

Lemma sig_binds_certB : forall (G:gamma) (A:atoms) (B:atom) (d:domain) (t:typ) fx, 
  typing G A (exp_cert (exp_base B) d) (t,fx) -> (exists t, binds B (binding_typ t) Signature).
Admitted.

Lemma in_eqns_consistency : forall G M S (b:atom) de_l d e tfx,
  MS_consistent M S -> binds b de_l M -> in_eqns d de_l e -> 
  typing G {} (exp_cert (exp_base b) d) tfx ->
  typing G {} e tfx.
Admitted. 

End LAIR_Model.

Lemma classical_eqns : forall (d:domain) (e: list dom_eqn), 
  (not_in_eqns d e) \/ (exists e', in_eqns d e e').
Admitted. 
