Require Import Lair_Tequiv.

Module LAIR_Progress (Import X : LAIR_PARAMS).
  Module Export TequivFacts := LAIR_Tequiv X.

Lemma canonical_lam0 : forall (G:gamma) (A:atoms) (e:exp) (t1:typ) (t2:typ) fx,  
  value e -> typing G A e ((typ_fun t1 t2), fx) -> 
  (exists e', exists tx', e = (exp_lam tx' e')) \/ (exists B, exists d, e = (exp_cert (exp_base B) d)).
intros until 1. rename H into Val_e.
intro T.
remember (typ_fun t1 t2, fx) as tfx.
revert t1 t2 fx Heqtfx.
induction T; try solve [inversion Val_e].
Case "type_NEW".
 intros. discriminate.
Case "type_ABS".
intros.
left. exists e.
injection Heqtfx; intros; subst.
           destruct (ptype_cases (typ_fun t_x t_e) A fx0) as [LL1 | RR1].
             rewrite LL1 in H3. injection H3. intros; subst. exists t1. reflexivity.
             rewrite RR1 in H3. discriminate.
Case "type_TAB".
intros.
        destruct (ptype_cases (typ_univ k NO_FX t_e) A fx) as [LL1 | RR1]. 
          rewrite LL1 in Heqtfx. discriminate.
          rewrite RR1 in Heqtfx; discriminate.
Case "type_TAB_FX".
intros.
        destruct (ptype_cases (typ_univ k FX t_e) A fx) as [LL1 | RR1]. 
          rewrite LL1 in Heqtfx. discriminate.
          rewrite RR1 in Heqtfx; discriminate.
Case "type_DROPPED".
intros. discriminate.
Case "type_AFN".
  intros.  injection Heqtfx; intros; subst. 
  eapply IHT; eauto.
Case "type_B1".
  intros; injection Heqtfx; intros; subst.
  right. exists B. exists (eort_exp v :: domain). reflexivity.
Case "type_B2".
  intros.  
  right. exists B. exists (eort_typ t' :: domain). reflexivity.
Case "type_B3".
  intros.  
  right. exists B. exists (nil:domain). reflexivity.
Case "type_CONV".
intros.
injection Heqtfx. intros; subst.
pose proof (tequiv_preserves_fun_shape _ _ _  H) as tfshape.
destruct tfshape as [t1' [t2' t_eq]].
eapply IHT; eauto. subst. reflexivity.
Qed.

Lemma canonical_lam1 : forall (G:gamma) (A:atoms) (e:exp) (t1:typ) (t2:typ) fx,  
  value e -> typing G A e (typ_qtype (typ_fun t1 t2), fx) -> 
  (exists e', exists tx', e = (exp_lam tx' e')) \/ (exists B, exists d, e = (exp_cert (exp_base B) d)).
intros until 1. rename H into Val_e.
intro T.
remember (typ_qtype (typ_fun t1 t2), fx) as tfx.
revert t1 t2 fx Heqtfx.
induction T; try solve [inversion Val_e].
Case "type_NEW".
 intros. discriminate.
Case "type_ABS".
intros.
left. exists e.
injection Heqtfx; intros; subst.
           destruct (ptype_cases ((typ_fun t_x t_e)) A fx0) as [LL1 | RR1].
             rewrite LL1 in H3. discriminate. 
             rewrite RR1 in H3. injection H3. intros; subst. exists t1. reflexivity. 
Case "type_TAB".
intros.
        destruct (ptype_cases (typ_univ k NO_FX t_e) A fx) as [LL1 | RR1]. 
          rewrite LL1 in Heqtfx. discriminate.
          rewrite RR1 in Heqtfx; discriminate.
Case "type_TAB_FX".
intros.
        destruct (ptype_cases (typ_univ k FX t_e) A fx) as [LL1 | RR1]. 
          rewrite LL1 in Heqtfx. discriminate.
          rewrite RR1 in Heqtfx; discriminate.
Case "type_DROPPED". 
intros; discriminate.
Case "type_AFN".
  intros.  injection Heqtfx; intros; subst. 
  eapply IHT; eauto.
Case "type_B1".
  intros; injection Heqtfx; intros; subst.
  right. exists B. exists (eort_exp v :: domain). reflexivity.
Case "type_B2".
  intros.  
  right. exists B. exists (eort_typ t' :: domain). reflexivity.
Case "type_B3".
  intros.  
  right. exists B. exists (nil:domain). reflexivity.
Case "type_CONV".
intros.
injection Heqtfx. intros; subst.
pose proof (tequiv_preserves_qtyp_shape _ _ _ H) as qtshape.
destruct qtshape as [t1' [t2' t_eq]]. subst.
eapply IHT; eauto.
Qed.


Lemma canonical_lam : forall (G:gamma) (A:atoms) (e:exp) (tx:typ) (t:typ) te fx,  
  value e -> (te = (typ_fun tx t) \/ te = typ_qtype(typ_fun tx t)) -> typing G A e (te,fx) -> 
  (exists e', exists tx', e = (exp_lam tx' e')) \/ (exists B, exists d, e = (exp_cert (exp_base B) d)).
intros until 1. rename H into Val_e.
intros disj T.
destruct disj as [L|R]; [subst te; eapply canonical_lam0; eauto | subst te; eapply canonical_lam1; eauto ].
Qed.

Lemma canonical_tab : forall (G:gamma) (A:atoms) (e:exp) (k:kind) (t:typ) te ef fx,
  value e -> (te = (typ_univ k ef t) \/ te = typ_qtype (typ_univ k ef t)) -> typing G A e (te,fx) -> 
  (exists e', e = (exp_tab k e')) \/ (exists B, exists d, e = (exp_cert (exp_base B) d)).
Admitted.

  Module Export ModelFacts := LAIR_Model X.

Theorem progress : forall (G:gamma) (A:atoms) (e:exp) (t:typ) (fx:atoms),
  MS_consistent Model Signature -> wf_env G A -> only_names G A -> 
  typing G A e (t,fx) -> 
  (exists e', eval e e') \/ (value e) \/ (e = exp_bot).
intros G A e t fx MS WF ON T.
assert (typing G A e (t,fx)); auto.
remember (t, fx) as tfx.
generalize dependent t.
generalize dependent fx.
induction T.
Case "typing_X".
intros.
assert (~ binds x (binding_typ t) G) as NB.
  eapply binds_no_vars. apply ON.
elim NB. assumption.
Case "typing_XA".
intros.
assert (~ binds x (binding_typ t) G) as NB.
  eapply binds_no_vars; apply ON.
elim NB. assumption.
Case "typing_NEW".
(* pose proof (only_names_split2 G (singleton a) A A' ON H2) as ONA. *)
destruct (IHT WF ON T fx t (refl_equal (t, fx))) as [Esteps | [Evalue | Ebot]]. auto.
  (* Sub-Case Esteps *)
  intros.
  elim Esteps. intros x Ex. left. exists (exp_new x).
  apply eval_new1. assumption.
  (* Sub-case Evalue *)
  intros.
  right. left. apply value_new. assumption.
  (* Sub-case Ebot *)
  intros.
  left. exists exp_bot. rewrite Ebot. apply eval_new1_bot. apply eval_bot_bot.
Case "typing_ABS".
intros.
right. left. apply value_abs.
eapply typing_lc. apply WF. apply H.
Case "typing_TAB".
intros.
right. left. apply value_tab.
eapply typing_lc. apply WF. apply H.
Case "typing_TAB_FX".
intros.
right. left. apply value_tab.
eapply typing_lc. apply WF. apply H.
Case "typing_APP".
intros.
left.
pose proof (wf_split1 G A1 A2 A WF H0) as wfGA1.
pose proof (wf_split2 G A1 A2 A WF H0) as wfGA2.
pose proof (only_names_split1 G A1 A2 A ON H0) as ONA1.
pose proof (only_names_split2 G A1 A2 A ON H0) as ONA2.
pose proof (refl_equal (t_e1, fx1)) as tfx_eq.
destruct (IHT1 wfGA1 ONA1 T1 _ _ tfx_eq) as [E1steps | [E1val | E1bot]]; auto. 
(* E1steps *)
elim E1steps. intros x E1x.
exists (exp_app x e2). apply eval_app1. assumption.
(* E1val *)
destruct (IHT2 wfGA2 ONA2 T2 _ _ (refl_equal (t2, fx2))) as [E2steps | [E2val | E2bot]]; auto.
   (* E2steps *)
    elim E2steps. intros x E2x. 
    exists (exp_app e1 x). apply eval_app2. assumption.
   (* E2val *)
   destruct (canonical_lam G A1 e1 _ _ _ _ E1val H1 T1) as [[x [tx' E1isLAMx]] | E1isCERT].
     (* E1 is lam *)
     exists (open_exp 0 e2 x). rewrite E1isLAMx. apply eval_beta. assumption.
     (* E1 is cert *)
     destruct E1isCERT as [B [D E1isBD]].
     rewrite E1isBD. subst.   
     destruct (sig_binds_certB _ _  _ _ _ _ T1) as [ Bt SigBindsB ].
     destruct (MS_binding Model Signature B Bt MS SigBindsB) as [De ModBindsB].
     destruct (classical_eqns ((eort_exp e2)::D) De) as [ NotInEqns | InEqns ].
     exists (exp_cert (exp_base B) ((eort_exp e2)::D)).
     eapply eval_b2.
     eapply (cert_is_value G A1 B D t_e1); auto. apply T1.
     apply E2val. apply ModBindsB. apply NotInEqns.
     destruct InEqns as [X EqReducesX].
     exists X. eapply eval_b1.
     eapply (cert_is_value G A1 B D t_e1); auto. apply T1.
     apply E2val. apply ModBindsB. apply EqReducesX. 
   (* E2bot *)
   exists exp_bot. eapply eval_app2_bot. rewrite E2bot; apply eval_bot_bot.
(* E1bot *)
exists exp_bot. eapply eval_app1_bot. rewrite E1bot; apply eval_bot_bot.
Case "typing_TAP".
intros.
left. 
destruct (IHT WF ON T _ _ (refl_equal (te, fx))) as [Esteps | [Eval | Ebot]]; auto.
(* Esteps *)
destruct Esteps as [x Ex].
exists (exp_tap x t).
apply eval_tap1. assumption.
(* Eval *)
destruct (canonical_tab G A e _ _ _ _ _ Eval H0 T) as [EisTAB | EisCERT].
  (* EisTAB *)
  destruct EisTAB as [x Etabx].
  exists (open_exp_typ 0 t x).
  rewrite Etabx. apply eval_tap.
  (* E is cert *)
  destruct EisCERT as [ B [D EisBD]].
  subst. 
  destruct (sig_binds_certB _  _ _ _ _ _ T) as [Bt SigBindsB].
  destruct (MS_binding Model Signature B Bt MS SigBindsB) as [De ModBindsB].
  destruct (classical_eqns ((eort_typ t)::D) De) as [ NotInEqns | InEqns ].
  exists (exp_cert (exp_base B) ((eort_typ t)::D)).
  eapply eval_b4.
  eapply cert_is_value. apply WF. apply ON. apply T. apply ModBindsB. apply NotInEqns.
  destruct InEqns as [X EqReducesX].
  exists X.
  eapply eval_b3.
  eapply cert_is_value. apply WF. apply ON. apply T. apply ModBindsB. apply EqReducesX.
(* Ebot *)
exists (exp_bot). rewrite Ebot. eapply eval_tap1_bot. apply eval_bot_bot.
Case "typing_TAP_FX".
intros.
left. 
destruct (IHT WF ON T _ _ (refl_equal (te, fx))) as [Esteps | [Eval | Ebot]]; auto.
(* Esteps *)
destruct Esteps as [x Ex].
exists (exp_tap x a).
apply eval_tap1. assumption.
(* Eval *)
destruct (canonical_tab G A e _ _ _ _ _ Eval H0 T) as [EisTAB | EisCERT].
  (* EisTAB *)
  destruct EisTAB as [x Etabx].
  exists (open_exp_typ 0 a x).
  rewrite Etabx. apply eval_tap.
  (* E is cert *)
  destruct EisCERT as [ B [D EisBD]].
  subst. 
  destruct (sig_binds_certB _  _ _ _ _ _ T) as [Bt SigBindsB].
  destruct (MS_binding Model Signature B Bt MS SigBindsB) as [De ModBindsB].
  destruct (classical_eqns ((eort_typ a)::D) De) as [ NotInEqns | InEqns ].
  exists (exp_cert (exp_base B) ((eort_typ a)::D)).
  eapply eval_b4.
  eapply cert_is_value. apply WF. apply ON. apply T. apply ModBindsB. apply NotInEqns.
  destruct InEqns as [X EqReducesX].
  exists X.
  eapply eval_b3.
  eapply cert_is_value. apply WF. apply ON. apply T. apply ModBindsB. apply EqReducesX.
(* Ebot *)
exists (exp_bot). rewrite Ebot. eapply eval_tap1_bot. apply eval_bot_bot.
Case "typing_BOT".
intros.
right; right. auto.
Case "typing_DROP".
intros.
eapply IHT; auto.
Case "typing_AFN".
intros.
pose proof (only_names_split1 G A1 A2 A ON H0) as ONA1.
pose proof (wf_split1 G A1 A2 A WF H0) as wfGA1.
eapply (IHT wfGA1 ONA1 T fx1 t (refl_equal (t, fx1))).
Case "typing_B".
intros.
left. inversion MS.
(* subcase signature empty *)
subst. rewrite <- H3 in H0. inversion H0.
(* subcase steps *)
assert (exists de, (binds B de Model)) as DE.
eapply MS_binding. apply MS. apply H0.
elim DE. intros x MbindsX.
exists (exp_cert (exp_base B) nil).
eapply eval_delta. apply MbindsX.
Case "typing_B1".
right; left.
eapply cert_is_value. subst. apply WF. apply ON. subst. apply H.
Case "typing_B2".
right; left.
eapply cert_is_value. subst. apply WF. apply ON. subst. apply H.
Case "typing_B3".
right; left.
eapply value_cert. eapply typing_lc. apply WF. apply T. constructor.
Case "typing_CONV".
intros.
eapply (IHT WF ON T _ _ (refl_equal (t, fx))). 
Qed.

End LAIR_Progress.
