Require Export Lair_Tequiv.

Module LAIR_Subst (Import X : LAIR_PARAMS).

Lemma inversion_sing : forall G A e t a fx, 
  value e -> typing G A e (typ_sing t a, fx) -> 
  ((exists e', e = exp_new e') /\ fx <> {}).
Admitted.

  Module Export TequivFacts := LAIR_Tequiv X.

Lemma inversion_empty_effect : forall G v t, 
  value v -> typing G {} v (t, {}) -> kinding G t kind_U.
Proof with inversion H.
intros.
assert (typing G {} v (t, {})) as T; auto.
remember (t, {}) as t_fx.
remember ({}) as fx.
generalize dependent t.
induction H0.
Case "type_X"...
Case "type_XA"...
Case "type_NEW"...
 intros. subst A. injection Heqt_fx; intros; subst a_fx.
  inversion H3. assert (a `in` {}). rewrite <- H6. fsetdec. fsetdec.
Case "type_ABS".
 intros; injection Heqt_fx; intros; subst A. subst fx. 
   rewrite H4 in T.
   destruct (typing_kinded_typ _ _ _ _ _ T) as [k_t K].
     unfold ptype in H4.
     case_eq (AtomSet.F.is_empty ({} `union` {})).
      intro empty_true. 
        rewrite empty_true in H4.
        subst t. inversion K. subst k_t; auto.
      intro empty_false.
        destruct (non_empty_1  _ empty_false) as [X X_in_nul]. fsetdec.
Case "type_TAB".
 intros; injection Heqt_fx; intros; subst A. subst fx. 
   rewrite H4 in T.
   destruct (typing_kinded_typ _ _ _ _ _ T) as [k_t K].
     unfold ptype in H4.
     case_eq (AtomSet.F.is_empty ({} `union` {})).
      intro empty_true. 
        rewrite empty_true in H4.
        subst t. inversion K. subst k_t; auto.
      intro empty_false.
        destruct (non_empty_1  _ empty_false) as [X X_in_nul]. fsetdec.
Case "type_TAB_FX".
 intros. injection Heqt_fx. intros eq_fx eq_t; subst A. subst fx.
   rewrite eq_t in T.
   destruct (typing_kinded_typ _ _ _ _ _ T) as [k_t K].
     unfold ptype in eq_t.
     case_eq (AtomSet.F.is_empty ({} `union` {})).
      intro empty_true. 
        rewrite empty_true in eq_t.
        subst t. inversion K. subst k_t; auto.
      intro empty_false.
        destruct (non_empty_1  _ empty_false) as [X X_in_nul]. fsetdec.
Case "type_APP"...
Case "type_TAP_NOFX"...
Case "type_TAP_FX"...
Case "type_BOT"...
Case "type_DROP".
  intros. injection Heqt_fx; intros; subst A; subst t0. subst fx.
  destruct (inversion_sing _ _ _ _ _ _ H H0) as [[e' e_eq_newe'] fx_neq_nul].
    destruct fx_neq_nul. auto.
Case "type_AFN".
  intros. injection Heqt_fx; intros; subst A; subst t0. subst fx.
  destruct (split_nil_nil A1 A2 H0).
  destruct (split_nil_nil fx1 fx2 H2). subst fx1. subst A1.
  eapply IHtyping; eauto.
Case "type_B"...
Focus 4.
Case "type_CONV".
  intros; injection Heqt_fx; intros; subst A; subst t0; subst fx.
  eapply tequiv_preserves_kinding; eauto using H1.
Case "type_B1". (* handled by well-formedness of signature : TODO *)
Admitted.
  
Lemma inversion_nonempty_effect : forall G A e t fx,
  (value e) ->
  typing G A e (t, fx) -> fx <> {} -> 
  (kinding G t kind_A \/ typing G A e (t, {})).
Proof with inversion H. 
intros.
assert (typing G A e (t, fx)) as T; auto.
remember (t, fx) as t_fx.
generalize dependent t.
generalize dependent fx.
induction H0.
Case "type_X"...
Case "type_XA"...
Case "type_NEW".
  intros.
  injection Heqt_fx; intros; subst t0.
  left.
  eapply kinding_N.
  eapply kinding_AFN. auto.
  eapply kinding_A; auto.
Case "type_ABS".
  intros.
  injection Heqt_fx; intros; subst t; subst fx0.
  left.
  unfold ptype.
  case_eq (AtomSet.F.is_empty (A `union` fx)).
  intro u_empty.
    destruct (empty_union _ _ u_empty) as [_ fx_empty]. 
    contradiction.
  intro.
   unfold ptype in T.
     rewrite H4 in T.
   destruct (typing_kinded_typ _ _ _ _ _ T) as [kk t_kind_k].
   inversion t_kind_k. subst kk; auto.
Case "type_TAB".
  intros.
  injection Heqt_fx. intros eq_fx eq_t. subst t. subst fx0.
  left.
  unfold ptype.
  case_eq (AtomSet.F.is_empty (A `union` fx)).
  intro u_empty.
    destruct (empty_union _ _ u_empty) as [_ fx_empty]; contradiction.
  intro.
    unfold ptype in T.
      rewrite H4 in T.
    destruct (typing_kinded_typ _ _ _ _ _ T) as [kk t_kind_k].
    inversion t_kind_k; subst kk; auto.
Case "type_TAB_FX".
  intros.
  injection Heqt_fx; intros; subst t. subst fx.
  left.
  unfold ptype.
  case_eq (AtomSet.F.is_empty (A `union` fx0)).
  intro u_empty.
    destruct (empty_union _ _ u_empty) as [_ fx_empty]; contradiction.
  intro.
    unfold ptype in T.
      rewrite H5 in T.
    destruct (typing_kinded_typ _ _ _ _ _ T) as [kk t_kind_k].
    inversion t_kind_k; subst kk; auto.
Case "type_APP"...
Case "type_TAP_NOFX"...
Case "type_TAP_FX"...
Case "type_BOT"...
Case "type_DROP".
  intros. injection Heqt_fx; intros; subst t0; subst fx0.
  destruct (typing_kinded_typ _ _ _ _ _ H0) as [kk t_kind_k].
  inversion t_kind_k.
  destruct (typing_kinded_typ _ _ _ _ _ T) as [kk' t_kind_d].
  inversion t_kind_d. subst.
  pose proof (kinding_uniqueness _ _ _ _ H5 H10); subst.
  left; auto.
Case "type_AFN".
  intros; injection Heqt_fx; intros; subst t0; subst fx0.
  (* Hmm ... not sure how to proceed here without excluded middle *)
  destruct (exclude_middle_empty_sets fx1) as [fx1_empty | fx1_nonempty].
  subst fx1; right.
    eapply type_AFN. apply H0. apply H1. apply (split_nil_A_1 {}).
    unfold effect_ok. fsetdec.
  destruct (IHtyping H H1 fx1 fx1_nonempty t (refl_equal (t, fx1))) as [KK | TT].
    left; auto.
    right.
      eapply type_AFN. apply H0. apply TT. apply (split_nil_A_1 {}).
      unfold effect_ok. fsetdec.
Case "type_B"...
Focus 4.
Case "type_CONV".
  intros; injection Heqt_fx; intros; subst t0; subst fx0.
  destruct (IHtyping H H0 fx H2 t (refl_equal (t,fx))) as [ KK | TT ].
  left.
    eapply (tequiv_preserves_kinding G). apply KK. apply H1.
  right.
    eapply type_CONV. apply TT. apply H1.
Case "type_B1". (* remaining cases handled by well-formedness of types in a Signature *)
(* TODO *)
Admitted.

  Module Export ModelFacts := LAIR_Model X.

(* Separate lemma instead of complicating the main subst lemma with disjuncts *)
Lemma nofree_subst_preserves_typing : forall G1 G2 A x tx e t v fx_e fx_v fx_ev ka,
  wf_signature Signature ->
  wf_env G1 {} -> kinding G1 tx ka -> (x `notin` A) -> 
  wf_env (G2 ++ [(x, binding_typ tx)] ++ G1) A ->
  typing (G2 ++ [(x, binding_typ tx)] ++ G1) A e (t, fx_e) ->
  value v -> typing G1 {} v (tx, fx_v) -> splitA fx_v fx_e fx_ev ->
  typing ((subst_gamma x v G2) ++ G1) A (subst_exp x v e) ((subst_typ x v t), fx_e).
Admitted.

  Module Export UtilFacts := LAIR_Util X.

(* Main substitution lemma *)
Lemma subst_preserves_typing : forall G1 G2 A x tx k e t v fx_e fx_v fx_ev, 
  wf_signature Signature -> 
  wf_env G1 {} -> kinding G1 tx k -> (x `notin` A) -> (* wf_env (G2 ++ G1) A ->  *)
  wf_env (G2 ++ [(x, binding_typ tx)] ++ G1) (aqual x k A) ->
  typing (G2 ++ [(x, binding_typ tx)] ++ G1) (aqual x k A) e (t, fx_e) ->
  value v -> typing G1 {} v (tx, fx_v) -> splitA fx_v fx_e fx_ev ->
  typing ((subst_gamma x v G2) ++ G1) A (subst_exp x v e) ((subst_typ x v t), fx_ev).
Proof.
intros until 1. rename H into wf_sig.
intros wf_G1 Kind_txk x_notin_A WF_G1xG2Ax T Val_v T_v split_fx.
assert (typing (G2 ++ [(x, binding_typ tx)] ++ G1) (aqual x k A) e (t, fx_e)) as Te; auto.
remember (G2 ++ [(x, binding_typ tx)] ++ G1) as G.
remember (aqual x k A) as Ax.
remember (t, fx_e) as t_fxe.
assert (wf_env G Ax) as WF_GAx; auto.
generalize dependent Te. 
generalize dependent HeqG.
generalize dependent HeqAx.
generalize dependent WF_GAx.
generalize dependent A.
generalize dependent t.
generalize dependent fx_v.
generalize dependent fx_e.
generalize dependent fx_ev.
generalize dependent G1.
generalize dependent G2.
(* Check typing_ind2. (* probably should gen dep k as well *) *)
destruct T using typing_ind2 with
  (P0 := 
    fun (G:gamma) (t:typ) (k:kind) (K:kinding G t k) =>
      forall G1 G2 fx_e fx_v fx_ev k_v, (G = (G2 ++ [(x, binding_typ tx)] ++ G1)) -> wf_env G {} -> 
        wf_env G1 {} ->  kinding G1 tx k_v -> typing G1 {} v (tx, fx_v) -> 
        splitA fx_v fx_e fx_ev -> kinding ((subst_gamma x v G2) ++ G1) (subst_typ x v t) k) 
  (P1 :=
    fun (t:typ) (t':typ) (Te: tequiv t t') =>
      forall x v, tequiv (subst_typ x v t) (subst_typ x v t')).
Case "typing_X".
intros G2 G1 wf_G1 Kind_txk fx_ev fx_e fx_v T_v split_fx t0 t_fxe A x_notin_A wf_GEmpty AxEmpty G_eq_G2xG1 T_x0.
injection t_fxe; intros. subst t0. subst fx_e.
pose proof (Ax_empty_A_empty A x k (sym_equal AxEmpty)) as A_Empty.
rewrite G_eq_G2xG1 in b.
destruct (binds_concat_inv binding x0 (binding_typ t) _ G2 b) as [ [x0_Notin_G2 binds_x0_xG1] | binds_x0_G2].
destruct (binds_concat_inv binding x0 (binding_typ t) G1 _ binds_x0_xG1) as [ [ x0_Notin_x binds_x0_G1 ] | binds_x0_x].
(* x0 in G1 *)
simpl.
destruct (x == x0) as [x_eq_x0 | x_new_x0].
subst. destruct x0_Notin_x. simpl. fsetdec.
rewrite A_Empty. 
   eapply type_AFN.
   eapply split_nil_A_1. (* first premise of T-AFN: {},{} *)
   eapply type_X. (* second premise : IH *)
      assert ((subst_typ x v t) = t) as subt_eq_t.
         assert (wf_typ t G1) as wf_t_G1.
           eapply binds_wf. apply wf_G1. apply binds_x0_G1.
         eapply subst_wf_typ_ext. apply wf_t_G1. eapply binds_disjoint_1.
         rewrite G_eq_G2xG1 in WF_G1xG2Ax. apply WF_G1xG2Ax.
      rewrite subt_eq_t.
      assert (dom G2 = dom (subst_gamma x v G2)) as dom_G2.
         eapply (subst_gamma_dom G2 x v).
      rewrite dom_G2 in x0_Notin_G2.
         eapply binds_tail; auto. 
      eapply IHT; eauto.
   eapply (split_commutes _ _ _ split_fx). (* third premise of T-AFN *)
   assert (effect_ok fx_v G1) as ok_G1. (* fourth premise of T-AFN *)
     eapply typing_effect_ok; auto. apply T_v.
   assert (fx_v = fx_ev) as fx_eq.
     eapply split_nil_eq_2; auto.
     assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2G1.
        eapply subst_gamma_wf. rewrite AxEmpty in wf_GEmpty. rewrite G_eq_G2xG1 in wf_GEmpty. apply wf_GEmpty.
        eapply typing_wf_exp. apply T_v. 
        apply x_notin_A.
     eapply effect_ok_weakening. subst; auto. apply wf_sG2G1.
(* binds_x0_x ... x0 in [(x, t)] *)
inversion binds_x0_x.
destruct (x0 == x) as [x0_eq_x | x0_neq_x].
rewrite x0_eq_x. simpl.
destruct (x==x).
assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2_G1.
   eapply subst_gamma_wf. rewrite AxEmpty in wf_GEmpty.  rewrite G_eq_G2xG1 in wf_GEmpty.  apply wf_GEmpty. 
   eapply typing_wf_exp. apply T_v. 
   apply x_notin_A.
pose proof (weakening G1 (subst_gamma x v G2) {} A A v (tx, fx_v) wf_G1 (split_nil_A_1 A) wf_sG2_G1 T_v) as T_v'.
assert (tx = t) as tx_eq_t.
  injection H0. auto. 
rewrite <- tx_eq_t.
assert (subst_typ x v tx = tx) as stx_tx.
  eapply subst_wf_typ_ext.
  eapply typing_wf_typ. apply T_v.
  eapply binds_disjoint_1. rewrite G_eq_G2xG1 in WF_G1xG2Ax. apply WF_G1xG2Ax.
assert (fx_v = fx_ev) as eq_fx.
  eapply split_nil_eq_2. assumption.
rewrite stx_tx. rewrite <- eq_fx. assumption.
destruct n. auto. 
discriminate.
(* x0 in G2 *)
simpl. destruct (x == x0) as [x_eq_x0 | x_neq_x0]. 
(* Subcase x = x0 (impos) *)
assert (x0 `notin` dom G2) as x0_notin_G2.
  eapply binds_disjoint_2.
  rewrite G_eq_G2xG1 in WF_G1xG2Ax. subst. apply WF_G1xG2Ax.
assert (x0 `in` dom G2) as x0_in_G2. eapply binds_In. apply binds_x0_G2. fsetdec.
(* Subcase x <> x0 *)
pose proof (Gamma_subst_binding_typ G2 x0 x t v binds_x0_G2 x_neq_x0) as binds_subx0_G2.
pose proof (binds_head binding x0 _ G1 _ binds_subx0_G2) as binds_subx0_G1G2.
rewrite A_Empty.
   eapply type_AFN.
   eapply split_nil_A_1. (* first premise of T-AFN: {},{} *)
   eapply type_X. (* second premise : IH *)
     assumption.
     eapply IHT; eauto.
   eapply (split_commutes _ _ _ split_fx). (* third premise of T-AFN *)
   assert (effect_ok fx_v G1) as ok_G1. (* fourth premise of T-AFN *)
     eapply typing_effect_ok; auto. apply T_v.
   assert (fx_v = fx_ev) as fx_eq.
     eapply split_nil_eq_2; auto.
     assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2G1.
        eapply subst_gamma_wf. rewrite AxEmpty in wf_GEmpty.  rewrite G_eq_G2xG1 in wf_GEmpty. apply wf_GEmpty.
        eapply typing_wf_exp. apply T_v.
        apply x_notin_A.
     eapply effect_ok_weakening. subst; auto. apply wf_sG2G1.
Case "typing_XA".
intros G2 G1 wf_G1 Kind_txk fx_ev fx_e fx_v T_v split_fx t0 t_fxe A x_notin_A wf_Gx0 Ax0_eq G_eq_G2xG1 T_x0.
injection t_fxe; intros. subst t0. subst fx_e.
rewrite G_eq_G2xG1 in b.
destruct (binds_concat_inv binding x0 (binding_typ t) _ G2 b) as [ [x0_Notin_G2 binds_x0_xG1] | binds_x0_G2].
destruct (binds_concat_inv binding x0 (binding_typ t) G1 _ binds_x0_xG1) as [ [ x0_Notin_x binds_x0_G1 ] | binds_x0_x].
(* x0 in G1 *)
simpl.
destruct (x == x0) as [x_eq_x0 | x_neq_x0].
subst. destruct x0_Notin_x. simpl. fsetdec.
assert (A = singleton x0) as A_is_x0.
   destruct (aqual_cases A x k) as [aq_is_A | aq_is_Ax].
   rewrite <- Ax0_eq in aq_is_A. auto.
   rewrite <- Ax0_eq in aq_is_Ax. 
   assert (x `in` singleton x0) as x_in_x0.
     rewrite aq_is_Ax. fsetdec.
     assert (x = x0) as contr. fsetdec. 
   contradiction.  
rewrite A_is_x0. 
   eapply type_AFN.
   eapply split_nil_A_2. (* first premise of T-AFN: {},{} *)
   eapply type_XA.
   assert ((subst_typ x v t) = t) as subt_eq_t.
     assert (wf_typ t G1) as wf_t_G1.
      eapply binds_wf. apply wf_G1. apply binds_x0_G1.
   eapply subst_wf_typ_ext. apply wf_t_G1. eapply binds_disjoint_1.
     rewrite G_eq_G2xG1 in WF_G1xG2Ax. apply WF_G1xG2Ax.
   rewrite subt_eq_t.
   assert (dom G2 = dom (subst_gamma x v G2)) as dom_G2.
     eapply (subst_gamma_dom G2 x v).
   rewrite dom_G2 in x0_Notin_G2.
   eapply binds_tail; auto. 
   eapply IHT; eauto using wf_split1, split_nil_A_1, split_nil_A_2.
   eapply (split_commutes _ _ _ split_fx). (* third premise of T-AFN *)
   assert (effect_ok fx_v G1) as ok_G1. (* fourth premise of T-AFN *)
     eapply typing_effect_ok; auto. apply T_v.
   assert (fx_v = fx_ev) as fx_eq.
     eapply split_nil_eq_2; auto.
     assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2G1.
        eapply subst_gamma_wf. rewrite Ax0_eq in wf_Gx0.  rewrite G_eq_G2xG1 in wf_Gx0. apply wf_Gx0.
        eapply typing_wf_exp. apply T_v.
        apply x_notin_A.
     eapply effect_ok_weakening. subst; auto. apply wf_sG2G1.
(* binds_x0_x ... x0 in [(x, t)] *)
inversion binds_x0_x.
destruct (x0 == x) as [x0_eq_x | x0_neq_x].
rewrite x0_eq_x. simpl.
destruct (x==x).
assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2_G1.
        eapply subst_gamma_wf. rewrite Ax0_eq in wf_Gx0.  rewrite G_eq_G2xG1 in wf_Gx0. apply wf_Gx0.
        eapply typing_wf_exp. apply T_v.
        apply x_notin_A.
pose proof (weakening G1 (subst_gamma x v G2) {} A A v (tx, fx_v) wf_G1 (split_nil_A_1 A) wf_sG2_G1 T_v) as T_v'.
assert (tx = t) as tx_eq_t.
  injection H0. auto. 
rewrite <- tx_eq_t.
assert (subst_typ x v tx = tx) as stx_tx.
  eapply subst_wf_typ_ext.
  eapply typing_wf_typ. apply T_v.
  eapply binds_disjoint_1. rewrite G_eq_G2xG1 in WF_G1xG2Ax. apply WF_G1xG2Ax.
assert (fx_v = fx_ev) as eq_fx. 
  eapply split_nil_eq_2. assumption.
rewrite stx_tx. rewrite <- eq_fx. assumption.
destruct n. auto. 
discriminate.
(* x0 in G2 *)
simpl. destruct (x == x0) as [x_eq_x0 | x_neq_x0]. 
(* Subcase x = x0 (impos) *)
assert (x0 `notin` dom G2) as x0_notin_G2.
  eapply binds_disjoint_2.
  rewrite G_eq_G2xG1 in WF_G1xG2Ax. subst. apply WF_G1xG2Ax.
assert (x0 `in` dom G2) as x0_in_G2. eapply binds_In. apply binds_x0_G2. fsetdec.
(* Subcase x <> x0 *)
pose proof (Gamma_subst_binding_typ G2 x0 x t v binds_x0_G2 x_neq_x0) as binds_subx0_G2.
pose proof (binds_head binding x0 _ G1 _ binds_subx0_G2) as binds_subx0_G1G2.
assert (A = singleton x0) as A_is_x0.
   destruct (aqual_cases A x k) as [aq_is_A | aq_is_Ax].
   rewrite <- Ax0_eq in aq_is_A. auto.
   rewrite <- Ax0_eq in aq_is_Ax. 
   assert (x `in` singleton x0) as x_in_x0.
     rewrite aq_is_Ax. fsetdec.
     assert (x = x0) as contr. fsetdec. 
   contradiction.  
rewrite A_is_x0.
   eapply type_AFN.
   eapply split_nil_A_2. (* first premise of T-AFN: {},{} *)
   eapply type_XA. (* second premise : IH *)
     assumption.
     eapply IHT; eauto using wf_split1, split_nil_A_1, split_nil_A_2.
   eapply (split_commutes _ _ _ split_fx). (* third premise of T-AFN *)
   assert (effect_ok fx_v G1) as ok_G1. (* fourth premise of T-AFN *)
     eapply typing_effect_ok; auto. apply T_v.
   assert (fx_v = fx_ev) as fx_eq.
     eapply split_nil_eq_2; auto.
     assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2G1.
        eapply subst_gamma_wf. rewrite Ax0_eq in wf_Gx0.  rewrite G_eq_G2xG1 in wf_Gx0. apply wf_Gx0.
        eapply typing_wf_exp. apply T_v.
        apply x_notin_A.
     eapply effect_ok_weakening. subst; auto. apply wf_sG2G1.
Case "typing_NEW".
intros G2 G1 wf_G1 Kind_txk fx_newe_v fx_newe fx_v T_v split_fx t0 t_fxe A0 x_notin_A0 wf_GA Ax0_eq G_eq T_newe.
injection t_fxe; intros. subst t0. subst a_fx. 
destruct (split_trans_2 _ _ _ _ _ split_fx s) as [fx_ev split_fxv_fxe].
simpl.
eapply type_NEW.
  eapply IHT; eauto using T_v, split_fxv_fxe. (* first premise *)
  eapply IHT0; eauto using G_eq, wf_split1, split_nil_A_1, split_nil_A_2. (* second premise *)
  assert (binds a (binding_kind kind_N) (G2 ++ G1)) as binds_G2G1_a.
    assert (x <> a) as x_neq_a.
      unfold not; intro x_eq_a.
      subst.
      pose proof (binds_mid_eq binding a _ _ G1 G2 b (wf_is_ok _ _ WF_G1xG2Ax)) as kind_N_eq_t.
      discriminate.
    eapply binds_remove_mid. rewrite G_eq in b. apply b. auto.
  assert ((dom G2) = dom (subst_gamma x v G2)) as dom_eq.
    eapply (subst_gamma_dom G2 x v).
  destruct (binds_concat_inv binding a (binding_kind kind_N) G1 G2 binds_G2G1_a) as [ [ a_Notin_G2 binds_a_G1 ] | binds_a_G2].
    (* case binds_a_G1 *)
    eapply binds_tail. assumption. rewrite dom_eq in a_Notin_G2. assumption.
    (* case binds_a_G2 *)
    assert (binds a (binding_kind kind_N) (subst_gamma x v G2)) as binds_a_sG2.
      eapply subst_gamma_kind_invariant; eauto.
    eapply binds_head; auto.
  (* final premise ... disjoint effects *)
  inversion s.
  inversion split_fx.  
  inversion split_fxv_fxe.
  rewrite <- H2.
  assert ((fx_v `union` (singleton a) `union` fx) = 
          ((singleton a) `union` (fx_v `union` fx))) as eq. fsetdec.
  rewrite eq.
  eapply split_A.
  rename H into inter_a_fx_nil.
  assert ((singleton a) `inter` fx_v = {}) as inter_a_fxv_nil.
    rewrite <- H2 in H3.
       destruct (inter_distributes fx_v (singleton a) fx H3) as [_ inter_fxv_a_nil].
       apply inter_symmetric. assumption.
  eapply inter_distributes_inv; auto.
Case "type_ABS".
intros G2 G1 wf_G1 Kind_txk fx_lamv fx_lam fx_v T_v split_fx t0 eq_tfx A_minus_x x_notin_Amx wf_GA A_eq G_eq T_lam.
injection eq_tfx; intros. subst t0. subst fx.
simpl.
assert ((subst_typ x v (ptype (typ_fun t_x t_e) A fx_lam)) = (ptype (subst_typ x v (typ_fun t_x t_e)) A fx_lam)) as subst_ptype_eq.
  eapply subst_ptype; auto.
rewrite subst_ptype_eq. simpl.
destruct (exclude_middle_empty_sets fx_v) as [fx_v_empty | fx_v_nonempty].
  (* SubCase 1: fx_v_empty *)
   subst fx_v.
   assert (fx_lam = fx_lamv) as fx_lam_eq_lamv.
     eapply split_nil_eq_1; eauto using split_fx.
   assert (A = A_minus_x) as A_eq_Amx.
     assert (kinding G1 tx kind_U) as kinding_U.
     eapply inversion_empty_effect; eauto using T_v.
     assert (k = kind_U) as k_is_U.
     eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
     subst k; unfold aqual in A_eq; simpl in A_eq; assumption.
   rewrite A_eq_Amx; rewrite fx_lam_eq_lamv.
   pick fresh y excluding (L `union` (singleton x) `union` dom G) and apply type_ABS.
   rename t_x into t_y. rename k0 into k_y.
    cutrewrite((open_exp 0 y (subst_exp x v e)) = (subst_exp x v (open_exp 0 y e))).
    cutrewrite((open_typ 0 y (subst_typ x v t_e)) = (subst_typ x v (open_typ 0 y t_e))).
    assert (y `notin` L) as y_notin_L. fsetdec.
    cut (wf_env ([(y, binding_typ t_y)] ++ G) (aqual y k_y A)). intro wf_yG.
    cut (aqual y k_y A = aqual x k (aqual y k_y A_minus_x)). intro aq_eq.
    cut (x `notin` aqual y k_y A_minus_x). intro x_notin_aqAmx.
    assert ([(y, binding_typ t_y)] ++ G = ([(y, binding_typ t_y)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as yG_eq.
      rewrite G_eq. simpl_env. reflexivity.
    pose proof (H y y_notin_L wf_yG (([(y, binding_typ t_y)] ++ G2)) G1 wf_G1 Kind_txk fx_lamv fx_lam {} T_v split_fx _ (refl_equal ((open_typ 0 y t_e), fx_lam))
                  (aqual y k_y A_minus_x) x_notin_aqAmx wf_yG aq_eq yG_eq (t y y_notin_L)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : x_notin_aqAmx *)
    unfold not. intro x_in_aqAmx.
    destruct (aqual_identity x y k_y A_minus_x x_in_aqAmx) as [x_eq_y | x_in_Amx].
      subst x. fsetdec.
      contradiction.
    (* cut 2 : aq_eq *)
    subst A. eapply aqual_commutes; auto.
    (* cut 3 : wf_yG *)
    pose proof (kinding_wf_typ G t_y k_y k1) as wf_t_y.
    destruct (aqual_cases A y k_y) as [ aq_eq_A | aq_eq_yA ].
      rewrite aq_eq_A.
      apply wf_env_typ. assumption. fsetdec. assumption.
      rewrite aq_eq_yA.
      apply wf_env_typ_A. assumption. fsetdec. assumption.
    (* cut 4 : (subst_typ x) (open_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_typ_open_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v). assumption.
    (* cut 5 : (subst_exp x) (open_exp y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v); assumption.
(* Second premise of type_ABS *)
    eapply IHT; eauto using wf_split1, split_nil_A_1, split_nil_A_2.
(*** SubCase 2 : fx_v_nonempty ***)
    destruct (inversion_nonempty_effect _ _ _ _ _ Val_v T_v fx_v_nonempty) as [ tx_kind_A | T_v_fxempty ].
   (** SubSubCase 2a. tx_kind_A **)
     SCase "tx_kind_A".
      assert (k = kind_A) as k_is_A.
        eapply kinding_uniqueness; eauto using tx_kind_A, Kind_txk.
      assert (x `in` A) as x_in_A.
        subst k; unfold aqual in A_eq; simpl in A_eq; fsetdec.
      destruct (non_empty_2 _ fx_v_nonempty) as [fx_elt fx_elt_in_v].
      assert (fx_elt `in` fx_lamv) as v_in_lamv.
        inversion split_fx; fsetdec.
      assert (AtomSet.F.is_empty (A `union` fx_lam) = false) as nonempty_L.
        eapply non_empty_3; [eauto || fsetdec].
      assert (AtomSet.F.is_empty (A_minus_x `union` fx_lamv) = false) as nonempty_R.
        eapply non_empty_3; [eauto || fsetdec].
      rewrite <- nonempty_R in nonempty_L.
      assert (ptype (typ_fun (subst_typ x v t_x) (subst_typ x v t_e)) A fx_lam = 
              ptype (typ_fun (subst_typ x v t_x) (subst_typ x v t_e)) A_minus_x fx_lamv) as rw.
        eapply ptype_union_equality; eauto using nonempty_L.
      rewrite rw. (* now starts the reconstruction of the derivation *)
      pick fresh y excluding (L `union` (singleton x) `union` dom G) and apply type_ABS.
      rename t_x into t_y. rename k0 into k_y.
      cutrewrite((open_exp 0 y (subst_exp x v e)) = (subst_exp x v (open_exp 0 y e))).
      cutrewrite((open_typ 0 y (subst_typ x v t_e)) = (subst_typ x v (open_typ 0 y t_e))).
      assert (y `notin` L) as y_notin_L. fsetdec.
      cut (wf_env ([(y, binding_typ t_y)] ++ G) (aqual y k_y A)). intro wf_yG.
      cut (aqual y k_y A = aqual x k (aqual y k_y A_minus_x)). intro aq_eq.
      cut (x `notin` aqual y k_y A_minus_x). intro x_notin_aqAmx.
      assert ([(y, binding_typ t_y)] ++ G = ([(y, binding_typ t_y)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as yG_eq.
      rewrite G_eq. simpl_env. reflexivity.
      pose proof (H y y_notin_L wf_yG (([(y, binding_typ t_y)] ++ G2)) G1 wf_G1 Kind_txk fx_lamv fx_lam fx_v T_v split_fx _ (refl_equal ((open_typ 0 y t_e), fx_lam))
        (aqual y k_y A_minus_x) x_notin_aqAmx wf_yG aq_eq yG_eq (t y y_notin_L)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
      (* cut 1 : x_notin_aqAmx *)
      unfold not. intro x_in_aqAmx.
      destruct (aqual_identity x y k_y A_minus_x x_in_aqAmx) as [x_eq_y | x_in_Amx].
      subst x. fsetdec.
      contradiction.
      (* cut 2 : aq_eq *)
      subst A. eapply aqual_commutes; auto.
               (* cut 3 : wf_yG *)
      pose proof (kinding_wf_typ G t_y k_y k1) as wf_t_y.
      destruct (aqual_cases A y k_y) as [ aq_eq_A | aq_eq_yA ].
      rewrite aq_eq_A.
      apply wf_env_typ. assumption. fsetdec. assumption.
      rewrite aq_eq_yA.
      apply wf_env_typ_A. assumption. fsetdec. assumption.
                                               (* cut 4 : (subst_typ x) (open_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_typ_open_typ_commute.
      fsetdec.
      destruct (typing_wf_exp G1 {} v tx fx_v T_v). assumption.
                                                  (* cut 5 : (subst_exp x) (open_exp y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_commute.
      fsetdec.
      destruct (typing_wf_exp G1 {} v tx fx_v T_v); assumption.
      (* Second premise of type_ABS *)
      eapply IHT; eauto using wf_split1, split_nil_A_1, split_nil_A_2.
      (*** Subcase 2b ***)
      SCase "T_v_fxempty".
      assert (typing (subst_gamma x v G2 ++ G1) A_minus_x
                     (exp_lam (subst_typ x v t_x) (subst_exp x v e))
                        (ptype (typ_fun (subst_typ x v t_x) (subst_typ x v t_e)) A fx_lam,
                         fx_lam)) as premise_TAFN.
     assert (A = A_minus_x) as A_eq_Amx.
       assert (kinding G1 tx kind_U) as kinding_U.
       eapply inversion_empty_effect.
          apply Val_v.
          apply T_v_fxempty.
       assert (k = kind_U) as k_is_U.
          eapply kinding_uniqueness.
            eapply Kind_txk.
            eapply kinding_U.
       subst k; unfold aqual in A_eq; simpl in A_eq; assumption.
     rewrite A_eq_Amx.
     pick fresh y excluding (L `union` (singleton x) `union` dom G) and apply type_ABS.
     rename t_x into t_y. rename k0 into k_y.
     cutrewrite((open_exp 0 y (subst_exp x v e)) = (subst_exp x v (open_exp 0 y e))).
    cutrewrite((open_typ 0 y (subst_typ x v t_e)) = (subst_typ x v (open_typ 0 y t_e))).
    assert (y `notin` L) as y_notin_L. fsetdec.
    cut (wf_env ([(y, binding_typ t_y)] ++ G) (aqual y k_y A)). intro wf_yG.
    cut (aqual y k_y A = aqual x k (aqual y k_y A_minus_x)). intro aq_eq.
    cut (x `notin` aqual y k_y A_minus_x). intro x_notin_aqAmx.
    assert ([(y, binding_typ t_y)] ++ G = ([(y, binding_typ t_y)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as yG_eq.
      rewrite G_eq. simpl_env. reflexivity.
      pose proof (H y y_notin_L wf_yG (([(y, binding_typ t_y)] ++ G2)) G1 wf_G1 Kind_txk fx_lam fx_lam {} T_v_fxempty 
                  (split_nil_A_1 fx_lam) _ (refl_equal ((open_typ 0 y t_e), fx_lam))
                  (aqual y k_y A_minus_x) x_notin_aqAmx wf_yG aq_eq yG_eq (t y y_notin_L)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : x_notin_aqAmx *)
    unfold not. intro x_in_aqAmx.
    destruct (aqual_identity x y k_y A_minus_x x_in_aqAmx) as [x_eq_y | x_in_Amx].
      subst x. fsetdec.
      contradiction.
    (* cut 2 : aq_eq *)
    subst A. eapply aqual_commutes; auto.
    (* cut 3 : wf_yG *)
    pose proof (kinding_wf_typ G t_y k_y k1) as wf_t_y.
    destruct (aqual_cases A y k_y) as [ aq_eq_A | aq_eq_yA ].
      rewrite aq_eq_A.
      apply wf_env_typ. assumption. fsetdec. assumption.
      rewrite aq_eq_yA.
      apply wf_env_typ_A. assumption. fsetdec. assumption.
    (* cut 4 : (subst_typ x) (open_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_typ_open_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v_fxempty). assumption.
    (* cut 5 : (subst_exp x) (open_exp y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v_fxempty); assumption.
(* Second premise of type_ABS *)
   eapply IHT.
      apply G_eq.
      eapply wf_split1.
        apply wf_GA.
        apply split_nil_A_1.
      apply wf_G1.
      apply Kind_txk.
      apply T_v.
      apply split_fx.
    (* Finally ... conclude with T-AFN *)
   eapply type_AFN.
      apply (split_nil_A_2 A_minus_x).
      apply premise_TAFN.
      eapply (split_commutes _ _ _ split_fx).
      subst G.
      eapply effect_ok_subst.
        apply wf_GA. 
        eapply typing_effect_ok. 
          apply T_lam.
        eapply typing_effect_ok.
          apply T_v. apply (split_commutes _ _ _ split_fx).
Case "typing_TAB_NOFX".
intros G2 G1 wf_G1 Kind_txk fx_tabv fx_tab fx_v T_v split_fx t0 eq_tfx A_minus_x x_notin_Amx wf_GA A_eq G_eq T_tab.
injection eq_tfx; intros. subst t0. subst fx.
simpl.
assert ((subst_typ x v (ptype (typ_univ k0 NO_FX t_e) A fx_tab)) = (ptype (subst_typ x v (typ_univ k0 NO_FX t_e)) A fx_tab)) as subst_ptype_eq.
  eapply subst_ptype; auto.
rewrite subst_ptype_eq. simpl.
destruct (exclude_middle_empty_sets fx_v) as [fx_v_empty | fx_v_nonempty].
  (* SubCase 1: fx_v_empty *)
   subst fx_v.
   assert (fx_tab = fx_tabv) as fx_tab_eq_tabv.
     eapply split_nil_eq_1; eauto using split_fx.
   assert (A = A_minus_x) as A_eq_Amx.
     assert (kinding G1 tx kind_U) as kinding_U.
     eapply inversion_empty_effect; eauto using T_v.
     assert (k = kind_U) as k_is_U.
     eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
     subst k; unfold aqual in A_eq; simpl in A_eq; assumption.
   rewrite A_eq_Amx; rewrite fx_tab_eq_tabv.
   pick fresh b excluding (L `union` (singleton x) `union` dom G) and apply type_TAB.
   intro b_notin_fxtabv.
   rename k0 into k_b.
    cutrewrite((open_exp_typ 0 b (subst_exp x v e)) = (subst_exp x v (open_exp_typ 0 b e))).
    cutrewrite((open_typ_typ 0 b (subst_typ x v t_e)) = (subst_typ x v (open_typ_typ 0 b t_e))).
    assert (b `notin` L) as b_notin_L. fsetdec.
    cut (wf_env ([(b, binding_kind k_b)] ++ G) A). intro wf_bG.
    cut (b `notin` fx_tab). intro b_notin_fxtab.  
(*     cut (aqual y k_y A = aqual x k (aqual y k_y A_minus_x)). intro aq_eq. *)
(*     cut (x `notin` aqual y k_y A_minus_x). intro x_notin_aqAmx. *)
    assert ([(b, binding_kind k_b)] ++ G = ([(b, binding_kind k_b)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
      rewrite G_eq. simpl_env. reflexivity.
    pose proof (H b b_notin_L b_notin_fxtab wf_bG (([(b, binding_kind k_b)] ++ G2)) G1 wf_G1 Kind_txk fx_tabv fx_tab {} T_v split_fx _ 
                (refl_equal ((open_typ_typ 0 b t_e), fx_tab))
                  A_minus_x x_notin_Amx wf_bG A_eq bG_eq (t b b_notin_L b_notin_fxtab)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : b notin fx_tab *)
       subst fx_tabv; auto.
    (* cut 2 : wf_bG *)
    eapply wf_env_kind; ( auto || fsetdec ).
    (* cut 3 : (subst_typ x) (open_typ_typ y) commute (when x <> y) *) 
     apply sym_equal.
      eapply subst_typ_open_typ_typ_commute.
         destruct (typing_wf_exp G1 {} v tx {} T_v). assumption.
    (* cut 4 : (subst_exp x) (open_exp_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_typ_commute.
         destruct (typing_wf_exp G1 {} v tx {} T_v); assumption.
(* Second premise of type_ABS *)
     assumption.
(*** SubCase 2 : fx_v_nonempty ***)
    destruct (inversion_nonempty_effect _ _ _ _ _ Val_v T_v fx_v_nonempty) as [ tx_kind_A | T_v_fxempty ].
   (** SubSubCase 2a. tx_kind_A **)
     SCase "tx_kind_A".
      assert (k = kind_A) as k_is_A.
        eapply kinding_uniqueness; eauto using tx_kind_A, Kind_txk.
      assert (x `in` A) as x_in_A.
        subst k; unfold aqual in A_eq; simpl in A_eq; fsetdec.
      destruct (non_empty_2 _ fx_v_nonempty) as [fx_elt fx_elt_in_v].
      assert (fx_elt `in` fx_tabv) as v_in_tabv.
        inversion split_fx; fsetdec.
      assert (AtomSet.F.is_empty (A `union` fx_tab) = false) as nonempty_L.
        eapply non_empty_3; [eauto || fsetdec].
      assert (AtomSet.F.is_empty (A_minus_x `union` fx_tabv) = false) as nonempty_R.
        eapply non_empty_3; [eauto || fsetdec].
      rewrite <- nonempty_R in nonempty_L.
      assert ((ptype (typ_univ k0 NO_FX (subst_typ x v t_e)) A fx_tab) =
              (ptype (typ_univ k0 NO_FX (subst_typ x v t_e)) A_minus_x fx_tabv)) as rw.
        eapply ptype_union_equality; eauto using nonempty_L.
      rewrite rw. (* now starts the reconstruction of the derivation *)
   pick fresh b excluding (L `union` (singleton x) `union` dom G) and apply type_TAB.
   intro b_notin_fxtabv.
   rename k0 into k_b.
    cutrewrite((open_exp_typ 0 b (subst_exp x v e)) = (subst_exp x v (open_exp_typ 0 b e))).
    cutrewrite((open_typ_typ 0 b (subst_typ x v t_e)) = (subst_typ x v (open_typ_typ 0 b t_e))).
    assert (b `notin` L) as b_notin_L. fsetdec.
    cut (wf_env ([(b, binding_kind k_b)] ++ G) A). intro wf_bG.
    cut (b `notin` fx_tab). intro b_notin_fxtab.  
(*     cut (aqual y k_y A = aqual x k (aqual y k_y A_minus_x)). intro aq_eq. *)
(*     cut (x `notin` aqual y k_y A_minus_x). intro x_notin_aqAmx. *)
    assert ([(b, binding_kind k_b)] ++ G = ([(b, binding_kind k_b)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
      rewrite G_eq. simpl_env. reflexivity.
    pose proof (H b b_notin_L b_notin_fxtab wf_bG (([(b, binding_kind k_b)] ++ G2)) G1 wf_G1 Kind_txk fx_tabv fx_tab fx_v T_v split_fx _ 
                (refl_equal ((open_typ_typ 0 b t_e), fx_tab))
                  A_minus_x x_notin_Amx wf_bG A_eq bG_eq (t b b_notin_L b_notin_fxtab)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : b notin fx_tab *)
       inversion split_fx. subst fx_tabv; fsetdec.
    (* cut 2 : wf_bG *)
    eapply wf_env_kind; ( auto || fsetdec ).
    (* cut 3 : (subst_typ x) (open_typ_typ y) commute (when x <> y) *) 
     apply sym_equal.
      eapply subst_typ_open_typ_typ_commute.
         destruct (typing_wf_exp G1 {} v tx fx_v T_v). assumption.
    (* cut 4 : (subst_exp x) (open_exp_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_typ_commute.
         destruct (typing_wf_exp G1 {} v tx fx_v T_v); assumption.
(* Second premise of type_ABS *)
     assumption.
      (*** Subcase 2b ***)
      SCase "T_v_fxempty".
      assert (typing (subst_gamma x v G2 ++ G1) A_minus_x (exp_tab k0 (subst_exp x v e))
                     (ptype (typ_univ k0 NO_FX (subst_typ x v t_e)) A fx_tab, fx_tab)) as premise_TAFN.
      assert (A = A_minus_x) as A_eq_Amx.
         assert (kinding G1 tx kind_U) as kinding_U.
            eapply inversion_empty_effect.
              apply Val_v.
              apply T_v_fxempty.
         assert (k = kind_U) as k_is_U.
            eapply kinding_uniqueness.
               apply Kind_txk.
               apply kinding_U.
         subst k; unfold aqual in A_eq; simpl in A_eq; assumption.
      rewrite A_eq_Amx.
      pick fresh b excluding (L `union` (singleton x) `union` dom G) and apply type_TAB.
      intro b_notin_fxtabv.
      rename k0 into k_b.
      cutrewrite((open_exp_typ 0 b (subst_exp x v e)) = (subst_exp x v (open_exp_typ 0 b e))).
      cutrewrite((open_typ_typ 0 b (subst_typ x v t_e)) = (subst_typ x v (open_typ_typ 0 b t_e))).
      assert (b `notin` L) as b_notin_L. fsetdec.
      cut (wf_env ([(b, binding_kind k_b)] ++ G) A). intro wf_bG.
      cut (b `notin` fx_tab). intro b_notin_fxtab.  
      assert ([(b, binding_kind k_b)] ++ G = ([(b, binding_kind k_b)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
         rewrite G_eq. simpl_env. reflexivity.
      pose proof (H b b_notin_L b_notin_fxtab wf_bG (([(b, binding_kind k_b)] ++ G2)) G1 wf_G1 Kind_txk fx_tab fx_tab {} T_v_fxempty 
                 (split_nil_A_1 fx_tab) _ (refl_equal ((open_typ_typ 0 b t_e), fx_tab))
                    A_minus_x x_notin_Amx wf_bG A_eq bG_eq (t b b_notin_L b_notin_fxtab)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : b notin fx_tab *)
       auto.
    (* cut 2 : wf_bG *)
    eapply wf_env_kind; ( auto || fsetdec ).
    (* cut 3 : (subst_typ x) (open_typ_typ y) commute (when x <> y) *) 
     apply sym_equal.
      eapply subst_typ_open_typ_typ_commute.
         destruct (typing_wf_exp G1 {} v tx {} T_v_fxempty). assumption.
    (* cut 4 : (subst_exp x) (open_exp_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_typ_commute.
         destruct (typing_wf_exp G1 {} v tx {} T_v_fxempty); assumption.
(* Second premise of type_ABS *)
     assumption.
   (* Finally ... conclude with T-AFN *)
   eapply type_AFN.
      apply (split_nil_A_2 A_minus_x).
      apply premise_TAFN.
      eapply (split_commutes _ _ _ split_fx).
      subst G.
      eapply effect_ok_subst.
        apply wf_GA. 
        eapply typing_effect_ok. 
          apply T_tab.
        eapply typing_effect_ok.
          apply T_v. apply (split_commutes _ _ _ split_fx).
Case "typing_TAB_FX".
intros G2 G1 wf_G1 Kind_txk fx_tabv fx_tab fx_v T_v split_fx t0 eq_tfx A_minus_x x_notin_Amx wf_GA A_eq G_eq T_tab.
injection eq_tfx; intros. subst t0. subst fx.
simpl.
assert ((subst_typ x v (ptype (typ_univ k0 FX t_e) A fx_tab)) = (ptype (subst_typ x v (typ_univ k0 FX t_e)) A fx_tab)) as subst_ptype_eq.
  eapply subst_ptype; auto.
rewrite subst_ptype_eq. simpl.
destruct (exclude_middle_empty_sets fx_v) as [fx_v_empty | fx_v_nonempty].
  (* SubCase 1: fx_v_empty *)
   subst fx_v.
   assert (fx_tab = fx_tabv) as fx_tab_eq_tabv.
     eapply split_nil_eq_1; eauto using split_fx.
   assert (A = A_minus_x) as A_eq_Amx.
     assert (kinding G1 tx kind_U) as kinding_U.
     eapply inversion_empty_effect; eauto using T_v.
     assert (k = kind_U) as k_is_U.
     eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
     subst k; unfold aqual in A_eq; simpl in A_eq; assumption.
   rewrite A_eq_Amx; rewrite fx_tab_eq_tabv.
   subst fx_tabv.
   pick fresh b excluding (L `union` (singleton x) `union` dom G `union` fx_tab) and apply type_TAB_FX.
    (* first premise: fx subset L *)
     fsetdec.
    (* second premise : e typing *)
    assert (b `notin` L) as b_notin_L. fsetdec.
    pose proof (t b b_notin_L) as T_e.
    rename k0 into k_b.
    cutrewrite((open_exp_typ 0 b (subst_exp x v e)) = (subst_exp x v (open_exp_typ 0 b e))).
    cutrewrite((open_typ_typ 0 b (subst_typ x v t_e)) = (subst_typ x v (open_typ_typ 0 b t_e))).
    cut (wf_env ([(b, binding_kind k_b)] ++ G) A). intro wf_bG.
    cut (b `notin` fx_tab). intro b_notin_fxtab.  
    assert ([(b, binding_kind k_b)] ++ G = ([(b, binding_kind k_b)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
      rewrite G_eq. simpl_env. reflexivity.
    pose proof (H b b_notin_L wf_bG (([(b, binding_kind k_b)] ++ G2)) G1 wf_G1 Kind_txk _ _ {} T_v 
                (split_nil_A_1 (fx_tab `union` singleton b))  _ (refl_equal ((open_typ_typ 0 b t_e), (fx_tab `union` singleton b)))
                  A_minus_x x_notin_Amx wf_bG A_eq bG_eq T_e) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : b notin fx_tab *)
       fsetdec.
    (* cut 2 : wf_bG *)
    eapply wf_env_kind; ( auto || fsetdec ).
    (* cut 3 : (subst_typ x) (open_typ_typ y) commute (when x <> y) *) 
     apply sym_equal.
      eapply subst_typ_open_typ_typ_commute.
         destruct (typing_wf_exp G1 {} v tx {} T_v). assumption.
    (* cut 4 : (subst_exp x) (open_exp_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_typ_commute.
         destruct (typing_wf_exp G1 {} v tx {} T_v); assumption.
(* Second premise of type_TAB_FX *)
     assumption.
(*** SubCase 2 : fx_v_nonempty ***)
    destruct (inversion_nonempty_effect _ _ _ _ _ Val_v T_v fx_v_nonempty) as [ tx_kind_A | T_v_fxempty ].
   (** SubSubCase 2a. tx_kind_A **)
     SCase "tx_kind_A".
      assert (k = kind_A) as k_is_A.
        eapply kinding_uniqueness; eauto using tx_kind_A, Kind_txk.
      assert (x `in` A) as x_in_A.
        subst k; unfold aqual in A_eq; simpl in A_eq; fsetdec.
      destruct (non_empty_2 _ fx_v_nonempty) as [fx_elt fx_elt_in_v].
      assert (fx_elt `in` fx_tabv) as v_in_tabv.
        inversion split_fx; fsetdec.
      assert (AtomSet.F.is_empty (A `union` fx_tab) = false) as nonempty_L.
        eapply non_empty_3; [eauto || fsetdec].
      assert (AtomSet.F.is_empty (A_minus_x `union` fx_tabv) = false) as nonempty_R.
        eapply non_empty_3; [eauto || fsetdec].
      rewrite <- nonempty_R in nonempty_L.
      assert ((ptype (typ_univ k0 FX (subst_typ x v t_e)) A fx_tab) =
              (ptype (typ_univ k0 FX (subst_typ x v t_e)) A_minus_x fx_tabv)) as rw.
        eapply ptype_union_equality; eauto using nonempty_L.
      rewrite rw. (* now starts the reconstruction of the derivation *)
   pick fresh b excluding (L `union` (singleton x) `union` dom G `union` fx_tabv) and apply type_TAB_FX.
    (* first premise : subset fx_tabv *)
       fsetdec.
    (* second premise : e typing *)
    rename k0 into k_b.
    cut (splitA fx_v (fx_tab `union` singleton b) (fx_tabv `union` singleton b)). intro splitter.
    cutrewrite((open_exp_typ 0 b (subst_exp x v e)) = (subst_exp x v (open_exp_typ 0 b e))).
    cutrewrite((open_typ_typ 0 b (subst_typ x v t_e)) = (subst_typ x v (open_typ_typ 0 b t_e))).
    assert (b `notin` L) as b_notin_L. fsetdec.
    cut (wf_env ([(b, binding_kind k_b)] ++ G) A). intro wf_bG.
    cut (b `notin` fx_tab). intro b_notin_fxtab.  
(*     cut (aqual y k_y A = aqual x k (aqual y k_y A_minus_x)). intro aq_eq. *)
(*     cut (x `notin` aqual y k_y A_minus_x). intro x_notin_aqAmx. *)
    assert ([(b, binding_kind k_b)] ++ G = ([(b, binding_kind k_b)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
      rewrite G_eq. simpl_env. reflexivity.
    pose proof (H b b_notin_L wf_bG (([(b, binding_kind k_b)] ++ G2)) G1 wf_G1 Kind_txk _ (fx_tab `union` singleton b) _ T_v
               splitter _ (refl_equal ((open_typ_typ 0 b t_e), (fx_tab `union` singleton b)))
                  A_minus_x x_notin_Amx wf_bG A_eq bG_eq (t b b_notin_L)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : b notin fx_tab *)
       inversion split_fx. subst fx_tabv; fsetdec.
    (* cut 2 : wf_bG *)
    eapply wf_env_kind; ( auto || fsetdec ).
    (* cut 3 : (subst_typ x) (open_typ_typ y) commute (when x <> y) *) 
     apply sym_equal.
      eapply subst_typ_open_typ_typ_commute.
         destruct (typing_wf_exp G1 {} v tx fx_v T_v). assumption.
    (* cut 4 : (subst_exp x) (open_exp_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_typ_commute.
         destruct (typing_wf_exp G1 {} v tx fx_v T_v); assumption.
    (* cut 5 : splitter *)
      eapply split_extension_R; (eauto || fsetdec).
(* Second premise of type_TAB_FX *)
     assumption.
      (*** Subcase 2b ***)
      SCase "T_v_fxempty".
      assert (typing (subst_gamma x v G2 ++ G1) A_minus_x (exp_tab k0 (subst_exp x v e))
                     (ptype (typ_univ k0 FX (subst_typ x v t_e)) A fx_tab, fx_tab)) as premise_TAFN.
      assert (A = A_minus_x) as A_eq_Amx.
         assert (kinding G1 tx kind_U) as kinding_U.
            eapply inversion_empty_effect; eauto using T_v.
         assert (k = kind_U) as k_is_U.
            eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
            subst k; unfold aqual in A_eq; simpl in A_eq; assumption.
      rewrite A_eq_Amx.
   pick fresh b excluding (L `union` (singleton x) `union` dom G `union` fx_tab) and apply type_TAB_FX.
    (* first premise: fx subset L *)
     fsetdec.
    (* second premise : e typing *)
    assert (b `notin` L) as b_notin_L. fsetdec.
    pose proof (t b b_notin_L) as T_e.
    rename k0 into k_b.
    cutrewrite((open_exp_typ 0 b (subst_exp x v e)) = (subst_exp x v (open_exp_typ 0 b e))).
    cutrewrite((open_typ_typ 0 b (subst_typ x v t_e)) = (subst_typ x v (open_typ_typ 0 b t_e))).
    cut (wf_env ([(b, binding_kind k_b)] ++ G) A). intro wf_bG.
    cut (b `notin` fx_tab). intro b_notin_fxtab.  
    assert ([(b, binding_kind k_b)] ++ G = ([(b, binding_kind k_b)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
      rewrite G_eq. simpl_env. reflexivity.
    pose proof (H b b_notin_L wf_bG (([(b, binding_kind k_b)] ++ G2)) G1 wf_G1 Kind_txk _ _ {} T_v_fxempty
                (split_nil_A_1 (fx_tab `union` singleton b))  _ (refl_equal ((open_typ_typ 0 b t_e), (fx_tab `union` singleton b)))
                  A_minus_x x_notin_Amx wf_bG A_eq bG_eq T_e) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : b notin fx_tab *)
       fsetdec.
    (* cut 2 : wf_bG *)
    eapply wf_env_kind; ( auto || fsetdec ).
    (* cut 3 : (subst_typ x) (open_typ_typ y) commute (when x <> y) *) 
     apply sym_equal.
      eapply subst_typ_open_typ_typ_commute.
         destruct (typing_wf_exp G1 {} v tx {} T_v_fxempty). assumption.
    (* cut 4 : (subst_exp x) (open_exp_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_typ_commute.
         destruct (typing_wf_exp G1 {} v tx {} T_v_fxempty); assumption.
(* Second premise of type_TAB_FX *)
     assumption.
   (* Finally ... conclude with T-AFN *)
   eapply type_AFN.
      apply (split_nil_A_2 A_minus_x).
      apply premise_TAFN.
      eapply (split_commutes _ _ _ split_fx).
      subst G.
      eapply effect_ok_subst.
        apply wf_GA. 
        eapply typing_effect_ok. 
          apply T_tab.
        eapply typing_effect_ok.
          apply T_v. apply (split_commutes _ _ _ split_fx).
Case "type_APP".
intros G2 G1 wf_G1 Kind_txk fx_appv fx_app fx_v T_v split_fx t0 eq_tfx A_minus_x x_notin_Amx wf_GA A_eq G_eq T_app.
injection eq_tfx; intros. subst t0. subst fx.
simpl.
assert (subst_typ x v (open_typ 0 e2 t) = open_typ 0 (subst_exp x v e2) (subst_typ x v t)) as subst_open_weak_eq.
  eapply subst_typ_open_typ_commute_weak; eauto using typing_lc, wf_G1, T_v.
rewrite subst_open_weak_eq.
destruct (exclude_middle_empty_sets fx_v) as [fx_v_empty | fx_v_nonempty].
  SCase "fx_v_empty".
    subst fx_v.
    assert (k = kind_U) as k_is_U.
      assert (kinding G1 tx kind_U) as kinding_U.
        eapply inversion_empty_effect; eauto using T_v.
      eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
    subst k.
    assert (fx_app = fx_appv) as fx_app_eq_appv.
     eapply split_nil_eq_1; eauto using split_fx.
    assert (A = A_minus_x) as A_eq_Amx.
     unfold aqual in A_eq; simpl in A_eq; assumption.
    rewrite <- fx_app_eq_appv.
    subst A.
    unfold aqual in s; simpl s.
    assert (x `notin` A1) as x_notin_A1.
      inversion s; fsetdec.
    assert (x `notin` A2) as x_notin_A2.
      inversion s; fsetdec.
    assert (wf_env G A1) as wf_GA1; eauto using wf_split1.
    assert (wf_env G A2) as wf_GA2; eauto using wf_split2.
    assert (forall A, A = aqual x kind_U A) as aq_U_id.
      unfold aqual; simpl. auto.
    pose proof (IHT1 wf_GA1 G2 G1 wf_G1 Kind_txk fx1 fx1 {} T_v 
                    (split_nil_A_1 fx1) _ (refl_equal (t_e1, fx1))
                    A1 x_notin_A1 wf_GA1 (aq_U_id A1) G_eq T1) as T_sub_e1.
    pose proof (IHT2 wf_GA2 G2 G1 wf_G1 Kind_txk fx2 fx2 {} T_v 
                    (split_nil_A_1 fx2) _ (refl_equal (t2, fx2))
                    A2 x_notin_A2 wf_GA2 (aq_U_id A2) G_eq T2) as T_sub_e2.
    clear IHT1 IHT2.
    eapply type_APP.
       apply s.
       destruct o as [ U | A ].
         left.
          assert (subst_typ x v t_e1 = typ_fun (subst_typ x v t2) (subst_typ x v t)) as EQ.
            subst t_e1; simpl. auto.
          apply EQ.
         right.
          assert (subst_typ x v t_e1 = typ_qtype (typ_fun (subst_typ x v t2) (subst_typ x v t))) as EQ.
            subst t_e1; simpl; auto.
          apply EQ.
       apply T_sub_e1. apply T_sub_e2. assumption.
  SCase "fx_v_nonempty".    
    destruct (inversion_nonempty_effect _ _ _ _ _ Val_v T_v fx_v_nonempty) as [ tx_kind_A | T_v_fxempty ].
   SSCase "tx_kind_A".
       assert (k = kind_A) as k_is_A.
        eapply kinding_uniqueness; eauto using tx_kind_A, Kind_txk.
       subst k.
       assert (x `in` A) as x_in_A.
        unfold aqual in A_eq; simpl in A_eq; fsetdec.
       assert ((x `in` A1 -> x `notin` A2) /\ (x `in` A2 -> x `notin` A1)) as x_splits.
         inversion s.
         split. 
           intro x_in_A1. unfold not. intro x_in_A2. 
           pose proof (AtomSet.F.inter_3 x_in_A1 x_in_A2) as for_contra. rewrite H in for_contra; fsetdec.
           intro x_in_A2. unfold not. intro x_in_A1. 
           pose proof (AtomSet.F.inter_3 x_in_A1 x_in_A2) as for_contra. rewrite H in for_contra; fsetdec.
       assert (x `in` A1 \/ x `in` A2) as x_somewhere.
         inversion s. rewrite <- H2 in x_in_A. auto using AtomSet.F.union_1.
       assert (forall A, aqual x kind_A A = singleton x `union` A) as aqA_identity; (unfold aqual; (simpl; auto)).
       destruct x_somewhere as [x_in_A1 | x_in_A2].
        SSSCase "x_in_A1, x_notin_A2".  
          assert (x `notin` A2) as x_notin_A2; eauto using x_splits.
          clear IHT2.
          assert (wf_env G A1) as wf_GA1; eauto using wf_split1.
          destruct (split_trans_3 _ _ _ _ _ s0 split_fx) as [fx_1_v split_fx_1_v]. 
          destruct (split_singleton x A1 x_in_A1) as [A1_minus_x [ x_notin_A1mx A1_eq]].
          rewrite <- (aqA_identity A1_minus_x) in A1_eq.
          pose proof (IHT1 wf_GA1 G2 G1 wf_G1 Kind_txk fx_1_v fx1 fx_v T_v
                          split_fx_1_v _ (refl_equal (t_e1, fx1))
                          A1_minus_x x_notin_A1mx wf_GA1 A1_eq G_eq T1) as T_sub_e1.
          destruct (split_trans_4 _ _ _ _ _ s0 split_fx) as [fx_2_v split_fx_2_v].
          assert (wf_env G A2) as wf_GA2; eauto using wf_split2. 
            rewrite G_eq in wf_GA2, T2.
          pose proof (nofree_subst_preserves_typing 
                        G1 G2 A2 x tx e2 t2 v fx2 fx_v fx_2_v _ wf_sig
                        wf_G1 Kind_txk x_notin_A2 wf_GA2 T2 Val_v T_v split_fx_2_v) as T_sub_e2.
          clear IHT1.
          cut (splitA A1_minus_x A2 A_minus_x). intro split_a1mx_a2.
          eapply type_APP.
            apply split_a1mx_a2.
            destruct o as [ K_U | K_A ].
              left.
                assert (subst_typ x v t_e1 = typ_fun (subst_typ x v t2) (subst_typ x v t)) as EQ.
                  subst t_e1; simpl. auto.
               apply EQ.
              right.
                assert (subst_typ x v t_e1 = typ_qtype (typ_fun (subst_typ x v t2) (subst_typ x v t))) as EQ.
                  subst t_e1; simpl; auto.
                apply EQ.
            apply T_sub_e1.
            apply T_sub_e2.
            eapply split_trans_5; eauto.
          (* cut : splitA A1_minus_x A2 A_minus_x *)
          rewrite (aqA_identity A_minus_x) in A_eq.
          rewrite (aqA_identity A1_minus_x) in A1_eq.
          eapply split_trans_7; eauto.
        SSSCase "x_in_A2, x_notin_A1".  
          assert (x `notin` A1) as x_notin_A1; eauto using x_splits.
          clear IHT1.
          assert (wf_env G A2) as wf_GA2; eauto using wf_split2.
          destruct (split_trans_4 _ _ _ _ _ s0 split_fx) as [fx_2_v split_fx_2_v].
          destruct (split_singleton x A2 x_in_A2) as [A2_minus_x [ x_notin_A2mx A2_eq]].
          rewrite <- (aqA_identity A2_minus_x) in A2_eq.
          pose proof (IHT2 wf_GA2 G2 G1 wf_G1 Kind_txk fx_2_v fx2 fx_v T_v
                          split_fx_2_v _ (refl_equal (t2, fx2))
                          A2_minus_x x_notin_A2mx wf_GA2 A2_eq G_eq T2) as T_sub_e2.
          destruct (split_trans_3 _ _ _ _ _ s0 split_fx) as [fx_1_v split_fx_1_v]. 
          assert (wf_env G A1) as wf_GA1; eauto using wf_split1. 
            rewrite G_eq in wf_GA1, T1.
          pose proof (nofree_subst_preserves_typing 
                        G1 G2 A1 x tx e1 _ v fx1 fx_v fx_1_v _ wf_sig
                        wf_G1 Kind_txk x_notin_A1 wf_GA1 T1 Val_v T_v split_fx_1_v) as T_sub_e1.
          simpl in T_sub_e1.
          clear IHT2.
          cut (splitA A1 A2_minus_x A_minus_x). intro split_a1_a2mx.
          eapply type_APP.
            apply split_a1_a2mx.
            destruct o as [ K_U | K_A ].
              left.
                assert (subst_typ x v t_e1 = typ_fun (subst_typ x v t2) (subst_typ x v t)) as EQ.
                  subst t_e1; simpl. auto.
                apply EQ.
              right.
                assert (subst_typ x v t_e1 = typ_qtype (typ_fun (subst_typ x v t2) (subst_typ x v t))) as EQ.
                  subst t_e1; simpl; auto.
                apply EQ.
            apply T_sub_e1.
            apply T_sub_e2.
            eapply split_trans_6; eauto.
          (* cut : splitA A1_minus_x A2 A_minus_x *)
          rewrite (aqA_identity A_minus_x) in A_eq.
          rewrite (aqA_identity A2_minus_x) in A2_eq.
          eapply split_commutes.
          eapply split_trans_7; eauto using split_commutes.
    SSCase "T_v_fxempty".
      assert (typing (subst_gamma x v G2 ++ G1) A_minus_x
                     (exp_app (subst_exp x v e1) (subst_exp x v e2))
                     (open_typ 0 (subst_exp x v e2) (subst_typ x v t), fx_app)) as premise_TAFN.
      assert (k = kind_U) as k_is_U.
        assert (kinding G1 tx kind_U) as kinding_U.
          eapply inversion_empty_effect; eauto using T_v.
        eapply kinding_uniqueness; eauto using kinding_U, Kind_txk. 
      subst k.
      assert (A = A_minus_x) as A_eq_Amx.
        unfold aqual in A_eq; simpl in A_eq; assumption.
      subst A.
      unfold aqual in s; simpl s.
      assert (x `notin` A1) as x_notin_A1.
        inversion s; fsetdec.
      assert (x `notin` A2) as x_notin_A2.
        inversion s; fsetdec.
      assert (wf_env G A1) as wf_GA1; eauto using wf_split1.
      assert (wf_env G A2) as wf_GA2; eauto using wf_split2.
      assert (forall A, A = aqual x kind_U A) as aq_U_id.
        unfold aqual; simpl. auto.
      pose proof (IHT1 wf_GA1 G2 G1 wf_G1 Kind_txk fx1 fx1 {} T_v_fxempty 
                      (split_nil_A_1 fx1) _ (refl_equal (t_e1, fx1))
                      A1 x_notin_A1 wf_GA1 (aq_U_id A1) G_eq T1) as T_sub_e1.
      pose proof (IHT2 wf_GA2 G2 G1 wf_G1 Kind_txk fx2 fx2 {} T_v_fxempty 
                      (split_nil_A_1 fx2) _ (refl_equal (t2, fx2))
                      A2 x_notin_A2 wf_GA2 (aq_U_id A2) G_eq T2) as T_sub_e2.
      clear IHT1 IHT2.
      eapply type_APP.
        eauto.
        destruct o as [ K_U | K_A ].
          left.
            assert (subst_typ x v t_e1 = typ_fun (subst_typ x v t2) (subst_typ x v t)) as EQ.
              subst t_e1; (simpl; auto).
            apply EQ.
          right.
            assert (subst_typ x v t_e1 = typ_qtype (typ_fun (subst_typ x v t2) (subst_typ x v t))) as EQ.
              subst t_e1; (simpl; auto).
            apply EQ.
        eauto. eauto. eauto. clear IHT1 IHT2. 
   (* Finally ... conclude with T-AFN *)
      eapply type_AFN.
        apply (split_nil_A_2 A_minus_x).
        apply premise_TAFN.
        eapply (split_commutes _ _ _ split_fx).
        subst G.
      eapply effect_ok_subst.
        apply wf_GA. 
        eapply typing_effect_ok. 
          apply T_app.
        eapply typing_effect_ok.
          apply T_v. apply (split_commutes _ _ _ split_fx).
Case "type_TAP_NOFX".
intros G2 G1 wf_G1 Kind_txk fx_tapv fx_tap fx_v T_v split_fx t0 eq_tfx A_minus_x x_notin_Amx wf_GA A_eq G_eq T_tap.
injection eq_tfx; intros. subst t0. subst fx.
simpl.
assert (subst_typ x v (open_typ_typ 0 t t') = open_typ_typ 0 (subst_typ x v t) (subst_typ x v t')) as subst_open_weak_eq.
  eapply subst_typ_open_typ_typ_commute_weak; eauto using typing_lc, wf_G1, T_v.
rewrite subst_open_weak_eq.
    pose proof (IHT WF_G1xG2Ax G2 G1 wf_G1 Kind_txk fx_tapv fx_tap fx_v T_v 
                    split_fx _ (refl_equal (te, fx_tap))
                    A_minus_x x_notin_Amx WF_G1xG2Ax A_eq G_eq T) as T_sub_e.
    assert (wf_env G {}) as wf_GEmpty. eauto using wf_split1, split_nil_A_1, split_nil_A_2.
    pose proof (IHT0 G1 G2 _ _ _ _ G_eq wf_GEmpty wf_G1 Kind_txk T_v split_fx) as K_sub_t.
    clear IHT IHT0.
    eapply type_TAP_NOFX.
       destruct o.
         left.
          assert (subst_typ x v te = typ_univ k0 NO_FX (subst_typ x v t')) as EQ.
            subst te; simpl. auto.
          apply EQ.
         right.
          assert (subst_typ x v te = typ_qtype (typ_univ k0 NO_FX (subst_typ x v t'))) as EQ.
            subst te; simpl; auto.
          apply EQ.
       apply T_sub_e. apply K_sub_t.
Case "type_TAP_FX".
intros G2 G1 wf_G1 Kind_txk fx_tapv fx_tap fx_v T_v split_fx t0 eq_tfx A_minus_x x_notin_Amx wf_GA A_eq G_eq T_tap.
injection eq_tfx; intros. subst t0. subst fx'.
simpl.
assert (subst_typ x v (open_typ_typ 0 a t') = open_typ_typ 0 (subst_typ x v a) (subst_typ x v t')) as subst_open_weak_eq.
  eapply subst_typ_open_typ_typ_commute_weak; eauto using typing_lc, wf_G1, T_v.
rewrite subst_open_weak_eq. clear subst_open_weak_eq. rename fx into fx_e.
    assert (exists fx_ev, splitA fx_v fx_e fx_ev) as FX_ev. eauto using split_trans_2.
    destruct FX_ev as [fx_ev split_fx_v_e].
    pose proof (IHT WF_G1xG2Ax G2 G1 wf_G1 Kind_txk fx_ev fx_e fx_v T_v 
                    split_fx_v_e _ (refl_equal (te, fx_e))
                    A_minus_x x_notin_Amx WF_G1xG2Ax A_eq G_eq T) as T_sub_e.
    assert (wf_env G {}) as wf_GEmpty. eauto using wf_split1, split_nil_A_1, split_nil_A_2.
    pose proof (IHT0 G1 G2 _ _ _ _ G_eq wf_GEmpty wf_G1 Kind_txk T_v split_fx) as K_sub_t.
    clear IHT IHT0. simpl in K_sub_t. simpl.
    eapply type_TAP_FX.
       destruct o.
         left.
          assert (subst_typ x v te = typ_univ kind_N FX (subst_typ x v t')) as EQ.
            subst te; simpl. auto.
          apply EQ.
         right.
          assert (subst_typ x v te = typ_qtype (typ_univ kind_N FX (subst_typ x v t'))) as EQ.
            subst te; simpl; auto.
          apply EQ.
       apply T_sub_e. apply K_sub_t.
         eauto using split_trans_6.
Case "type_BOT".
intros.
injection Heqt_fxe; intros; subst fx.
simpl.
eapply type_AFN.
  eauto using split_nil_A_1.
  eauto using type_BOT.
  apply split_fx.
  assert (effect_ok fx_v G1) as ok_G1. (* fourth premise of T-AFN *)
     eapply typing_effect_ok; auto. apply T_v.
  assert (effect_ok fx_e G) as ok_G.
     eapply typing_effect_ok; eauto using T_v.
  subst G. rewrite HeqAx in WF_G1xG2Ax.
  assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2G1.
     eapply subst_gamma_wf. eauto.
     eapply typing_wf_exp. apply T_v. 
     apply x_notin_A.
  eapply effect_ok_subst; eauto using split_commutes.
Case "type_DROP".
intros. injection Heqt_fxe; intros; subst t0; subst fx.
simpl.
eapply type_DROP.
  pose proof (IHT WF_G1xG2Ax G2 G1 wf_G1 Kind_txk fx_ev fx_e fx_v T_v split_fx _ (refl_equal (typ_sing t a, fx_e)) A0 x_notin_A WF_G1xG2Ax HeqAx HeqG T) as fromIH.
  simpl in fromIH.
  apply fromIH.
Case "typing_AFN".
intros G2 G1 wf_G1 Kind_txk fx_ev fx_e fx_v T_v split_fx t0 inj Amx x_notin_Amx wf_GA A_eq G_eq T_e.
injection inj; intros; subst fx; subst t0.
destruct (aqual_cases Amx x k) as [A_eq_Amx | A_eq_xAmx ].
  SCase "A_eq_Amx".
    destruct (exclude_middle_empty_sets fx_v) as [fx_v_empty | fx_v_nonempty].
       SSCase "fx_v_empty".
         subst fx_v.
         assert (fx_e = fx_ev) as fx_e_eq_ev.
           eapply split_nil_eq_1; eauto using split_fx.
         subst fx_ev.
         rewrite A_eq_Amx in A_eq. subst A.         
         assert (k = kind_U) as k_is_U.
           assert (kinding G1 tx kind_U) as kinding_U.
             eapply inversion_empty_effect; eauto using T_v.
           eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
         subst k.
         assert (x `notin` A1) as x_notin_A1.
           inversion s; fsetdec.
         eapply type_AFN.
            apply s.
            assert (wf_env G A1) as wf_GA1; eauto using wf_split1.
            assert (aqual x kind_U A1 = A1) as aqU_A1; simpl; auto.
            apply (IHT wf_GA1 G2 G1 wf_G1 Kind_txk fx1 fx1 {} T_v (split_nil_A_1 fx1) t (refl_equal (t, fx1)) A1 x_notin_A1 wf_GA1 aqU_A1 G_eq T).
            apply s0.
            subst G.
            eapply effect_ok_subst; eauto using typing_effect_ok, T_v, T_e, split_nil_A_2.
       SSCase "fx_v_nonempty".
         destruct (inversion_nonempty_effect _ _ _ _ _ Val_v T_v fx_v_nonempty) as [ tx_kind_A | T_v_fxempty ].
   (** SubSubCase 2a. tx_kind_A **)
         SSSCase "tx_kind_A".
           assert (k = kind_A) as k_is_A.
             eapply kinding_uniqueness; eauto using tx_kind_A, Kind_txk.
           assert (x `in` A) as x_in_A.
             subst k; unfold aqual in A_eq; simpl in A_eq; fsetdec.
           subst k. subst A. rewrite <- A_eq_Amx in x_notin_Amx. simpl in x_notin_Amx. fsetdec.
         SSSCase "T_v_fxempty".
           assert (k = kind_U) as k_is_U.
             assert (kinding G1 tx kind_U) as kinding_U.
               eapply inversion_empty_effect; eauto using T_v_fxempty.
             eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
           subst k. simpl in A_eq. subst A.
           assert (x `notin` A1) as x_notin_A1.
            inversion s. fsetdec.
           assert (exists fx_v_2, splitA fx_v fx2 fx_v_2) as FX_V2. 
              eapply split_trans_4. apply s0. apply split_fx.
           destruct FX_V2 as [fx_v_2 split_fx_v_2].
           assert (splitA fx1 fx_v_2 fx_ev) as FX.
               eapply split_trans_6.
                 apply split_fx.
                 apply s0.
                 apply split_fx_v_2.
           eapply type_AFN.
            apply s.
            assert (wf_env G A1) as wf_GA1; eauto using wf_split1.
            assert (aqual x kind_U A1 = A1) as aqU_A1; simpl; auto.
            apply (IHT wf_GA1 G2 G1 wf_G1 Kind_txk fx1 fx1 {} T_v_fxempty (split_nil_A_1 fx1) t (refl_equal (t, fx1)) A1 x_notin_A1 wf_GA1 aqU_A1 G_eq T).
            apply FX.
            subst G.
            eapply effect_ok_subst.
              apply WF_G1xG2Ax.
              eapply typing_effect_ok.
                 apply T_e.
              eapply typing_effect_ok.
                 apply T_v.
              apply (split_commutes _ _ _ split_fx).
   SCase "A_eq_xAmx".
       assert (x `in` A) as x_in_A.
         rewrite A_eq_xAmx in A_eq. fsetdec.
       assert ((x `in` A1 -> x `notin` A2) /\ (x `in` A2 -> x `notin` A1)) as x_splits.
         inversion s.
         split. 
           intro x_in_A1. unfold not. intro x_in_A2. 
           pose proof (AtomSet.F.inter_3 x_in_A1 x_in_A2) as for_contra. rewrite H in for_contra; fsetdec.
           intro x_in_A2. unfold not. intro x_in_A1. 
           pose proof (AtomSet.F.inter_3 x_in_A1 x_in_A2) as for_contra. rewrite H in for_contra; fsetdec.
       assert (x `in` A1 \/ x `in` A2) as x_somewhere.
         inversion s. rewrite <- H2 in x_in_A. auto using AtomSet.F.union_1.
       assert (forall A, aqual x kind_A A = singleton x `union` A) as aqA_identity; (unfold aqual; (simpl; auto)).
       assert (k = kind_A) as k_is_A.
         unfold aqual in A_eq_xAmx.
           destruct k; (auto || (rewrite A_eq_xAmx in x_notin_Amx; fsetdec)). 
       subst k.
       destruct x_somewhere as [x_in_A1 | x_in_A2].
       SSCase "x_in_A1, x_notin_A2".  
         simpl in A_eq.
         rewrite A_eq in s.
         assert (exists A1mx, x `notin` A1mx /\ A1 = singleton x `union` A1mx) as A1MXInner; eauto using split_singleton.
         destruct A1MXInner as [A1mx [x_notin_A1mx A1_eq]].
         assert (splitA A1mx A2 Amx) as split_A1mx_A2; eauto using split_trans_7.
         assert (wf_env G A1) as wf_GA1; (subst A; eauto using wf_split1).
         assert (exists fx_1_v, splitA fx_v fx1 fx_1_v) as FX_V1.
             eapply split_trans_2; eauto using split_commutes.
         destruct FX_V1 as [fx_1_v split_fx_1_v].
         assert (splitA fx2 fx_1_v fx_ev) as FX.
            eapply split_trans_6.
              apply split_fx.
              apply (split_commutes _ _ _ s0).
              apply split_fx_1_v.
         rewrite <- (aqA_identity A1mx) in A1_eq.
         eapply type_AFN. 
            apply split_A1mx_A2.
            apply (IHT wf_GA1 G2 G1 wf_G1 Kind_txk fx_1_v fx1 fx_v T_v split_fx_1_v _ (refl_equal (t, fx1)) 
                        A1mx x_notin_A1mx wf_GA1 A1_eq G_eq T).
            apply (split_commutes _ _ _ FX).
            subst G.
            eapply effect_ok_subst.
              apply WF_G1xG2Ax.
              eapply typing_effect_ok.
                apply T_e.
              eapply typing_effect_ok.
                apply T_v.
              apply (split_commutes _ _ _ split_fx).
       SSCase "x_in_A2, x_notin_A1".
         assert (x `notin` A1) as x_notin_A1; eauto using x_splits.
         assert (wf_env G A1) as wf_GA1; eauto using wf_split1.
         clear IHT.
         assert (exists fx_1_v, splitA fx_v fx1 fx_1_v) as FX_V1.
             eapply split_trans_2; eauto using split_commutes.
         destruct FX_V1 as [fx_1_v split_fx_1_v].
         subst G.
         pose proof (nofree_subst_preserves_typing G1 G2 A1 x tx e t v fx1 fx_v fx_1_v _ wf_sig
                          wf_G1 Kind_txk x_notin_A1 wf_GA1 T Val_v T_v split_fx_1_v) as T_sub.
         assert (exists fx_v_2, splitA fx_v fx2 fx_v_2) as FX_V2.
             eapply split_trans_2. apply split_fx. eapply s0.
         destruct FX_V2 as [fx_v_2 split_fx_v_2].
         assert (exists A2mx, x `notin` A2mx /\ A2 = singleton x `union` A2mx) as A2MXInner; eauto using split_singleton.
         destruct A2MXInner as [A2mx [x_notin_A2mx A2_eq]].
         assert (splitA A2mx A1 Amx) as split_A1_A2mx.
           eapply split_trans_7.
             apply (split_commutes _ _ _ s).
             rewrite A_eq_xAmx in A_eq; apply A_eq.
             apply A2_eq.
             apply x_notin_Amx.
             apply x_notin_A2mx.
        assert (splitA fx1 fx_v_2 fx_ev) as FX.
            eapply split_trans_6.
              apply split_fx.
              apply s0.
              apply split_fx_v_2.
         eapply type_AFN.
            apply (split_commutes _ _ _ split_A1_A2mx).
            apply T_sub.
            apply FX.
            eapply effect_ok_subst.
              apply WF_G1xG2Ax.
              eapply typing_effect_ok.
                apply T_e.
              eapply typing_effect_ok.
                apply T_v.
              apply (split_commutes _ _ _ split_fx).
Case "type_B".
intros. injection Heqt_fxe;intros;subst t0;subst fx_e.
simpl.
unfold wf_signature in wf_sig.
  assert (wf_typ t nil) as wft. 
     destruct (wf_sig B t kind_U) as [wft _].
     eauto using wft.
  assert (subst_typ x v t = t) as tid; eauto using subst_wf_typ_ext.
  rewrite tid.
  assert (fx_v = fx_ev). 
    eapply split_nil_eq_2; eauto.
  subst G. subst fx_ev.
  assert (wf_env (subst_gamma x v G2 ++ G1) A) as wf.
    eapply subst_gamma_wf. rewrite HeqAx in WF_G1xG2Ax. apply WF_G1xG2Ax.
    eauto using typing_wf_exp. 
    assumption.
eapply type_AFN.
  eapply split_nil_A_1; eauto.
  eapply type_B. assumption.
  eapply split_nil_A_1; eauto.
  eapply effect_ok_weakening; eauto using typing_effect_ok.
Case "typing_B1".
rename G1 into G1_1.
rename G2 into G1_2.
intros. injection Heqt_fxe;intros; subst t0; subst fx.
simpl.
destruct (cert_is_wf _ _ _ _ _ _ Te) as [G' [wf_cert_G' onG']].
assert (subst_typ x v t = t) as tid.
  eapply subst_wf_exp_typing_on; eauto.
rewrite tid.
pose proof (IHT WF_G1xG2Ax G2 G1 wf_G1 Kind_txk  _ _ _ T_v split_fx _ 
                (refl_equal (t, fx_e)) A0 x_notin_A WF_G1xG2Ax HeqAx HeqG T) as fromIH.
simpl in fromIH.
rewrite tid in fromIH.
   assert (subst_exp x v v0 = v0) as v0id.
     assert (x `notin` dom G1_1) as xnotin_G1_1.
       eapply binds_no_vars_sub; eauto.
   eapply subst_wf_exp_ext; eauto.
rewrite v0id in fromIH.
destruct (only_names_prefix _ _ _ _ _ _  HeqG o) as [G1_1' G1_eq].
subst G1.
rewrite_env ((subst_gamma x v G2 ++ G1_1') ++ G1_1).
eapply type_B1; (eauto || (simpl_env; eauto)).
Case "typing_B2".
rename G1 into G1_1.
rename G2 into G1_2.
intros. injection Heqt_fxe;intros; subst t0; subst fx.
simpl.
destruct (cert_is_wf _ _ _ _ _ _ Te) as [G' [wf_cert_G' onG']].
assert (subst_typ x v t = t) as tid.
  eapply subst_wf_exp_typing_on; eauto.
rewrite tid.
pose proof (IHT WF_G1xG2Ax G2 G1 wf_G1 Kind_txk  _ _ _ T_v split_fx _ 
                (refl_equal (t, fx_e)) A0 x_notin_A WF_G1xG2Ax HeqAx HeqG T) as fromIH.
simpl in fromIH.
rewrite tid in fromIH.
assert (subst_typ x v t' = t') as tid'.
   eapply subst_wf_typ_ext; eauto using binds_no_vars_sub.
rewrite tid' in fromIH.
destruct (only_names_prefix _ _ _ _ _ _  HeqG o) as [G1_1' G1_eq].
subst G1.
rewrite_env ((subst_gamma x v G2 ++ G1_1') ++ G1_1).
eapply type_B2; (eauto || (simpl_env; eauto)).
Case "type_B3".
intros. injection Heqt_fxe;intros; subst t0; subst fx_e.
simpl.
destruct (cert_is_wf _ _ _ _ _ _ Te) as [G' [wf_cert_G' onG']].
assert (subst_typ x v t = t) as tid.
  eapply subst_wf_exp_typing_on; eauto.
eapply type_B3.
  change (typing (subst_gamma x v G2 ++ G1) A0 (subst_exp x v (exp_base B)) (subst_typ x v t, fx_ev)).
  eauto using IHT.
Case "typing_CONV".
intros G2 G1 wf_G1 Kind_txk fx_ev fx_e fx_v T_v split_fx t1 t_fxe Amx x_notin_Amx wf_GA A_eq G_eq T_e0.
injection t_fxe; intros; subst t1; subst fx.
pose proof (IHT wf_GA G2 G1 wf_G1 Kind_txk fx_ev fx_e fx_v T_v split_fx _ (refl_equal (t, fx_e)) Amx x_notin_Amx 
             wf_GA A_eq G_eq T) as fromIHT.
pose proof (IHT0 x v) as fromIHT0.
eapply type_CONV; eauto.
(* ******************** KINDING JUDGMENT CASES *********************)
Case "kinding_A".
rename WF_G1xG2Ax into G1.
intros G2 _ fx_v _ _ G_eq wf_G wf_G1 _ T_v _.
simpl. 
subst G.
assert (binds a (binding_kind k0) (subst_gamma x v G2 ++ G1)) as binds.
destruct (binds_concat_inv _ _ _ _ _ b) as [[notin_G2 G1x_binds] | G2Binds].
 SCase "G1x_binds".
destruct (binds_concat_inv _ _ _ _ _ G1x_binds) as [[notin_x G1_binds] | xBinds].
 SSCase "G1_binds".
   assert (dom G2 = dom (subst_gamma x v G2)) as dom_eq.
     eauto using subst_gamma_dom.
   rewrite dom_eq in notin_G2.
   eapply binds_tail; eauto.
 SSCase "xBinds".
   pose proof (binds_singleton_inv _ _ _ _ _ xBinds) as contra; destruct contra.
   discriminate.
 SSCase "G2Binds".
   eapply binds_head; eauto using subst_gamma_kind_invariant.
eapply kinding_A; eauto.
Case "kinding_N".
rename WF_G1xG2Ax into G1.
intros G2 fx_e fx_v fx_ev k_v G_eq wf_G wf_G1 Kind_txk T_v split_fx.
pose proof (IHT G1 G2 _ fx_v _ _ G_eq wf_G wf_G1 Kind_txk T_v split_fx) as P1.
pose proof (IHT0 G1 G2 _ fx_v _ _ G_eq wf_G wf_G1 Kind_txk T_v split_fx) as P2.
simpl.
eapply kinding_N; eauto using P1, P2.
Case "kinding_TC".
rename WF_G1xG2Ax into G1.
intros. simpl.
eapply kinding_TC; eauto.
Case "kinding_AFN".
rename WF_G1xG2Ax into G1.
intros G2 fx_e fx_v fx_ev k_v G_eq wf_G wf_G1 Kind_txk T_v split_fx.
pose proof (IHT G1 G2 _ fx_v _ _ G_eq wf_G wf_G1 Kind_txk T_v split_fx) as P1.
simpl.
eapply kinding_AFN; eauto.
Case "kinding_ALL".
rename WF_G1xG2Ax into G1.
intros G2 fx_e fx_v fx_ev k_v G_eq wf_G wf_G1 Kind_txk T_v split_fx. simpl.
pick fresh b excluding ( L `union` singleton x `union` dom G) and apply kinding_ALL.
assert ([(b, binding_kind k0)] ++ G = ([(b, binding_kind k0)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
  subst G. simpl_env. reflexivity.
assert (b `notin` L) as b_notin_L. fsetdec.
assert (wf_env ([(b, binding_kind k0)]++G) {}) as wf_bG.
  eapply wf_env_kind. 
    assumption.
    fsetdec.
pose proof (H b b_notin_L G1 _ _ _ _ _ bG_eq wf_bG wf_G1 Kind_txk T_v split_fx) as P1.
assert (subst_typ x v (open_typ_typ 0 b t) = (open_typ_typ 0 b (subst_typ x v t))) as eq.
  eapply subst_typ_open_typ_typ_commute.
    eapply typing_lc. 
     apply wf_G1. apply T_v.
rewrite <- eq.
apply P1.
Case "kinding_FUN".
rename WF_G1xG2Ax into G1.
intros G2 fx_e fx_v fx_ev k_v G_eq wf_G wf_G1 Kind_txk T_v split_fx.
simpl.
pick fresh y excluding ( L `union` singleton x `union` dom G) and apply kinding_FUN.
  eapply IHT; eauto.
  rename t_x into ty.
  assert ([(y, binding_typ ty)] ++ G = ([(y, binding_typ ty)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as yG_eq.
    subst G. simpl_env. reflexivity.
  assert (y `notin` L) as y_notin_L. fsetdec.
  assert (wf_env ([(y, binding_typ ty)]++G) {}) as wf_yG.
    eapply wf_env_typ. 
    assumption.
    fsetdec.
    eapply kinding_wf_typ; eauto.
pose proof (H y y_notin_L G1 _ _ _ _ _ yG_eq wf_yG wf_G1 Kind_txk T_v split_fx) as P1.
assert (subst_typ x v (open_typ 0 y t) = (open_typ 0 y (subst_typ x v t))) as eq.
  eapply subst_typ_open_typ_commute.
    fsetdec.
    eapply typing_lc. 
     apply wf_G1. apply T_v.
rewrite <- eq.
apply P1.
Case "kinding_TAP".
rename WF_G1xG2Ax into G1.
intros G2 fx_e fx_v fx_ev k_v G_eq wf_G wf_G1 Kind_txk T_v split_fx.
simpl.
pose proof (IHT _ _ _ _ _ _ G_eq wf_G wf_G1 Kind_txk T_v split_fx) as P1.
pose proof (IHT0 _ _ _ _ _ _ G_eq wf_G wf_G1 Kind_txk T_v split_fx) as P2.
eapply kinding_TAP; eauto.
Case "kinding_DEP".
rename WF_G1xG2Ax into G1.
intros G2 fx_e fx_v fx_ev k_v G_eq wf_G wf_G1 Kind_txK T_v split_fx.
simpl.
assert (affines G = {}) as afG; (unfold affines; (simpl; auto)).
assert (x `notin` affines G) as x_notin_afG.
   rewrite afG; fsetdec.
rewrite <- afG in wf_G. subst G.
pose proof (nofree_subst_preserves_typing _ _ _ _ _ _ _ _ _ _ _ _
               wf_sig wf_G1 Kind_txK x_notin_afG wf_G T Val_v T_v (split_nil_A_2 fx_v)) as P1.
pose proof (IHT G1 G2 _ _ _ _ (refl_equal  (G2 ++ [(x, binding_typ tx)] ++ G1))
                wf_G wf_G1 Kind_txK T_v split_fx) as P2.
clear IHT IHT0.
assert (wf_typ t' nil) as wf_t'_nil.
  assert (wf_kind (kind_dcon t' K)) as wf_k.
    eapply kinding_wf_kind; eauto.
  inversion wf_k.
    inversion H.
    eapply kinding_wf_typ; eauto.
assert (subst_typ x v t' = t') as st'_eq.
  eapply subst_wf_typ_ext; eauto.
rewrite st'_eq in P1.
eapply kinding_DEP; eauto.
Case "kinding_DROPPED".
rename WF_G1xG2Ax into G1.
intros G2 fx_e fx_v fx_ev k_v G_eq wf_G wf_G1 Kind_txK T_v split_fx.
simpl.
eapply kinding_DROPPED; (eapply IHT; eauto).
(****************************** tequiv judgment cases *********************)
Case "te_ID".
rename WF_G1xG2Ax into y.
intros. apply te_ID.
Case "te_SYM".
rename WF_G1xG2Ax into y.
intros. eapply te_SYM; eauto. 
Case "te_CTX_FUN_1".
rename WF_G1xG2Ax into y.
intros. simpl. 
eapply te_CTX_FUN_1; eauto.
Case "te_CTX_FUN_2".
rename WF_G1xG2Ax into y.
intros. simpl. 
eapply te_CTX_FUN_2; eauto.
Case "te_CTX_UNIV".
rename WF_G1xG2Ax into y.
intros. simpl.
eapply te_CTX_UNIV; eauto.
Case "te_CTX_Q".
rename WF_G1xG2Ax into y.
intros. simpl.
eapply te_CTX_Q; eauto.
Case "te_CTX_APP1".
rename WF_G1xG2Ax into y.
intros. simpl.
eapply te_CTX_APP1; eauto.
Case "te_CTX_APP2".
rename WF_G1xG2Ax into y.
intros. simpl.
eapply te_CTX_APP2; eauto.
Case "te_CTX_SING".
rename WF_G1xG2Ax into y.
intros. simpl.
eapply te_CTX_SING; eauto.
Case "te_CTX_DEP_1".
rename WF_G1xG2Ax into y.
intros. simpl.
eapply te_CTX_DEP_1; eauto.
Case "te_RED".
rename WF_G1xG2Ax into y.
intros. simpl.
assert (wf_exp e nil) as wf_e. eauto using typing_wf_exp.
assert (wf_exp e' nil) as wf_e'. eauto using typing_wf_exp.
assert (subst_exp y v0 e = e) as e_eq. eapply subst_wf_exp_ext; ( eauto || fsetdec).
assert (subst_exp y v0 e' = e') as e_eq'. eapply subst_wf_exp_ext; ( eauto || fsetdec).
rewrite e_eq. rewrite e_eq'.
eapply te_RED; eauto.
Case "te_CTX_DROPPED".
rename WF_G1xG2Ax into y.
intros. simpl.
eapply te_CTX_DROPPED; (eapply IHT; eauto).
Qed.

End LAIR_Subst.
