Require Export Lair_Infrastructure.

Module LAIR_Tequiv (Import X : LAIR_PARAMS).
  Module Export InfrastructureFacts := LAIR_Infrastructure X.

Lemma tequiv_fun_injection : forall t1 t2 t1' t2', 
  tequiv (typ_fun t1 t2) (typ_fun t1' t2') -> (tequiv t1 t1') /\ (tequiv t2 t2').
Proof.
intros.
remember (typ_fun t1 t2) as T1.
remember (typ_fun t1' t2') as T2.
revert t1 t2 t1' t2' HeqT1 HeqT2.
induction H; try solve [intros; discriminate].
Case "te_ID".
intros. subst. injection HeqT2; intros; subst. split; eauto using te_ID. 
Case "te_SYM".
intros.
pose proof (IHtequiv _ _ _ _ HeqT2 HeqT1) as IH. 
destruct IH; eauto using te_SYM.
Case "te_CTX_FUN_1".
intros. injection HeqT1; intros; subst. injection HeqT2; intros; subst.
split; [assumption | eauto using te_ID].
Case "te_CTX_FUN_2".
intros. injection HeqT1; intros; subst. injection HeqT2; intros; subst.
split; [eauto using te_ID | assumption ].
Qed.

Lemma tequiv_qtype_injection : forall t1 t2, 
  tequiv (typ_qtype t1) (typ_qtype t2) -> tequiv t1 t2.
Proof.
 intros t1 t2 Te.
remember (typ_qtype t1) as T1.
remember (typ_qtype t2) as T2.
revert t1 t2 HeqT1 HeqT2.
induction Te; try solve [intros; subst; (discriminate || eauto)].
Case "te_ID".
intros. subst.
 injection HeqT2; intros; subst; eapply te_ID; eauto.
Case "te_SYM".
 intros. subst.
  pose proof (IHTe _ _ (refl_equal (typ_qtype t2)) (refl_equal (typ_qtype t1))) as IH.
  eauto using te_SYM.
Case "te_CTX_Q".
 intros.
   injection HeqT1; injection HeqT2; intros; subst.
 assumption.
Qed.


Lemma tequiv_preserves_fun_shape_helper : forall t1 t2, 
  tequiv t1 t2 -> ((forall t1_1 t1_2, t1 = typ_fun t1_1 t1_2 -> exists t2_1, exists t2_2, t2 = typ_fun t2_1 t2_2) /\
                   (forall t2_1 t2_2, t2 = typ_fun t2_1 t2_2 -> exists t1_1, exists t1_2, t1 = typ_fun t1_1 t1_2)).
Proof.
intros t1 t2 Te.
induction Te; try solve [split; (intros ti_1 ti_2 teq; (discriminate || eauto))].
Case "te_SYM".
destruct IHTe as [IHT1 IHT2].
split; (intros; subst; eauto using IHT1, IHT2).
Qed.

Lemma tequiv_preserves_fun_shape : forall t1 t2 t, 
  tequiv t (typ_fun t1 t2)-> exists t1', exists t2', t = typ_fun t1' t2'.
Proof.
intros.
destruct (tequiv_preserves_fun_shape_helper _ _ H).
eapply H1; eauto using refl_equal.
Qed.

Lemma tequiv_preserves_univ_shape_helper : forall t1 t2, 
  tequiv t1 t2 -> ((forall t1' k fxtag, t1 = typ_univ k fxtag t1' -> exists t2', t2 = (typ_univ k fxtag t2') /\ tequiv t1' t2') /\
                   (forall t2' k fxtag, t2 = typ_univ k fxtag t2' -> exists t1', t1 = (typ_univ k fxtag t1') /\ tequiv t2' t1')).
Proof.
intros t1 t2 Te.
induction Te; try solve [split; (intros ti' k fxtag tieq; (discriminate || eauto using te_ID))].
Case "te_SYM".
destruct IHTe as [IHT1 IHT2]; (split; intros; subst; eauto using IHT1, IHT2).
Case "te_CTX_UNIV".
split.
 intros ti' k0 fxtag tieq; injection tieq; intros; subst.
   exists t'. auto. 
 intros ti' k0 fxtag tieq; injection tieq; intros; subst.
   exists t. eauto using te_SYM.
Qed.

Lemma tequiv_preserves_univ_shape : forall t1 t2 k fxtag, 
  tequiv t1 (typ_univ k fxtag t2) -> (exists t1', t1 = (typ_univ k fxtag t1') /\ tequiv t2 t1').
Proof.
intros.
destruct (tequiv_preserves_univ_shape_helper _ _ H).
eapply H1; eauto using refl_equal, te_SYM.
Qed.

Lemma tequiv_preserves_qtyp_shape_helper : forall t1 t2,
  tequiv t1 t2 -> ((forall t1', t1 = typ_qtype t1' -> exists t2', t2 = typ_qtype t2' /\ tequiv t1' t2') /\
                   (forall t2', t2 = typ_qtype t2' -> exists t1', t1 = typ_qtype t1' /\ tequiv t1' t2')).
Proof.
intros t1 t2 Te.
induction Te; try solve [split; (intros ti' teq; (discriminate || eauto using te_ID))].
Case "te_SYM".
rename t' into t1.
rename t into t2.
destruct IHTe as [IHT1 IHT2].
split.
  intros. 
  pose proof (IHT2 t1' H) as IH.
   destruct IH as [t2' [t1_eq teqv]].
   eauto using te_SYM.
  intros.
  pose proof (IHT1 t2' H) as IH.
    destruct IH as [t1' [t2_eq teqv]].
    eauto using te_SYM.
Case "te_CTX_Q".
split; (intros; injection H; intros; subst; eauto).
Qed.

Lemma tequiv_preserves_qtyp_shape : forall t1 t2 t, 
  tequiv t (typ_qtype (typ_fun t1 t2)) -> exists t1', exists t2',  t = (typ_qtype (typ_fun t1' t2')).
Proof.
intros.
destruct (tequiv_preserves_qtyp_shape_helper _  _ H) as [Q1 Q2].
destruct (Q2 _ (refl_equal (typ_qtype (typ_fun t1 t2)))) as [t' [teq teqv]].
destruct (tequiv_preserves_fun_shape _ _ _ teqv) as [t1' [t2' teqv']].
subst. eauto.
Qed.

Lemma tequiv_preserves_qtyp_univ_shape : forall t1 k fxtag t2, 
  tequiv t1 (typ_qtype (typ_univ k fxtag t2)) -> (exists t1', t1 = (typ_qtype (typ_univ k fxtag t1')) /\ tequiv t1' t2).
Proof.
intros.
destruct (tequiv_preserves_qtyp_shape_helper _  _ H) as [Q1 Q2].
destruct (Q2 _ (refl_equal (typ_qtype (typ_univ k fxtag t2)))) as [t' [teq teqv]].
destruct (tequiv_preserves_univ_shape _ _ _ _ teqv) as [t1' [t2' teqv']].
subst. eauto using te_SYM.
Qed.

Lemma tequiv_preserves_kinding : forall G t t' k, 
  kinding G t k -> tequiv t t' -> kinding G t' k.
Admitted.

Lemma wf_env_tequiv : forall G1 G2 x A tx tx',
  wf_env (G2 ++ [(x, binding_typ tx)] ++ G1) A ->
  tequiv tx tx' -> 
  wf_env (G2 ++ [(x, binding_typ tx')] ++ G1) A.
Admitted.

Lemma tequiv_soundness : forall G1 G2 A x tx tx' e t fx, 
  wf_env (G2 ++ [(x, binding_typ tx)] ++ G1) A ->
  typing (G2 ++ [(x, binding_typ tx)] ++ G1) A e (t, fx) -> 
  tequiv tx tx' -> 
  typing (G2 ++ [(x, binding_typ tx')] ++ G1) A e (t, fx).
Proof.
intros until 1. rename H into wf_G.
intros T Tequiv.
assert (typing (G2 ++ [(x, binding_typ tx)] ++ G1) A e (t, fx)) as Te; auto.
remember (G2 ++ [(x, binding_typ tx)] ++ G1) as G.
remember (t, fx) as tfx.
generalize dependent Te. 
generalize dependent HeqG.
generalize dependent wf_G.
generalize dependent t.
generalize dependent fx.
generalize dependent G2.
destruct T using typing_ind2 with
  (P0 := 
    fun (G:gamma) (t:typ) (k:kind) (K:kinding G t k) =>
      forall G1 G2, (G = (G2 ++ [(x, binding_typ tx)] ++ G1)) -> wf_env G {} -> 
        tequiv tx tx' -> kinding (G2 ++ [(x, binding_typ tx')] ++ G1) t k)
  (P1 :=
    fun (t:typ) (t':typ) (Te: tequiv t t') => True).
Case "type_X".
intros.
subst.
destruct (binds_concat_inv binding x0 (binding_typ t) _ G2 b) as [ [x0_Notin_G2 binds_x0_xG1] | binds_x0_G2].
SCase "xG1".
destruct (binds_concat_inv binding x0 (binding_typ t) G1 _ binds_x0_xG1) as [ [ x0_Notin_x binds_x0_G1 ] | binds_x0_x].
 SSCase "G1".
 assert (binds x0 (binding_typ t) (G2 ++ [(x, binding_typ tx')] ++ G1)) as bx0.
   eapply binds_tail; eauto using binds_tail.
 eapply type_X; eauto.
 SSCase "x0".
   pose proof (binds_singleton_inv _ _ _ _ _ binds_x0_x) as conj.
   destruct conj; subst.
   injection Heqtfx; intros; subst.
   injection H0; intros; subst.
   assert (binds x (binding_typ tx') (G2 ++ [(x, binding_typ tx')] ++ G1)) as bx.
     eapply binds_tail; eauto.
   eapply type_CONV.
       eapply type_X; eauto.
       eapply tequiv_preserves_kinding; eauto using IHT, wf_split1, split_nil_A_1.
       eapply te_SYM; eauto.
SCase "G2".
  assert (binds x0 (binding_typ t) (G2 ++ [(x, binding_typ tx')] ++ G1)) as bx0.
    eapply binds_head; eauto.
     eapply type_X; eauto. 
Case "type_XA".
intros.
subst.
destruct (binds_concat_inv binding x0 (binding_typ t) _ G2 b) as [ [x0_Notin_G2 binds_x0_xG1] | binds_x0_G2].
SCase "xG1".
destruct (binds_concat_inv binding x0 (binding_typ t) G1 _ binds_x0_xG1) as [ [ x0_Notin_x binds_x0_G1 ] | binds_x0_x].
 SSCase "G1".
 assert (binds x0 (binding_typ t) (G2 ++ [(x, binding_typ tx')] ++ G1)) as bx0.
   eapply binds_tail; eauto using binds_tail.
 eapply type_XA; eauto 6 using IHT, wf_split1, split_nil_A_1. 
 SSCase "x0".
   pose proof (binds_singleton_inv _ _ _ _ _ binds_x0_x) as conj.
   destruct conj; subst.
   injection Heqtfx; intros; subst.
   injection H0; intros; subst.
   assert (binds x (binding_typ tx') (G2 ++ [(x, binding_typ tx')] ++ G1)) as bx.
     eapply binds_tail; eauto.
   eapply type_CONV.
       eapply type_XA; eauto.
       eapply tequiv_preserves_kinding; eauto using IHT, wf_split1, split_nil_A_1.
       eapply te_SYM; eauto.
SCase "G2".
  assert (binds x0 (binding_typ t) (G2 ++ [(x, binding_typ tx')] ++ G1)) as bx0.
    eapply binds_head; eauto.
  eapply type_XA; eauto 6 using IHT, wf_split1, split_nil_A_1. 
Case "type_NEW".
intros. subst.
  assert (binds a (binding_kind kind_N) (G2 ++ [(x,binding_typ tx')] ++ G1)) as ba.
    assert (x <> a) as x_neq_a.
      unfold not; intro x_eq_a.
      subst.
      pose proof (binds_mid_eq binding a _ _ G1 G2 b (wf_is_ok _ _ wf_G)) as kind_N_eq_t.
      discriminate.
    assert (binds a (binding_kind kind_N) (G2 ++ G1)) as ba'.
      eapply binds_remove_mid; eauto. 
    eapply binds_weaken; eauto using wf_is_ok, wf_env_tequiv.
  eapply type_NEW; eauto using IHT, IHT0, wf_split1, split_nil_A_1.
Case "type_ABS".
intros. subst.
pick fresh y and apply type_ABS.
assert (y `notin` L) as y_notin_L. fsetdec.
rename t_x into t_y.
cut (wf_env ([(y, binding_typ t_y)] ++ G2 ++ [(x, binding_typ tx)] ++ G1) (aqual y k A)). intro wf_G'.
assert ((([(y, binding_typ t_y)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) = 
        ([(y, binding_typ t_y)] ++ G2 ++ [(x, binding_typ tx)] ++ G1)) as G_eq'.
   simpl_env. reflexivity.
apply (H y y_notin_L ([(y, binding_typ t_y)] ++ G2) _ _ (refl_equal (open_typ 0 y t_e, fx)) 
              wf_G' G_eq' (t y y_notin_L)).
(* cut *)
  eapply wf_env_aqual.
    eapply wf_env_typ.
     assumption.
     simpl_env. fsetdec.
     eapply kinding_wf_typ; eauto.
    eapply binds_head; eauto.
eapply IHT; eauto using wf_split1, split_nil_A_1.
Case "type_TAB".
(* TODO: Complete these cases *)
Admitted.

Lemma tequiv_open_red : forall e1 e2 t,
  eval e1 e2 -> tequiv (open_typ 0 e1 t) (open_typ 0 e2 t).
Admitted.

Lemma tequiv_open_equiv : forall t t' x, 
  tequiv t t' -> tequiv (open_typ 0 x t) (open_typ 0 x t').
Admitted.

Lemma tequiv_open_typ_equiv : forall t t' a, 
  tequiv t t' -> tequiv (open_typ_typ 0 a t) (open_typ_typ 0 a t').
Admitted. 

End LAIR_Tequiv.
