Require Import Lair_Definitions.
(*Require Import Lair.*)
Require Import Lair_Subst.

(* I think this notation should be built in. I wonder if it is in newer versions. *)
Notation "[ a , .. , b ]" := (cons a .. (cons b nil) ..).
Notation "[]" := nil.

Parameters
  Int
  LT
  Principal
  Protected
  N
  P
  less_than
  acts_for
  : atom.

Definition prelude : gamma :=
  [
  (Int, binding_kind kind_U),
  (LT, binding_kind (kind_dcon (typ_const Int) (kind_dcon (typ_const Int) kind_U))),
  (Principal, binding_kind kind_U),
  (Protected, binding_kind (kind_tcon kind_U (kind_tcon kind_A kind_U))),
  (N, binding_typ (typ_const Int)),
  (P, binding_typ (typ_const Principal)),
  (less_than, binding_typ (typ_fun (typ_const Int) (typ_fun (typ_const Int) (typ_dep (typ_dep (typ_const LT) 1) 0))))
  ].

  Parameters
    Label
    SLabel
    User
    Labeled
    Zero
    Bool
    TRUE FALSE
    member
    access
    P1 P2
    : atom.

(* This lets you prove anything given the contradictory hypotheses that
   (1) a list [L] has no duplicates, and
   (2) two different elements in [L] are equal
*)
Ltac nodup_neq noDupHyp Heq :=
  (* The actual goal doesn't matter; we'll prove it by proving [False] *)
  cut False ; [apply False_ind | idtac] ;
  (* Get rid of the context, keeping only [Heq] *)
  clear - Heq ;
  (* Put the [NoDup] hypothesis in the context *)
  pose proof noDupHyp ;
  (* Get names for the two sides of [Heq] *)
  match type of Heq with ?LHS = ?RHS =>
  (* Loop, doing [inversion] on the [NoDup] hypothesis, until you
     reach one of the atoms mentioned in [Heq] *)
  repeat (
    match goal with [ HNoDups : NoDup _ |- _ ] =>
      (* Do [inversion] on the current [NoDup] hypothesis; then clear it *)
      inversion HNoDups ; clear HNoDups ; subst ;
      (* Find the name of the new [~In] hypothesis *)
      match goal with [ Hnotin : ~In ?a _ |- _ ] =>
        (* See if the atom in the [~In] hypothesis matches [Heq]. *)
        (* The two matching cases are almost identical, except that LHS *)
        (* requires us to flip [Heq] around. *)
        (* If we match, apply [Hnotin] and simplify the new [In] goal, *)
        (* which leaves a big disjunction. We break down the disjunction *)
        (* until we find the equality we need. *)
        match a with
        | LHS => apply sym_equal in Heq ; apply Hnotin ; simpl ;
           repeat match goal with
                  | [ |- _ = a \/ _ ] => apply or_introl ; exact Heq
                  | _ => apply or_intror
                  end
        | RHS => apply Hnotin ; simpl ;
           repeat match goal with
                  | [ |- _ = a \/ _ ] => apply or_introl ; exact Heq
                  | _ => apply or_intror
                  end
        | _ => clear Hnotin
        end
      end
    end)
  end.

(* This [destruct]s anything of the form [if _ == _ then _ else _].
   It discharges impossible cases when the equality is [x == x]. *)
Ltac destructAtomEqs :=
  try match goal with
      | [ |- context [if ?b == ?b then _ else _] ] =>
        let H := fresh in
        destruct (b==b) as [_ | H] ; [destructAtomEqs | destruct H ; apply refl_equal]
      | [ _ : context [if ?b == ?b then _ else _] |- _ ] =>
        let H := fresh in
        destruct (b==b) as [_ | H] ; [destructAtomEqs | destruct H ; apply refl_equal]
      | [ _ : context [if ?b == ?x then _ else _] |- _ ]  => destruct (b == x) ; destructAtomEqs
      | [ |- context [if ?b == ?x then _ else _] ]  => destruct (b == x) ; destructAtomEqs
      end.

Module Import LairAccessControlParams <: LAIR_PARAMS.

  Definition Signature : gamma :=
    prelude ++
    [
    (Label, binding_kind kind_U),
    (SLabel, binding_kind (kind_dcon (typ_const Label) kind_U)),
    (User, binding_typ (typ_const Label)),
    (Labeled, binding_kind (kind_tcon kind_U (kind_dcon (typ_const Label) kind_U))),
    (Bool, binding_kind kind_U),
    (TRUE, binding_typ (typ_const Bool)),
    (FALSE, binding_typ (typ_const Bool)),
    (member, binding_typ (typ_fun (typ_const Label) (typ_fun (typ_const Label)
      (typ_const Bool)))),
    (access, binding_typ (typ_univ kind_U NO_FX
      (typ_fun (typ_dep (typ_const SLabel) (exp_base User))
        (typ_fun (typ_dep (typ_app (typ_const Labeled) (typ_const Int)) (exp_bvar 0))
          (typ_fun (typ_const Label)
            (typ_fun (typ_dep (typ_app (typ_const Labeled) (typ_btvar 0)) (exp_bvar 0))
              (typ_btvar 0))))))),
    (P1, binding_typ (typ_const User)),
    (P2, binding_typ (typ_const User))
    ].

  (* This is needed to be able to deduce that the base terms in
     [Signature] are distinct. Of course, it is important that no base
     term is repeated within [Signature], because then this axiom
     would be false. *)
  Axiom no_dups_Signature : NoDup (map (@fst atom binding) Signature).

  (* The tactic [distinct] only requires the equality hypothesis.
     It automatically uses the axiom that [Signature] is duplicate-free. *)
  Ltac distinct := nodup_neq no_dups_Signature.

  (* I don't have a good sense of what a Model should be, so I don't have an
     example *)
  Definition Model : model :=
    [].

  Inductive eval : exp -> exp -> Prop :=
  
  | eval_bot_bot : eval exp_bot exp_bot
  
  | eval_app1 : forall e1 e1' e2, 
    eval e1 e1' -> eval (exp_app e1 e2) (exp_app e1' e2)
    
  | eval_app1_bot : forall e1 e2, 
    eval e1 exp_bot -> eval (exp_app e1 e2) exp_bot
    
  | eval_app2 : forall e1 e2 e2', 
    eval e2 e2' -> eval (exp_app e1 e2) (exp_app e1 e2')
  
  | eval_app2_bot : forall e1 e2, 
    eval e2 exp_bot -> eval (exp_app e1 e2) exp_bot
  
  | eval_beta : forall t e v, 
    (value v) -> (eval (exp_app (exp_lam t e) v) (open_exp 0 v e))
  
  | eval_tap1 : forall e1 e1' t, 
    (eval e1 e1') -> (eval (exp_tap e1 t) (exp_tap e1' t))
  
  | eval_tap1_bot : forall e1 t, 
    (eval e1 exp_bot) -> (eval (exp_tap e1 t) exp_bot)
  
  | eval_tap : forall e1 k t, 
    (eval (exp_tap (exp_tab k e1) t) (open_exp_typ 0 t e1))
  
  | eval_new1 : forall e1 e1', 
    (eval e1 e1') ->  (eval (exp_new e1) (exp_new e1'))
  
  | eval_new1_bot : forall e1, 
    (eval e1 exp_bot) ->  (eval (exp_new e1) exp_bot)
  
  | eval_delta : forall a eqns, 
    (binds a eqns Model) -> (eval (exp_base a) (exp_cert (exp_base a) nil))
  
  | eval_b1 : forall b d v eqns v',
    (value (exp_cert (exp_base b) d)) -> (value v) -> 
    (binds b eqns Model) -> (in_eqns ((eort_exp v)::d) eqns v') -> 
    (eval (exp_app (exp_cert (exp_base b) d) v) v')
  
  | eval_b2 : forall b d v eqns,
    (value (exp_cert (exp_base b) d)) -> (value v) -> 
    (binds b eqns Model) -> (not_in_eqns ((eort_exp v)::d) eqns) -> 
    (eval (exp_app (exp_cert (exp_base b) d) v)  (exp_cert (exp_base b) ((eort_exp v)::d)))
  
  | eval_b3 : forall b d t eqns v',
    (value (exp_cert (exp_base b) d)) -> 
    (binds b eqns Model) -> (in_eqns ((eort_typ t)::d) eqns v') -> 
    (eval (exp_tap (exp_cert (exp_base b) d) t)  v')
  
  | eval_b4 : forall b d t eqns,
    (value (exp_cert (exp_base b) d)) -> 
    (binds b eqns Model) -> (not_in_eqns ((eort_typ t)::d) eqns) -> 
    (eval (exp_tap (exp_cert (exp_base b) d) t)  (exp_cert (exp_base b) ((eort_typ t)::d))).
  
  (* Static semantics *)
  
  Inductive typing : gamma -> atoms -> exp -> (typ * atoms) -> Prop:=
  | type_X : 
    forall G x t, 
      (binds x (binding_typ t) G) -> 
      (kinding G t kind_U) ->
      (typing G {} x (t, {}))
  
  | type_XA :
    forall G x t,
      (binds x (binding_typ t) G) ->
      (kinding G t kind_A) ->
      (typing G (singleton x) x (t, {}))
      
  | type_NEW :
    forall G A e t a fx a_fx,
      (typing G A e (t, fx)) ->
      (kinding G t kind_U) ->
      (binds a (binding_kind kind_N) G) ->
      (splitA (singleton a) fx a_fx) ->
      (typing G A (exp_new e) ((typ_sing (typ_qtype t) a), a_fx))
  
  | type_ABS :
    forall L G A t_x k e t_e fx, 
      (forall x, x `notin` L ->
        (typing ([(x,binding_typ t_x)] ++ G) (aqual x k A) (open_exp 0 x e) ((open_typ 0 x t_e), fx))) ->
      (kinding G t_x k) ->
      (typing G A (exp_lam t_x e) ((ptype (typ_fun t_x t_e) A fx), fx))
      
  | type_TAB :
    forall L G A k e t_e fx, 
      (forall a, a `notin` L -> a `notin` fx ->
        (typing ([(a,binding_kind k)] ++ G) A (open_exp_typ 0 a e) ((open_typ_typ 0 a t_e), fx))) ->
      (simple_kind k) ->
      (typing G A (exp_tab k e) ((ptype (typ_univ k NO_FX t_e) A fx), fx))
      
  | type_TAB_FX :
    forall L G A k e t_e fx,
      (AtomSet.F.Subset fx L) ->
      (forall a, a `notin` L -> 
        (typing ([(a,binding_kind k)] ++ G) A (open_exp_typ 0 a e) ((open_typ_typ 0 a t_e), fx `union` singleton a))) ->
      (simple_kind k) ->
      (typing G A (exp_tab k e) ((ptype (typ_univ k FX t_e) A fx), fx))
      
  | type_APP :
    forall G A A1 A2 e1 e2 t2 t fx1 fx2 fx t_e1, 
      (splitA A1 A2 A) ->
      (t_e1 = (typ_fun t2 t) \/ t_e1 = (typ_qtype (typ_fun t2 t))) ->
      (typing G A1 e1 (t_e1, fx1)) ->
      (typing G A2 e2 (t2, fx2)) -> 
      (splitA fx1 fx2 fx) ->
      (typing G A (exp_app e1 e2) ((open_typ 0 e2 t), fx))
  
  (* | type_QAPP : *)
  (*   forall G A A1 A2 e1 e2 t2 t fx1 fx2 fx,  *)
  (*     (splitA A1 A2 A) -> *)
  (*     (typing G A1 e1 (typ_qtype (typ_fun t2 t), fx1)) -> *)
  (*     (typing G A2 e2 (t2, fx2)) ->  *)
  (*     (splitA fx1 fx2 fx) -> *)
  (*     (typing G A (exp_app e1 e2) ((open_typ 0 e2 t), fx)) *)
  
  | type_TAP_NOFX :
    forall G A e t k t' fx te,
      (te = (typ_univ k NO_FX t') \/ te = typ_qtype (typ_univ k NO_FX t')) ->
      (typing G A e (te, fx)) ->
      (kinding G t k) ->
      (typing G A (exp_tap e t) ((open_typ_typ 0 t t'), fx))
  
  | type_TAP_FX :
    forall G A e a t' fx' fx te,
      (te = (typ_univ kind_N FX t') \/ te = typ_qtype (typ_univ kind_N FX t')) ->
      (typing G A e (te, fx)) ->
      (kinding G (typ_ftvar a) kind_N) ->
      (splitA (singleton a) fx fx') -> 
      (typing G A (exp_tap e (typ_ftvar a)) ((open_typ_typ 0 (typ_ftvar a) t'), fx'))
  
  | type_BOT :
    forall G t fx,
      (typing G {} exp_bot (t, fx))
      
  | type_DROP :
    forall G A e t a fx,
      (typing G A e ((typ_sing t a), fx)) ->
      (typing G A e ((typ_dropped t), fx))
  
  | type_AFN :
    forall G A A1 A2 e t fx fx1 fx2, 
      (splitA A1 A2 A) ->
      (typing G A1 e (t, fx1)) -> 
      (splitA fx1 fx2 fx) -> 
      (effect_ok fx G) ->
      (typing G A e (t, fx))
      
  | type_B :
    forall G B t,
      (binds B (binding_typ t) Signature) ->
      (typing G {} (exp_base B) (t, {}))
  
  | type_B1 :
    forall G1 G2 A B domain v t fx,
      (only_names G1 {}) -> 
      (wf_exp v G1) -> 
      (typing (G2++G1) A (exp_app (exp_cert (exp_base B) domain) v) (t,fx)) -> 
      (typing (G2++G1) A (exp_cert (exp_base B) ((eort_exp v)::domain)) (t,fx))
      
  | type_B2 :
    forall G1 G2 A B domain t' t fx,
      (only_names G1 {}) -> 
      (wf_typ t' G1) ->
      (typing (G2++G1) A (exp_tap (exp_cert (exp_base B) domain) t') (t,fx)) -> 
      (typing (G2++G1) A (exp_cert (exp_base B) ((eort_typ t')::domain)) (t,fx))
      
  | type_B3 :
    forall G A B t fx,
      (typing G A (exp_base B) (t, fx)) -> 
      (typing G A (exp_cert (exp_base B) nil) (t,fx))
  
  | type_CONV :
    forall G A e t t' fx,
      (typing G A e (t,fx)) -> 
      (tequiv t t') ->
      (typing G A e (t',fx))
      
  with kinding : gamma -> typ -> kind -> Prop :=
  
  | kinding_A :
    forall G a k,
      (binds a (binding_kind k) G) ->
      (kinding G a k)
      
  | kinding_N :
    forall G t a,
      (kinding G t kind_A) ->
      (kinding G a kind_N) ->
      (kinding G (typ_sing t a) kind_A)
  
  | kinding_TC :
    forall G tc K,
      (binds tc (binding_kind K) Signature) ->
      (kinding G (typ_const tc) K)
  
  | kinding_AFN :
    forall G t,
      (kinding G t kind_U) ->
      (kinding G (typ_qtype t) kind_A)
  
  | kinding_ALL :
    forall L G k t k' fx,  
      (forall a, a `notin` L ->
        (kinding ([(a,binding_kind k)] ++ G) (open_typ_typ 0 a t) k')) ->
      (kinding G (typ_univ k fx t) kind_U)
      
  | kinding_FUN :
    forall L G t_x t k_x k,
      (kinding G t_x k_x) ->
      (forall x, x `notin` L ->
        (kinding ([(x,binding_typ t_x)] ++ G) (open_typ 0 x t) k)) -> 
      (kinding G (typ_fun t_x t) kind_U)
  
  | kinding_TAP :
    forall G t t' k K,
      (kinding G t (kind_tcon k K)) ->
      (kinding G t' k) ->
      (kinding G (typ_app t t') K)
  
  | kinding_DEP :
    forall G t e t' K,
      (kinding G t (kind_dcon t' K)) ->
      (typing G (affines G) e (t',{})) -> (* fix this to include a separate type-level phase *)
      (kinding G (typ_dep t e) K)
  
  | kinding_DROPPED :
    forall G t k,
      (kinding G t k) ->
      (kinding G (typ_dropped t) k)
  
  with tequiv : typ -> typ -> Prop := 
  | te_ID : forall t, tequiv  t t
    
  | te_SYM : forall  t t', tequiv  t' t -> tequiv  t t'
    
  | te_CTX_FUN_1 : 
    forall  tx tx' t, 
      tequiv  tx tx' -> 
      tequiv  (typ_fun tx t) (typ_fun tx' t)
      
  | te_CTX_FUN_2 : 
    forall  tx t t', 
      tequiv  t t' -> 
      tequiv  (typ_fun tx t) (typ_fun tx t')
  
  | te_CTX_UNIV :
    forall  k t t' ef, 
      tequiv  t t' -> 
      tequiv  (typ_univ k ef t) (typ_univ k ef t')
  
  | te_CTX_Q :
    forall  t t', 
      tequiv  t t' ->
      tequiv  (typ_qtype t) (typ_qtype t')
      
  | te_CTX_APP1 :
    forall  t1 t1' t2, 
      tequiv  t1 t1' -> 
      tequiv  (typ_app t1 t2) (typ_app t1' t2)
  
  | te_CTX_APP2 :
    forall  t1 t2 t2', 
      tequiv  t2 t2' -> 
      tequiv  (typ_app t1 t2) (typ_app t1 t2')
  
  | te_CTX_SING :
    forall  t t' n, 
      tequiv  t t' -> 
      tequiv  (typ_sing t n) (typ_sing t' n)
  
  | te_CTX_DEP_1:
    forall  t t' e, 
      tequiv  t t' -> 
      tequiv  (typ_dep t e) (typ_dep t' e)
  
  | te_RED:
    forall  t_e fx t e e', 
      typing nil {} e (t_e, fx) -> 
      typing nil {} e' (t_e, fx) -> 
      eval e e' ->
      tequiv  (typ_dep t e) (typ_dep t e')
  
  | te_CTX_DROPPED:
    forall  t t', 
      tequiv  t t' ->
      tequiv  (typ_dropped t) (typ_dropped t').

  Scheme typing_ind2 :=
    Induction for typing Sort Prop
  with kinding_ind2 :=
    Induction for kinding Sort Prop
  with tequiv_ind2 :=
    Induction for tequiv Sort Prop.
  
  Inductive wf_kind : kind -> Prop :=
  | wf_kind_simple : forall k, simple_kind k -> wf_kind k
  
  | wf_kind_tcon : forall k k', 
    simple_kind k -> wf_kind k' -> wf_kind (kind_tcon k k')
  
  | wf_kind_dcon : forall t tk k, 
    kinding nil t tk -> wf_kind k -> wf_kind (kind_dcon t k).

End LairAccessControlParams.

(* Simple lemma which uses [distinct] *)
Lemma Label_neq_User : Label <> User.
Proof.
  intro. distinct H.
Qed.

(* Lemma stating that [access] has the type we gave it. 
   The proof uses both [destructAtomEqs] and [distinct]. *)
Lemma access_has_right_type :
  typing Signature {} (exp_base access) ((typ_univ kind_U NO_FX
    (typ_fun (typ_dep (typ_const SLabel) (exp_base User))
      (typ_fun (typ_dep (typ_app (typ_const Labeled) (typ_const Int)) (exp_bvar 0))
        (typ_fun (typ_const Label)
          (typ_fun (typ_dep (typ_app (typ_const Labeled) (typ_btvar 0)) (exp_bvar 0))
            (typ_btvar 0)))))),{}).
Proof.
  constructor. compute ; destructAtomEqs ; try distinct e. auto.
Qed.

(* LairAccessControl is a module which instantiates all of the facts
   (proved abstractly in the other files) with the concrete signature
   and model given above. *)
Module LairAccessControl := LAIR_Subst LairAccessControlParams.

(* Here is a more complicated proof, showing that [Signature] is
   well-formed. *)
Lemma wf_sig : LairAccessControl.ModelFacts.wf_signature Signature.
Proof.
  unfold LairAccessControl.ModelFacts.wf_signature. split ; intro H.
    split ; unfold binds in H ; simpl in H ; destructAtomEqs ;
      pick fresh x ;
      try (discriminate ||
           inversion H ; repeat (simpl ; (econstructor || apply Fr))).
    destruct K ; [repeat constructor|repeat constructor|repeat constructor|apply wf_kind_tcon|eapply wf_kind_dcon].
      inversion H ; destructAtomEqs ; try discriminate ; inversion H1 ; constructor.
      inversion H ; destructAtomEqs ; try discriminate ; inversion H1 ; [apply wf_kind_tcon|eapply wf_kind_dcon] ; repeat constructor. compute ; destructAtomEqs ; try solve [distinct e0] ; auto.
      inversion H ; destructAtomEqs ; try discriminate ; inversion H1 ; constructor ; compute ; destructAtomEqs ; try solve [distinct e0] ; auto.
      inversion H ; destructAtomEqs ; try discriminate ; inversion H1.
        eapply wf_kind_dcon.
          constructor. compute ; destructAtomEqs ; try solve [distinct e0] ; auto.
          repeat constructor.
        repeat constructor.
Qed.
