Require Import Metatheory.

Set Undo 100000.

Inductive exp : Type :=
  | exp_bvar : nat -> exp
  | exp_fvar : atom -> exp
  | exp_base : atom -> exp
  | exp_lam  : typ -> exp -> exp
  | exp_app  : exp -> exp -> exp
  | exp_tab  : kind -> exp -> exp
  | exp_tap  : exp -> typ -> exp
(*   | exp_cap  : exp -> exp -> exp *)
(*   | exp_case : exp -> (pat * exp) -> exp -> exp *)
  | exp_new  : exp -> exp
  | exp_bot  : exp
  | exp_cert : exp -> (list exp_or_typ) -> exp
with typ : Type :=
  | typ_btvar  : nat -> typ
  | typ_ftvar  : atom -> typ
  | typ_const  : atom -> typ
  | typ_fun    : typ -> typ -> typ
  | typ_univ   : kind -> effect -> typ -> typ
(*   | typ_constr : typ -> typ -> typ *)
  | typ_qtype  : typ -> typ
  | typ_app    : typ -> typ -> typ
  | typ_dep    : typ -> exp -> typ
  | typ_sing   : typ -> typ -> typ
with kind : Type :=
  | kind_A    : kind 
  | kind_U    : kind
  | kind_N    : kind
  | kind_tcon : kind -> kind -> kind
  | kind_dcon : typ -> kind -> kind
with exp_or_typ : Type :=
  | eort_exp : exp -> exp_or_typ
  | eort_typ : typ -> exp_or_typ
with effect : Type :=
  | FX : effect
  | NO_FX : effect. 

Notation "[ x ]" := (x :: nil).

Fixpoint open_exp (k : nat) (f : exp) (e : exp) {struct e} : exp :=
  match e with
    | exp_bvar i => if k === i then f else (exp_bvar i)
    | exp_fvar x => exp_fvar x
    | exp_base x => exp_base x
    | exp_lam t e1 => exp_lam (open_typ k f t) (open_exp (S k) f e1)
    | exp_app e1 e2 => exp_app (open_exp k f e1) (open_exp k f e2)
    | exp_tab kk e1 => exp_tab kk (open_exp k f e1)
    | exp_tap e1 t => exp_tap (open_exp k f e1) (open_typ k f t)
    | exp_new e1 => exp_new (open_exp k f e1)
    | exp_bot => exp_bot
    | exp_cert e1 et_l => exp_cert e1 et_l (* Fix this *)
  end
with open_typ (k : nat) (f : exp) (t : typ) {struct t} : typ :=
  match t with 
    | typ_btvar i => typ_btvar i
    | typ_ftvar x => typ_ftvar x
    | typ_const x => typ_const x
    | typ_fun t1 t2 => typ_fun (open_typ k f t1) (open_typ (S k) f t2)
    | typ_univ kk ef t1 => typ_univ kk ef (open_typ k f t1)
    | typ_qtype t1 => typ_qtype (open_typ k f t1)
    | typ_app t1 t2 => typ_app (open_typ k f t1) (open_typ k f t2)
    | typ_dep t1 e2 => typ_dep (open_typ k f t1) (open_exp k f e2)
    | typ_sing t1 t2 => typ_sing (open_typ k f t1) (open_typ k f t2)
  end.


Fixpoint open_exp_typ (k : nat) (t : typ) (e : exp) {struct e} : exp :=
  match e with
    | exp_bvar i => exp_bvar i
    | exp_fvar x => exp_fvar x
    | exp_base x => exp_base x
    | exp_lam t1 e1 => exp_lam (open_typ_typ k t t1) (open_exp_typ k t e1)
    | exp_app e1 e2 => exp_app (open_exp_typ k t e1) (open_exp_typ k t e2)
    | exp_tab kk e1 => exp_tab kk (open_exp_typ (S k) t e1)
    | exp_tap e1 t2 => exp_tap (open_exp_typ k t e1) (open_typ_typ k t t2)
    | exp_new e1 => exp_new (open_exp_typ k t e1)
    | exp_bot => exp_bot
    | exp_cert e1 et_l => exp_cert e1 et_l (* Fix this *)
  end
with open_typ_typ (k : nat) (t : typ) (t' : typ) {struct t'} : typ :=
  match t' with 
    | typ_btvar i => if k === i then t else typ_btvar i
    | typ_ftvar x => typ_ftvar x
    | typ_const x => typ_const x
    | typ_fun t1 t2 => typ_fun (open_typ_typ k t t1) (open_typ_typ k t t2)
    | typ_univ kk ef t1 => typ_univ kk ef (open_typ_typ (S k) t t1)
    | typ_qtype t1 => typ_qtype (open_typ_typ k t t1)
    | typ_app t1 t2 => typ_app (open_typ_typ k t t1) (open_typ_typ k t t2)
    | typ_dep t1 e2 => typ_dep (open_typ_typ k t t1) (open_exp_typ k t e2)
    | typ_sing t1 t2 => typ_sing (open_typ_typ k t t1) (open_typ_typ k t t2)
  end.

Coercion typ_btvar : nat >-> typ.
Coercion typ_ftvar : atom >-> typ.
Coercion exp_bvar : nat >-> exp.
Coercion exp_fvar : atom >-> exp.
Notation domain := (list exp_or_typ).

(* Local closure for types and terms *)
Inductive lc_typ : typ -> Prop :=
| lc_typ_fvar  : forall a, lc_typ (typ_ftvar a)
| lc_typ_const : forall a, lc_typ (typ_const a)
| lc_typ_fun   : 
  forall L t1 t2 a, a `notin` L ->
    (lc_typ t1) -> 
    (lc_typ (open_typ O a t2)) -> 
    (lc_typ (typ_fun t1 t2))
| lc_typ_univ : 
  forall L t1 ef a k, a `notin` L ->
    (lc_typ (open_typ_typ 0 a t1)) -> 
    (lc_typ (typ_univ k ef t1))
| lc_typ_qtype :
  forall t1, (lc_typ t1) -> (lc_typ (typ_qtype t1))
| lc_typ_app : 
  forall t1 t2, (lc_typ t1) -> (lc_typ t2) -> (lc_typ (typ_app t1 t2))
| lc_typ_dep :
  forall t1 e1, (lc_typ t1) -> (lc_exp e1) -> (lc_typ (typ_dep t1 e1))
| lc_typ_sing :
  forall t1 t2, (lc_typ t1) -> (lc_typ t2) -> (lc_typ (typ_sing t1 t2))
with lc_exp : exp -> Prop :=
| lc_exp_fvar : forall x:atom, lc_exp (exp_fvar x)
| lc_exp_base : forall x:atom, lc_exp (exp_base x)
| lc_exp_lam  : 
  forall L t e a, a `notin` L ->
    (lc_typ t) -> (lc_exp (open_exp 0 a e)) -> (lc_exp (exp_lam t e))
| lc_exp_app  : 
  forall e1 e2, 
    (lc_exp e1) -> (lc_exp e2) -> (lc_exp (exp_app e1 e2))
| lc_exp_tab  :
  forall L k e a, a `notin` L -> 
    (lc_exp (open_exp_typ 0 a e)) -> (lc_exp (exp_tab k e))
| lc_exp_tap  : 
  forall e t, (lc_exp e) -> (lc_typ t) -> (lc_exp (exp_tap e t))
| lc_exp_new  : 
  forall e, (lc_exp e) -> (lc_exp (exp_new e))
| lc_exp_bot  : (lc_exp exp_bot)
| lc_exp_cert :
  forall e el,
    (lc_exp e) -> (lc_domain el) -> (lc_exp (exp_cert e el))
with lc_domain : domain -> Prop :=
| lc_domain_nil : lc_domain nil
| lc_domain_cons_exp : 
  forall e tl, (lc_exp e) -> (lc_domain tl) -> 
    (lc_domain ((eort_exp e)::tl))
| lc_domain_cons_typ : 
  forall t tl, (lc_typ t) -> (lc_domain tl) -> 
    (lc_domain ((eort_typ t)::tl)).
  
Inductive value : exp -> Prop :=
| value_abs : forall (t:typ) (e:exp),
  (lc_exp (exp_lam t e)) -> value (exp_lam t e)
| value_tab : forall (k:kind) (e:exp),
  (lc_exp (exp_tab k e)) -> value (exp_tab k e)
| value_new : forall (e:exp), 
  (value e) -> (value (exp_new e))
| value_cert : forall (a:atom) (d:domain), 
  (lc_exp (exp_base a)) -> (val_domain d) -> (value (exp_cert (exp_base a) d))
with val_domain : domain -> Prop :=
| val_domain_nil : (val_domain nil)
| val_domain_t : forall (t:typ) (tl:domain), 
  (lc_typ t) -> (val_domain tl) -> val_domain ((eort_typ t)::tl)
| val_domain_e : forall (e:exp) (tl:domain), 
  (lc_exp e) -> (val_domain tl) -> val_domain ((eort_exp e)::tl).

(* Dynamic semantics *)

Notation dom_eqn := (prod domain exp).
Parameter Model : (list (atom * (list dom_eqn))).
Inductive in_eqns  : domain -> list dom_eqn -> exp -> Prop := .
Inductive not_in_eqns : domain -> list dom_eqn -> Prop := .

Lemma classical_eqns : forall (d:domain) (e: list dom_eqn), 
  (not_in_eqns d e) \/ (exists e', in_eqns d e e').
Admitted.

Inductive eval : exp -> exp -> Prop :=

| eval_bot_bot : eval exp_bot exp_bot

| eval_app1 : forall e1 e1' e2, 
  eval e1 e1' -> eval (exp_app e1 e2) (exp_app e1' e2)
  
| eval_app1_bot : forall e1 e2, 
  eval e1 exp_bot -> eval (exp_app e1 e2) exp_bot
  
| eval_app2 : forall e1 e2 e2', 
  eval e2 e2' -> eval (exp_app e1 e2) (exp_app e1 e2')

| eval_app2_bot : forall e1 e2, 
  eval e2 exp_bot -> eval (exp_app e1 e2) exp_bot

| eval_beta : forall t e v, 
  (value v) -> (eval (exp_app (exp_lam t e) v) (open_exp 0 v e))

| eval_tap1 : forall e1 e1' t, 
  (eval e1 e1') -> (eval (exp_tap e1 t) (exp_tap e1' t))

| eval_tap1_bot : forall e1 t, 
  (eval e1 exp_bot) -> (eval (exp_tap e1 t) exp_bot)

| eval_tap : forall e1 k t, 
  (eval (exp_tap (exp_tab k e1) t) (open_exp_typ 0 t e1))

| eval_new1 : forall e1 e1', 
  (eval e1 e1') ->  (eval (exp_new e1) (exp_new e1'))

| eval_new1_bot : forall e1, 
  (eval e1 exp_bot) ->  (eval (exp_new e1) exp_bot)

| eval_delta : forall a eqns, 
  (binds a eqns Model) -> (eval (exp_base a) (exp_cert (exp_base a) nil))

| eval_b1 : forall b d v eqns v',
  (value (exp_cert (exp_base b) d)) -> (value v) -> 
  (binds b eqns Model) -> (in_eqns ((eort_exp v)::d) eqns v') -> 
  (eval (exp_app (exp_cert (exp_base b) d) v) v')

| eval_b2 : forall b d v eqns,
  (value (exp_cert (exp_base b) d)) -> (value v) -> 
  (binds b eqns Model) -> (not_in_eqns ((eort_exp v)::d) eqns) -> 
  (eval (exp_app (exp_cert (exp_base b) d) v)  (exp_cert (exp_base b) ((eort_exp v)::d)))

| eval_b3 : forall b d t eqns v',
  (value (exp_cert (exp_base b) d)) -> 
  (binds b eqns Model) -> (in_eqns ((eort_typ t)::d) eqns v') -> 
  (eval (exp_tap (exp_cert (exp_base b) d) t)  v')

| eval_b4 : forall b d t eqns,
  (value (exp_cert (exp_base b) d)) -> 
  (binds b eqns Model) -> (not_in_eqns ((eort_typ t)::d) eqns) -> 
  (eval (exp_tap (exp_cert (exp_base b) d) t)  (exp_cert (exp_base b) ((eort_typ t)::d))).

(* Static semantics *)

Inductive binding : Type :=
| binding_typ : typ -> binding
| binding_kind : kind -> binding.

(* Coercion binding_typ : typ >-> binding. *)
(* Coercion binding_kind : kind >-> binding.  *)

Notation gamma := (list (atom * binding)).
Notation env   := (prod gamma atoms).

Inductive simple_kind : kind -> Prop := 
| sk_A : simple_kind kind_A
| sk_U : simple_kind kind_U
| sk_N : simple_kind kind_N.

(* Relative closure for types and terms *)
Inductive rc_typ : typ -> gamma -> Prop :=
| rc_typ_bvar : forall n g, rc_typ (typ_btvar n) g

| rc_typ_fvar : forall a g k, 
  (binds a (binding_kind k) g) -> rc_typ (typ_ftvar a) g

| rc_typ_const : forall T g, 
  rc_typ (typ_const T) g
  
| rc_typ_fun : forall t1 t2 g, 
  (rc_typ t1 g) -> (rc_typ t2 g) -> rc_typ (typ_fun t1 t2) g
  
| rc_typ_univ : forall k t ef g, 
  (rc_typ t g) -> rc_typ (typ_univ k ef t) g

| rc_typ_qtype : forall t g, 
  rc_typ t g -> rc_typ (typ_qtype t) g

| rc_typ_app : forall t1 t2 g, 
  rc_typ t1 g -> rc_typ t2 g -> rc_typ (typ_app t1 t2) g

| rc_typ_dep : forall t e g, 
  rc_typ t g -> rc_exp e g -> rc_typ (typ_dep t e) g

| rc_typ_sing : forall t a g, 
  rc_typ t g -> rc_typ a g -> rc_typ (typ_sing t a) g

with rc_exp : exp -> gamma -> Prop :=
| rc_exp_bvar : forall n g, rc_exp (exp_bvar n) g

| rc_exp_fvar : forall x g t,
  (binds x (binding_typ t) g) -> rc_exp (exp_fvar x) g
  
| rc_exp_base : forall b g,
  rc_exp (exp_base b) g

| rc_exp_lam : forall t e g,
  (rc_typ t g) -> (rc_exp e g) -> rc_exp (exp_lam t e) g

| rc_exp_app : forall e1 e2 g, 
  (rc_exp e1 g) -> (rc_exp e2 g) -> rc_exp (exp_app e1 e2) g
  
| rc_exp_tab : forall k e g,
  (simple_kind k) -> (rc_exp e g) -> rc_exp (exp_tab k e) g

| rc_exp_tap  :  forall e t g, 
  (rc_exp e g) -> (rc_typ t g) -> (rc_exp (exp_tap e t) g)

| rc_exp_new  : forall e g, 
  (rc_exp e g) -> (rc_exp (exp_new e) g)

| rc_exp_bot  : forall g, 
  rc_exp exp_bot g

| rc_exp_cert : forall e el g,
    (rc_exp e g) -> (rc_domain el g) -> rc_exp (exp_cert e el) g

with rc_domain : domain -> gamma -> Prop :=
| rc_domain_nil : forall g, 
  rc_domain nil g

| rc_domain_cons_exp :  forall e tl g, 
  (rc_exp e g) -> (rc_domain tl g) -> rc_domain ((eort_exp e)::tl) g

| rc_domain_cons_typ :  forall t tl g, 
  (rc_typ t g) -> (rc_domain tl g) -> 
  (rc_domain ((eort_typ t)::tl) g).

Definition wf_typ (t:typ) (g:gamma) := lc_typ t /\  rc_typ t g.
Definition wf_exp (e:exp) (g:gamma) := lc_exp e /\ rc_exp e g.

Parameter Signature : gamma.
    
Inductive splitA : atoms -> atoms -> atoms -> Prop := 
| split_A : forall A1 A2, (A1 `inter` A2 = {}) -> splitA A1 A2 (A1 `union` A2).

Lemma split_nil_nil : forall A1 A2, splitA A1 A2 {} -> A1 = {} /\ A2 = {}.
Admitted. 

Lemma split_nil_A_1 : forall A, splitA {} A A.
Admitted. 

Lemma split_nil_A_2 : forall A, splitA A {} A.
Admitted. 

Lemma split_nil_eq_1 : forall A B, splitA {} A B -> A = B.
Admitted.

Lemma split_nil_eq_2 : forall A B, splitA A {} B -> A = B.
Admitted.

Lemma split_commutes : forall A B C, splitA A B C -> splitA B A C.
Admitted. 

(* Auxiliary functions a(x,A) and p(A) *)
Definition aqual (x:atom) (k:kind) (A:atoms) : atoms := 
  match k with 
    | kind_A => (singleton x) `union` A
    | _      => A
  end.

Definition ptype (t:typ) (A:atoms) (fx:atoms) : typ :=
  match AtomSet.F.is_empty (A `union` fx) with 
    | true => t
    | false => (typ_qtype t)
  end.

Definition affines (G:gamma) : atoms := {}. (* expand this *)

Definition effect_ok fx G :=  forall a, a `in` fx -> (binds a (binding_kind kind_N) G).

Inductive typing : gamma -> atoms -> exp -> (typ * atoms) -> Prop:=
| type_X : 
  forall G x t, 
    (binds x (binding_typ t) G) -> 
    (kinding G t kind_U) ->
    (typing G {} x (t, {}))

| type_XA :
  forall G x t,
    (binds x (binding_typ t) G) ->
    (kinding G t kind_A) ->
    (typing G (singleton x) x (t, {}))
    
| type_NEW :
  forall G A e t a fx a_fx,
    (typing G A e (t, fx)) ->
    (kinding G t kind_U) ->
    (binds a (binding_kind kind_N) G) ->
    (splitA (singleton a) fx a_fx) ->
    (typing G A (exp_new e) ((typ_sing (typ_qtype t) a), a_fx))

| type_ABS :
  forall L G A t_x k e t_e fx, 
    (forall x, x `notin` L ->
      (typing ([(x,binding_typ t_x)] ++ G) (aqual x k A) (open_exp 0 x e) ((open_typ 0 x t_e), fx))) ->
    (kinding G t_x k) ->
    (typing G A (exp_lam t_x e) ((ptype (typ_fun t_x t_e) A fx), fx))
    
| type_TAB :
  forall L G A k e t_e fx, 
    (forall a, a `notin` L -> a `notin` fx ->
      (typing ([(a,binding_kind k)] ++ G) A (open_exp_typ 0 a e) ((open_typ_typ 0 a t_e), fx))) ->
    (simple_kind k) ->
    (typing G A (exp_tab k e) ((ptype (typ_univ k NO_FX t_e) A fx), fx))
    
| type_TAB_FX :
  forall L G A k e t_e fx,
    (AtomSet.F.Subset fx L) ->
    (forall a, a `notin` L -> 
      (typing ([(a,binding_kind k)] ++ G) A (open_exp_typ 0 a e) ((open_typ_typ 0 a t_e), fx `union` singleton a))) ->
    (simple_kind k) ->
    (typing G A (exp_tab k e) ((ptype (typ_univ k FX t_e) A fx), fx))
    
| type_APP :
  forall G A A1 A2 e1 e2 t2 t fx1 fx2 fx, 
    (splitA A1 A2 A) ->
    (typing G A1 e1 ((typ_fun t2 t), fx1)) ->
    (typing G A2 e2 (t2, fx2)) -> 
    (splitA fx1 fx2 fx) ->
    (typing G A (exp_app e1 e2) ((open_typ 0 e2 t), fx))

| type_QAPP :
  forall G A A1 A2 e1 e2 t2 t fx1 fx2 fx, 
    (splitA A1 A2 A) ->
    (typing G A1 e1 (typ_qtype (typ_fun t2 t), fx1)) ->
    (typing G A2 e2 (t2, fx2)) -> 
    (splitA fx1 fx2 fx) ->
    (typing G A (exp_app e1 e2) ((open_typ 0 e2 t), fx))

| type_TAP_NOFX :
  forall G A e t k t' fx,
    (typing G A e ((typ_univ k NO_FX t'), fx)) ->
    (kinding G t k) ->
    (typing G A (exp_tap e t) ((open_typ_typ 0 t t'), fx))

| type_TAP_FX :
  forall G A e a t' fx' fx,
    (typing G A e ((typ_univ kind_N FX t'), fx)) ->
    (kinding G (typ_ftvar a) kind_N) ->
    (splitA (singleton a) fx fx') -> 
    (typing G A (exp_tap e (typ_ftvar a)) ((open_typ_typ 0 (typ_ftvar a) t'), fx'))

| type_QTAP_NOFX :
  forall G A e t k t' fx,
    (typing G A e (typ_qtype (typ_univ k NO_FX t'), fx)) ->
    (kinding G t k) ->
    (typing G A (exp_tap e t) ((open_typ_typ 0 t t'), fx))

| type_QTAP_FX :
  forall G A e a t' fx' fx,
    (typing G A e (typ_qtype (typ_univ kind_N FX t'), fx)) ->
    (kinding G (typ_ftvar a) kind_N) ->
    (splitA (singleton a) fx fx') -> 
    (typing G A (exp_tap e (typ_ftvar a)) ((open_typ_typ 0 (typ_ftvar a) t'), fx'))

| type_BOT :
  forall G t fx,
    (typing G {} exp_bot (t, fx))
    
| type_DROP :
  forall G A e t a fx, 
    (typing G A e ((typ_sing t a), fx)) -> 
    (typing G A e (t,fx))

| type_AFN :
  forall G A A1 A2 e t fx fx1 fx2, 
    (splitA A1 A2 A) ->
    (typing G A1 e (t, fx1)) -> 
    (splitA fx1 fx2 fx) -> 
    (effect_ok fx G) ->
    (typing G A e (t, fx))
    
| type_B :
  forall G B t,
    (binds B (binding_typ t) Signature) ->
    (typing G {} (exp_base B) (t, {}))

| type_B1 :
  forall G A B domain v t fx,
    (value v) -> 
    (typing G A (exp_app (exp_cert (exp_base B) domain) v) (t,fx)) -> 
    (typing G A (exp_cert (exp_base B) ((eort_exp v)::domain)) (t,fx))
    
| type_B2 :
  forall G A B domain t' t fx,
    (typing G A (exp_tap (exp_cert (exp_base B) domain) t') (t,fx)) -> 
    (typing G A (exp_cert (exp_base B) ((eort_typ t')::domain)) (t,fx))

| type_B3 :
  forall G A B t fx,
    (typing G A (exp_base B) (t, fx)) -> 
    (typing G A (exp_cert (exp_base B) nil) (t,fx))

| type_CONV :
  forall G A e t t' fx,
    (typing G A e (t,fx)) -> 
    (tequiv G (affines G) t t') ->
    (typing G A e (t',fx))
    
with kinding : gamma -> typ -> kind -> Prop :=

| kinding_A :
  forall G a k,
    (binds a (binding_kind k) G) ->
    (kinding G a k)
    
| kinding_N :
  forall G t a,
    (kinding G t kind_A) ->
    (kinding G a kind_N) ->
    (kinding G (typ_sing t a) kind_A)

| kinding_TC :
  forall G tc K,
    (binds tc (binding_kind K) Signature) ->
    (kinding G (typ_const tc) K)

| kinding_AFN :
  forall G t,
    (kinding G t kind_U) ->
    (kinding G (typ_qtype t) kind_A)

| kinding_ALL :
  forall L G k t k' a fx, a `notin` L ->
    (kinding ([(a,binding_kind k)] ++ G) (open_typ_typ 0 a t) k') ->
    (kinding G (typ_univ k fx t) kind_U)

| kinding_FUN :
  forall L G t_x t k_x k x, x `notin` L ->
    (kinding G t_x k_x) ->
    (kinding ([(x,binding_typ t_x)] ++ G) (open_typ 0 x t) k) ->
    (kinding G (typ_fun t_x t) kind_U)

| kinding_TAP :
  forall G t t' k K,
    (kinding G t (kind_tcon k K)) ->
    (kinding G t' k) ->
    (kinding G (typ_app t t') K)

| kinding_DEP :
  forall G t e t' K fx,
    (kinding G t (kind_dcon t' K)) ->
    (typing G (affines G) e (t',fx)) ->
    (kinding G (typ_dep t e) K)

with tequiv : gamma -> atoms -> typ -> typ -> Prop := 
| te_ID : forall g a t, tequiv g a t t
  
| te_SYM : forall g a t t', tequiv g a t' t -> tequiv g a t t'
  
| te_CTX_FUN_1 : 
  forall g a tx tx' t, 
    tequiv g a tx tx' -> 
    tequiv g a (typ_fun tx t) (typ_fun tx' t)
    
| te_CTX_FUN_2 : 
  forall g a tx t t', 
    tequiv g a t t' -> 
    tequiv g a (typ_fun tx t) (typ_fun tx t')

| te_CTX_UNIV :
  forall g a k t t' ef, 
    tequiv g a t t' -> 
    tequiv g a (typ_univ k ef t) (typ_univ k ef t')

| te_CTX_Q :
  forall g a t t', 
    tequiv g a t t' ->
    tequiv g a (typ_qtype t) (typ_qtype t')
    
| te_CTX_APP1 :
  forall g a t1 t1' t2, 
    tequiv g a t1 t1' -> 
    tequiv g a (typ_app t1 t2) (typ_app t1' t2)

| te_CTX_APP2 :
  forall g a t1 t2 t2', 
    tequiv g a t2 t2' -> 
    tequiv g a (typ_app t1 t2) (typ_app t1 t2')

| te_CTX_SING :
  forall g a t t' n, 
    tequiv g a t t' -> 
    tequiv g a (typ_sing t n) (typ_sing t' n)

| te_CTX_DEP_1:
  forall g a t t' e, 
    tequiv g a t t' -> 
    tequiv g a (typ_dep t e) (typ_dep t' e)

| te_RED:
  forall g a t_e fx t e e', 
    typing g a e (t_e, fx) -> 
    typing g a e' (t_e, fx) -> 
    eval e e' -> 
    tequiv g a (typ_dep t e) (typ_dep t e').


(* Scheme typing_ind2 := *)
(*   Induction for typing Sort Prop *)
(* with kinding_ind2 := *)
(*   Induction for kinding Sort Prop. *)

Notation model := (list (atom * (list dom_eqn))).

Inductive MS_consistent : model -> gamma -> Prop :=
| MS_base : forall (M:model), MS_consistent M nil
| MS_cons : 
  forall (M:model) (S:gamma) (a:atom) de (t:typ), 
    MS_consistent M S -> (binds a de M) -> 
    MS_consistent M ([(a, binding_typ t)] ++ S).

Lemma MS_binding : forall (m:model) (s:gamma) (a:atom) (t:typ),
  MS_consistent m s -> (binds a (binding_typ t) s) -> (exists de, binds a de m).
Admitted.

(** An environment G,A is well-formed, denoted [(wf_env G A)], if each
    atom is bound at most at once and if each binding is to a
    well-formed type. *)

Inductive wf_env : gamma -> atoms -> Prop :=
| wf_env_empty :
  wf_env nil {}

| wf_env_typ : 
  forall (G : gamma) (A : atoms) (X : atom) (T : typ),
    wf_env G A ->
(*     (kinding G T K) ->  *)
(*     (simple_kind K) ->  *)
    X `notin` dom G ->
    wf_typ T G ->
    wf_env ([(X, binding_typ T)] ++ G) A
    
| wf_env_typ_A : 
  forall (G : gamma) (A : atoms) (X : atom) (T : typ),
    wf_env G A ->
(*     splitA ((singleton X)) A A' ->  *)
(*     (kinding G T kind_A) ->  *)
    X `notin` dom G ->
    wf_typ T G -> 
    wf_env ([(X, binding_typ T)] ++ G) ((singleton X) `union` A)

| wf_env_kind : 
  forall (G : gamma) (A : atoms) (X : atom) (K : kind),
    wf_env G A ->
    X `notin` dom G ->
    wf_env ([(X, binding_kind K)] ++ G) A.

Lemma wf_is_ok : forall g A, 
  wf_env g A -> ok g.
Admitted.

Inductive only_names : gamma -> atoms -> Prop :=
| only_names_empty :  
   only_names nil {}

| only_names_extra_in_G : forall G X, 
  only_names G {} -> 
  only_names (([(X, binding_kind kind_N)])++G) {} 
   
| only_names_cons :
  forall G A A' (X:atom), 
    only_names G A -> 
    splitA (singleton X) A A' ->
    only_names ([(X, binding_kind kind_N)] ++ G) A'.

Ltac case_eq H := generalize (refl_equal H); pattern H at -1; case H.

Lemma binds_no_vars : forall (G:gamma) (A:atoms) (x:atom) (t:typ), 
  only_names G A -> not (binds x (binding_typ t) G).
  intros G A x t ON. induction ON.
  Case "empty".
  unfold not. intros. inversion H.
  Case "extra".
  unfold not. intros binds_x_XG.
  destruct (binds_concat_inv binding x (binding_typ t) G ([(X, binding_kind kind_N)]) binds_x_XG) as [ bindsG | bindsHD ].
  destruct bindsG as [xNotinHd binds_x_F]. auto.
  inversion bindsHD.
  destruct (x==X); discriminate.
  Case "cons".
  unfold not. intros binds_x_XG.
  destruct (binds_concat_inv binding x (binding_typ t) G ([(X, binding_kind kind_N)]) binds_x_XG) as [ bindsG | bindsHD ].
  destruct bindsG as [xNotinHd binds_x_F]. auto.
  inversion bindsHD.
  destruct (x == X); discriminate.
Qed.

Lemma typing_lc : forall (G:gamma) (A:atoms) (e:exp) (t:typ) (fx:atoms),
  wf_env G A -> typing G A e (t,fx) -> lc_exp e.
Admitted.

Lemma only_names_split1 : forall (G:gamma) (A1:atoms) (A2:atoms) (A:atoms), 
  only_names G A -> splitA A1 A2 A -> only_names G A1.
Admitted.

Lemma only_names_split2 : forall (G:gamma) (A1:atoms) (A2:atoms) (A:atoms), 
  only_names G A -> splitA A1 A2 A -> only_names G A2.
Admitted.

Lemma wf_split1 : forall (G:gamma) (A1:atoms) (A2:atoms) (A:atoms), 
  wf_env G A -> splitA A1 A2 A -> wf_env G A1.
Admitted.

Lemma wf_split2 : forall (G:gamma) (A1:atoms) (A2:atoms) (A:atoms), 
  wf_env G A -> splitA A1 A2 A -> wf_env G A2.
Admitted.

Lemma wf_cat : forall (G1 : gamma) (G2 : gamma) A A1 A2, 
  wf_env G1 A -> splitA A A1 A2 ->  wf_env (G2 ++ G1) A2 ->
  wf_env (G2 ++ G1) A.
Admitted.

Lemma splitA_empty1 : forall (A1:atoms) (A2:atoms), 
  (splitA A1 A2 {}) -> A1 = {}.
Admitted.

Lemma splitA_empty2 : forall (A1:atoms) (A2:atoms), 
  (splitA A1 A2 {}) -> A2 = {}.
Admitted.

Lemma inversion_lam : forall (G:gamma) (A:atoms) (e:exp) (tx:typ) (t:typ) fx, 
  value e -> typing G A e ((typ_fun tx t),fx) -> 
  (exists e', e = (exp_lam tx e')) \/ (exists B, exists d, e = (exp_cert (exp_base B) d)).
Admitted.

(* Basic strategy for inversion_lam *)
(* rewrite <- H6 in T1. inversion T1. cut (A1 = {}). intro. subst. *)
(* cut ((ptype (typ_univ k fx3 t_e) {}) = (typ_univ k fx3 t_e)). *)
(* intro. rewrite H6 in H9. discriminate. apply ptype_empty.  *)
(* eapply splitA_empty1. subst. apply H0. *)
(* inversion H11.  *)
(* cut ((ptype (typ_univ k fx3 t_e) A1) = (typ_univ k fx3 t_e)). *)
(* intro Pt. rewrite Pt in H9. discriminate. *)
(* eapply ptype_empty. eapply splitA_empty1. rewrite HeqA in H0. apply H0. *)
(* inversion H11. *)
(* cut (ptype (typ_univ k fx4 t_e) A1 = typ_sing (typ_fun t2 fx' t0) a). *)
(* intro Pt. rewrite Pt in H15.  *)

(* unfold ptype in H9. case_eq (AtomSet.F.is_empty {}). *)
(* intro. rewrite H6 in H9. discriminate. intro. rewrite H6 in H9. discriminate. *)
(* case_eq (AtomSet.F.is_empty {}). *)
(* intros. eapply splitA_empty1. subst. apply H0. *)
(* intros. eapply splitA_empty1. subst. apply H0. *)
(* inversion H11. *)

Lemma inversion_qlam : forall (G:gamma) (A:atoms) (e:exp) (tx:typ) (t:typ) fx,
  value e -> typing G A e ((typ_qtype (typ_fun tx t), fx)) -> 
  (exists e', e = (exp_lam tx e')) \/ (exists B, exists d, e = (exp_cert (exp_base B) d)).
Admitted.

Lemma inversion_tab : forall (G:gamma) (A:atoms) (e:exp) (k:kind) (t:typ) ef fx,
  value e -> typing G A e ((typ_univ k ef t),fx) -> 
  (exists e', e = (exp_tab k e')) \/ (exists B, exists d, e = (exp_cert (exp_base B) d)).
Admitted.

Lemma inversion_qtab : forall (G:gamma) (A:atoms) (e:exp) (k:kind) (t:typ) ef fx, 
  value e -> typing G A e ((typ_qtype (typ_univ k ef t)), fx) -> 
  (exists e', e = (exp_tab k e')) \/ (exists B, exists d, e = (exp_cert (exp_base B) d)).
Admitted.  

Lemma cert_is_value : forall (G:gamma) (A:atoms) (B:atom) (d:domain) (t:typ) fx, 
  wf_env G A -> only_names G A -> typing G A (exp_cert (exp_base B) d) (t,fx) -> 
  value (exp_cert (exp_base B) d).
Admitted.

Lemma sig_binds_certB : forall (G:gamma) (A:atoms) (B:atom) (d:domain) (t:typ) fx, 
  typing G A (exp_cert (exp_base B) d) (t,fx) -> (exists t, binds B (binding_typ t) Signature).
Admitted.

Set Hyps Limit 20.

Tactic Notation "decompose" "records" :=
  repeat (
    match goal with
      | H: _ |- _ => progress (decompose record H); clear H
    end).

Theorem progress : forall (G:gamma) (A:atoms) (e:exp) (t:typ) (fx:atoms),
  MS_consistent Model Signature -> wf_env G A -> only_names G A -> 
  typing G A e (t,fx) -> 
  (exists e', eval e e') \/ (value e) \/ (e = exp_bot).
intros G A e t fx MS WF ON T.
assert (typing G A e (t,fx)); auto.
remember (t, fx) as tfx.
generalize dependent t.
generalize dependent fx.
induction T.
Case "typing_X".
intros.
assert (~ binds x (binding_typ t) G) as NB.
  eapply binds_no_vars. apply ON.
elim NB. assumption.
Case "typing_XA".
intros.
assert (~ binds x (binding_typ t) G) as NB.
  eapply binds_no_vars; apply ON.
elim NB. assumption.
Case "typing_NEW".
(* pose proof (only_names_split2 G (singleton a) A A' ON H2) as ONA. *)
destruct (IHT WF ON T fx t (refl_equal (t, fx))) as [Esteps | [Evalue | Ebot]]. auto.
  (* Sub-Case Esteps *)
  intros.
  elim Esteps. intros x Ex. left. exists (exp_new x).
  apply eval_new1. assumption.
  (* Sub-case Evalue *)
  intros.
  right. left. apply value_new. assumption.
  (* Sub-case Ebot *)
  intros.
  left. exists exp_bot. rewrite Ebot. apply eval_new1_bot. apply eval_bot_bot.
Case "typing_ABS".
intros.
right. left. apply value_abs.
eapply typing_lc. apply WF. apply H.
Case "typing_TAB".
intros.
right. left. apply value_tab.
eapply typing_lc. apply WF. apply H.
Case "typing_TAB_FX".
intros.
right. left. apply value_tab.
eapply typing_lc. apply WF. apply H.
Case "typing_APP".
intros.
left.
pose proof (wf_split1 G A1 A2 A WF H0) as wfGA1.
pose proof (wf_split2 G A1 A2 A WF H0) as wfGA2.
pose proof (only_names_split1 G A1 A2 A ON H0) as ONA1.
pose proof (only_names_split2 G A1 A2 A ON H0) as ONA2.
pose proof (refl_equal (typ_fun t2 t, fx1)) as tfx_eq.
destruct (IHT1 wfGA1 ONA1 T1 _ _ tfx_eq) as [E1steps | [E1val | E1bot]]; auto. 
(* E1steps *)
elim E1steps. intros x E1x.
exists (exp_app x e2). apply eval_app1. assumption.
(* E1val *)
destruct (IHT2 wfGA2 ONA2 T2 _ _ (refl_equal (t2, fx2))) as [E2steps | [E2val | E2bot]]; auto.
   (* E2steps *)
    elim E2steps. intros x E2x. 
    exists (exp_app e1 x). apply eval_app2. assumption.
   (* E2val *)
   destruct (inversion_lam G A1 e1 _ _ _  E1val T1) as [E1isLAM | E1isCERT].
     (* E1 is lam *)
     elim E1isLAM. intros x E1isLAMx.
     exists (open_exp 0 e2 x). rewrite E1isLAMx. apply eval_beta. assumption.
     (* E1 is cert *)
     destruct E1isCERT as [B [D E1isBD]].
     rewrite E1isBD. subst.   
     destruct (sig_binds_certB _ _  _ _ _ _ T1) as [ Bt SigBindsB ].
     destruct (MS_binding Model Signature B Bt MS SigBindsB) as [De ModBindsB].
     destruct (classical_eqns ((eort_exp e2)::D) De) as [ NotInEqns | InEqns ].
     exists (exp_cert (exp_base B) ((eort_exp e2)::D)).
     eapply eval_b2.
     eapply (cert_is_value G A1 B D (typ_fun t2 t)); auto. apply T1.
     apply E2val. apply ModBindsB. apply NotInEqns.
     destruct InEqns as [X EqReducesX].
     exists X. eapply eval_b1.
     eapply (cert_is_value G A1 B D (typ_fun t2 t)); auto. apply T1.
     apply E2val. apply ModBindsB. apply EqReducesX. 
   (* E2bot *)
   exists exp_bot. eapply eval_app2_bot. rewrite E2bot; apply eval_bot_bot.
(* E1bot *)
exists exp_bot. eapply eval_app1_bot. rewrite E1bot; apply eval_bot_bot.
Case "typing_QAPP".
intros.
left.
pose proof (wf_split1 G A1 A2 A WF H0) as wfGA1.
pose proof (wf_split2 G A1 A2 A WF H0) as wfGA2.
pose proof (only_names_split1 G A1 A2 A ON H0) as ONA1.
pose proof (only_names_split2 G A1 A2 A ON H0) as ONA2.
pose proof (refl_equal (typ_qtype (typ_fun t2 t), fx1)) as tfx_eq.
destruct (IHT1 wfGA1 ONA1 T1 _ _ tfx_eq) as [E1steps | [E1val | E1bot]]; auto. 
(* E1steps *)
elim E1steps. intros x E1x.
exists (exp_app x e2). apply eval_app1. assumption.
(* E1val *)
destruct (IHT2 wfGA2 ONA2 T2 _ _ (refl_equal (t2, fx2))) as [E2steps | [E2val | E2bot]]; auto.
   (* E2steps *)
    elim E2steps. intros x E2x. 
    exists (exp_app e1 x). apply eval_app2. assumption.
   (* E2val *)
   destruct (inversion_qlam G A1 e1 _ _ _  E1val T1) as [E1isLAM | E1isCERT].
     (* E1 is lam *)
     elim E1isLAM. intros x E1isLAMx.
     exists (open_exp 0 e2 x). rewrite E1isLAMx. apply eval_beta. assumption.
     (* E1 is cert *)
     destruct E1isCERT as [B [D E1isBD]].
     rewrite E1isBD. subst.   
     destruct (sig_binds_certB _ _  _ _ _ _ T1) as [ Bt SigBindsB ].
     destruct (MS_binding Model Signature B Bt MS SigBindsB) as [De ModBindsB].
     destruct (classical_eqns ((eort_exp e2)::D) De) as [ NotInEqns | InEqns ].
     exists (exp_cert (exp_base B) ((eort_exp e2)::D)).
     eapply eval_b2.
     eapply (cert_is_value G A1 B D (typ_qtype (typ_fun t2 t))); auto. apply T1.
     apply E2val. apply ModBindsB. apply NotInEqns.
     destruct InEqns as [X EqReducesX].
     exists X. eapply eval_b1.
     eapply (cert_is_value G A1 B D (typ_qtype (typ_fun t2 t))); auto. apply T1.
     apply E2val. apply ModBindsB. apply EqReducesX. 
   (* E2bot *)
   exists exp_bot. eapply eval_app2_bot. rewrite E2bot; apply eval_bot_bot.
(* E1bot *)
exists exp_bot. eapply eval_app1_bot. rewrite E1bot; apply eval_bot_bot.
Case "typing_TAP".
intros.
left. 
destruct (IHT WF ON T _ _ (refl_equal ((typ_univ k NO_FX t'), fx))) as [Esteps | [Eval | Ebot]]; auto.
(* Esteps *)
destruct Esteps as [x Ex].
exists (exp_tap x t).
apply eval_tap1. assumption.
(* Eval *)
destruct (inversion_tab G A e _ _ _ _ Eval T) as [EisTAB | EisCERT].
  (* EisTAB *)
  destruct EisTAB as [x Etabx].
  exists (open_exp_typ 0 t x).
  rewrite Etabx. apply eval_tap.
  (* E is cert *)
  destruct EisCERT as [ B [D EisBD]].
  subst. 
  destruct (sig_binds_certB _  _ _ _ _ _ T) as [Bt SigBindsB].
  destruct (MS_binding Model Signature B Bt MS SigBindsB) as [De ModBindsB].
  destruct (classical_eqns ((eort_typ t)::D) De) as [ NotInEqns | InEqns ].
  exists (exp_cert (exp_base B) ((eort_typ t)::D)).
  eapply eval_b4.
  eapply cert_is_value. apply WF. apply ON. apply T. apply ModBindsB. apply NotInEqns.
  destruct InEqns as [X EqReducesX].
  exists X.
  eapply eval_b3.
  eapply cert_is_value. apply WF. apply ON. apply T. apply ModBindsB. apply EqReducesX.
(* Ebot *)
exists (exp_bot). rewrite Ebot. eapply eval_tap1_bot. apply eval_bot_bot.
Case "typing_TAP_FX".
intros.
left. 
destruct (IHT WF ON T _ _ (refl_equal ((typ_univ kind_N FX t'), fx))) as [Esteps | [Eval | Ebot]]; auto.
(* Esteps *)
destruct Esteps as [x Ex].
exists (exp_tap x a).
apply eval_tap1. assumption.
(* Eval *)
destruct (inversion_tab G A e _ _ _ _ Eval T) as [EisTAB | EisCERT].
  (* EisTAB *)
  destruct EisTAB as [x Etabx].
  exists (open_exp_typ 0 a x).
  rewrite Etabx. apply eval_tap.
  (* E is cert *)
  destruct EisCERT as [ B [D EisBD]].
  subst. 
  destruct (sig_binds_certB _  _ _ _ _ _ T) as [Bt SigBindsB].
  destruct (MS_binding Model Signature B Bt MS SigBindsB) as [De ModBindsB].
  destruct (classical_eqns ((eort_typ a)::D) De) as [ NotInEqns | InEqns ].
  exists (exp_cert (exp_base B) ((eort_typ a)::D)).
  eapply eval_b4.
  eapply cert_is_value. apply WF. apply ON. apply T. apply ModBindsB. apply NotInEqns.
  destruct InEqns as [X EqReducesX].
  exists X.
  eapply eval_b3.
  eapply cert_is_value. apply WF. apply ON. apply T. apply ModBindsB. apply EqReducesX.
(* Ebot *)
exists (exp_bot). rewrite Ebot. eapply eval_tap1_bot. apply eval_bot_bot.
Case "typing_QTAP".
intros.
left. 
destruct (IHT WF ON T _ _ (refl_equal (typ_qtype (typ_univ k NO_FX t'), fx))) as [Esteps | [Eval | Ebot]]; auto.
(* Esteps *)
destruct Esteps as [x Ex].
exists (exp_tap x t).
apply eval_tap1. assumption.
(* Eval *)
destruct (inversion_qtab G A e _ _ _ _ Eval T) as [EisTAB | EisCERT].
  (* EisTAB *)
  destruct EisTAB as [x Etabx].
  exists (open_exp_typ 0 t x).
  rewrite Etabx. apply eval_tap.
  (* E is cert *)
  destruct EisCERT as [ B [D EisBD]].
  subst. 
  destruct (sig_binds_certB _  _ _ _ _ _ T) as [Bt SigBindsB].
  destruct (MS_binding Model Signature B Bt MS SigBindsB) as [De ModBindsB].
  destruct (classical_eqns ((eort_typ t)::D) De) as [ NotInEqns | InEqns ].
  exists (exp_cert (exp_base B) ((eort_typ t)::D)).
  eapply eval_b4.
  eapply cert_is_value. apply WF. apply ON. apply T. apply ModBindsB. apply NotInEqns.
  destruct InEqns as [X EqReducesX].
  exists X.
  eapply eval_b3.
  eapply cert_is_value. apply WF. apply ON. apply T. apply ModBindsB. apply EqReducesX.
(* Ebot *)
exists (exp_bot). rewrite Ebot. eapply eval_tap1_bot. apply eval_bot_bot.
Case "typing_QTAP_FX".
intros.
left. 
destruct (IHT WF ON T _ _ (refl_equal (typ_qtype (typ_univ kind_N FX t'), fx))) as [Esteps | [Eval | Ebot]]; auto.
(* Esteps *)
destruct Esteps as [x Ex].
exists (exp_tap x a).
apply eval_tap1. assumption.
(* Eval *)
destruct (inversion_qtab G A e _ _ _ _ Eval T) as [EisTAB | EisCERT].
  (* EisTAB *)
  destruct EisTAB as [x Etabx].
  exists (open_exp_typ 0 a x).
  rewrite Etabx. apply eval_tap.
  (* E is cert *)
  destruct EisCERT as [ B [D EisBD]].
  subst. 
  destruct (sig_binds_certB _  _ _ _ _ _ T) as [Bt SigBindsB].
  destruct (MS_binding Model Signature B Bt MS SigBindsB) as [De ModBindsB].
  destruct (classical_eqns ((eort_typ a)::D) De) as [ NotInEqns | InEqns ].
  exists (exp_cert (exp_base B) ((eort_typ a)::D)).
  eapply eval_b4.
  eapply cert_is_value. apply WF. apply ON. apply T. apply ModBindsB. apply NotInEqns.
  destruct InEqns as [X EqReducesX].
  exists X.
  eapply eval_b3.
  eapply cert_is_value. apply WF. apply ON. apply T. apply ModBindsB. apply EqReducesX.
(* Ebot *)
exists (exp_bot). rewrite Ebot. eapply eval_tap1_bot. apply eval_bot_bot.
Case "typing_BOT".
intros.
right; right. auto.
Case "typing_DROP".
intros.
eapply IHT; auto.
Case "typing_AFN".
intros.
pose proof (only_names_split1 G A1 A2 A ON H0) as ONA1.
pose proof (wf_split1 G A1 A2 A WF H0) as wfGA1.
eapply (IHT wfGA1 ONA1 T fx1 t (refl_equal (t, fx1))).
Case "typing_B".
intros.
left. inversion MS.
(* subcase signature empty *)
subst. rewrite <- H3 in H0. inversion H0.
(* subcase steps *)
assert (exists de, (binds B de Model)) as DE.
eapply MS_binding. apply MS. apply H0.
elim DE. intros x MbindsX.
exists (exp_cert (exp_base B) nil).
eapply eval_delta. apply MbindsX.
Case "typing_B1".
right; left.
eapply cert_is_value. subst. apply WF. apply ON. subst. apply H.
Case "typing_B2".
right; left.
eapply cert_is_value. subst. apply WF. apply ON. subst. apply H.
Case "typing_B3".
right; left.
eapply value_cert. eapply typing_lc. apply WF. apply T. constructor.
Case "typing_CONV".
intros.
eapply (IHT WF ON T _ _ (refl_equal (t, fx))). 
Qed.


(* ******************************************************************************** *)
(* Subject reduction *)
(* ******************************************************************************** *)

Fixpoint subst_exp (x:atom) (ex:exp) (e:exp) {struct e} : exp :=
  match e with
    | exp_bvar i => exp_bvar i
    | exp_fvar y => if x == y then ex else exp_fvar y
    | exp_base b => exp_base b
    | exp_lam t e1 => exp_lam (subst_typ x ex t) (subst_exp x ex e1)
    | exp_app e1 e2 => exp_app (subst_exp x ex e1) (subst_exp x ex e2)
    | exp_tab kk e1 => exp_tab kk (subst_exp x ex e1)
    | exp_tap e1 t => exp_tap (subst_exp x ex e1) (subst_typ x ex t)
    | exp_new e1 => exp_new (subst_exp x ex e1)
    | exp_bot => exp_bot
    | exp_cert e1 et_l => exp_cert e1 et_l (* These are values ... ie. closed exps *)
  end
with subst_typ (x : atom) (ex : exp) (t : typ) {struct t} : typ :=
  match t with 
    | typ_btvar i => typ_btvar i
    | typ_ftvar tv => typ_ftvar tv
    | typ_const c => typ_const c
    | typ_fun t1 t2 => typ_fun (subst_typ x ex t1)  (subst_typ x ex t2) (* fx is only names *)
    | typ_univ kk ef t1 => typ_univ kk ef (subst_typ x ex t1)
    | typ_qtype t1 => typ_qtype (subst_typ x ex t1)
    | typ_app t1 t2 => typ_app (subst_typ x ex t1) (subst_typ x ex t2)
    | typ_dep t1 e2 => typ_dep (subst_typ x ex t1) (subst_exp x ex e2)
    | typ_sing t1 t2 => typ_sing (subst_typ x ex t1) (subst_typ x ex t2)
  end.

Definition subst_binding (x:atom) (ex:exp) (ab: atom * binding) : (atom * binding) :=
  match ab with 
    | (a, binding_typ t) => (a, binding_typ (subst_typ x ex t))
    | (a, binding_kind k) =>  (a, binding_kind k)
  end.

Lemma subst_exp_open_exp_commute : forall x y vx e, 
  x <> y -> lc_exp vx -> 
  (subst_exp x vx (open_exp 0 y e)) = (open_exp 0 y (subst_exp x vx e)).
Admitted.

Lemma subst_typ_open_typ_commute : forall x y vx t, 
  x <> y -> lc_exp vx -> 
  (subst_typ x vx (open_typ 0 y t)) = (open_typ 0 y (subst_typ x vx t)).
Admitted.

Lemma subst_typ_open_typ_typ_commute : forall x b v t, 
  x <> b -> lc_exp v -> 
  (subst_typ x v (open_typ_typ 0 b t) = open_typ_typ 0 b (subst_typ x v t)).
Admitted.

Lemma subst_exp_open_exp_typ_commute : forall x b v e, 
  x <> b -> lc_exp v -> 
  (subst_exp x v (open_exp_typ 0 b e) = open_exp_typ 0 b (subst_exp x v e)).
Admitted.

Check List.map.
Check gamma.

Definition subst_gamma (x:atom) (ex:exp) (G:gamma) : gamma :=
  List.map (subst_binding x ex) G.

Scheme exp_ind2 :=
  Induction for exp Sort Prop
with typ_ind2 :=
  Induction for typ Sort Prop
with kind_ind2 :=
  Induction for kind Sort Prop
with exp_or_typ_ind2 :=
  Induction for exp_or_typ Sort Prop.

Scheme typing_ind2 :=
  Induction for typing Sort Prop
with kinding_ind2 :=
  Induction for kinding Sort Prop.

Check exp_ind2.
Check exp_ind.
Check typing_ind2.

Lemma subst_gamma_dom : forall G x v, 
  dom G = dom (subst_gamma x v G).
Admitted.

Lemma subst_wf_exp : forall G e x v, 
  wf_exp e G -> wf_exp v G -> wf_exp (subst_exp x v e) G.
Admitted.

Lemma subst_wf_typ : forall G1 G2 t x tx v A,
  wf_env (G2 ++ [(x, binding_typ tx)] ++ G1) A -> 
  wf_typ t (G2 ++ [(x, binding_typ tx)] ++ G1) ->
  wf_exp v G1 -> wf_typ (subst_typ x v t) ((subst_gamma x v G2) ++ G1).
Admitted.

Lemma wf_exp_weakening : forall G1 G2 v, 
  wf_exp v G1 -> wf_exp v (G2 ++ G1).
Admitted.

Lemma wf_typ_weakening : forall G1 G2 t, 
  wf_typ t G1 -> wf_typ t (G2 ++ G1).
Admitted.

Unset Hyps Limit.

Lemma aqual_cases : forall A x k, 
  (aqual x k A) = A  \/ (aqual x k A) = (singleton x `union` A).
Admitted.

Lemma wf_env_G_A_containment : forall G A, 
  wf_env G A -> AtomSet.F.Subset A (dom G).
intros.
generalize dependent A.
induction G.
Case "Gnil".
  intros.
  inversion H. simpl. fsetdec.
Case "Gcons".
  intros.
  inversion H.
  SCase "wf_env_typ".
    simpl.
    pose proof (IHG A H2) as IH. fsetdec.
  SCase "wf_env_typ_A".
    simpl.
    pose proof (IHG A0 H2) as IH. fsetdec.
  SCase "wf_env_kind".
    simpl.
    pose proof (IHG A H2) as IH. fsetdec.
Qed.

Lemma union_identity : forall x A B, 
  (singleton x `union` A) = (singleton x `union` B) -> x `notin` A -> x `notin` B -> A = B.
Admitted. 

Lemma aqual_identity : forall X x k A, 
  X `in` aqual x k A -> (X = x \/ X `in` A).
Admitted.

Lemma aqual_identity_0 : forall X x k A, 
  X `in` A -> X `in` (aqual x k A).
Admitted.

Lemma aqual_union_identity_1 : forall X A0 x k A, 
  (singleton X `union` A0) = (aqual x k A) -> (X `notin` A0) -> (x `notin` A) -> 
  exists A1, (A0 = aqual x k A1 /\ x `notin` A1).
Admitted.

Lemma aqual_union_identity_2 : forall X x k A1 A, 
  (singleton X `union` A1 = aqual x k A) -> 
  (x <> X) ->
  exists A2, (A = singleton X `union` A2 /\ X `notin` A2).
Admitted.

Lemma aqual_identity_3 : forall X x k A A1 A2, 
  X `notin` A1 -> X `notin` A2 -> x `notin` A1 -> x `notin` A -> 
  (singleton X `union` aqual x k A1) = aqual x k A -> 
  (singleton X `union` A2) = A -> A1 = A2.
Admitted.
  
Lemma subst_gamma_wf : forall G1 G2 A x t_x v k, 
  wf_env (G2 ++ [(x, binding_typ t_x)] ++ G1) (aqual x k A) -> 
  wf_exp v G1 -> (x `notin` A) ->
  wf_env ((subst_gamma x v G2) ++ G1) A.
  intros G1 G2 A x t_x v k wf_G2_x_G1_Ax wf_v x_notin_A.
  generalize dependent A.
  induction G2.
  Case "G2nil".
  intros.
  simpl.
  simpl_env in wf_G2_x_G1_Ax.
  destruct (aqual_cases A x k) as [Ax_is_A | Ax ].
     SCase "Ax_is_A".
     rewrite Ax_is_A in wf_G2_x_G1_Ax.
     inversion wf_G2_x_G1_Ax. 
       (* case wf_env_typ *) assumption.
       (* case wf_env_typ_A *) rewrite <- H3 in x_notin_A. fsetdec.
     SCase "Ax".
     rewrite Ax in wf_G2_x_G1_Ax.
     inversion wf_G2_x_G1_Ax.
     (* Case wf_env_typ *)
       pose proof (wf_env_G_A_containment G1 (singleton x `union` A) H2) as for_contra. 
       destruct H4. unfold AtomSet.F.Subset in for_contra. 
       eapply for_contra. fsetdec.
     (* Case wf_env_typ_A *)
       pose proof (wf_env_G_A_containment G1 A0 H3) as A0_sub_G1.
       assert (x `notin` A0) as x_notin_A0.
       unfold AtomSet.F.Subset in A0_sub_G1.
       unfold not. intro. 
       pose proof (A0_sub_G1 x H6) as for_contra. contradiction.
       assert (A0 = A) as A0_eq_A. eapply union_identity. apply H2. auto. auto.
       subst A0. auto.
  Case "G2cons".
  intros.
  simpl.
  inversion wf_G2_x_G1_Ax.
       SCase "wf_env_typ". 
       pose proof (IHG2 A H1 x_notin_A) as wf_TL.
       rewrite_env (([(X, binding_typ (subst_typ x v T))]) ++ (subst_gamma x v G2 ++ G1)).
       pose proof (dom_concat _ G1 G2) as domEq.
       pose proof (dom_concat _ G1 (subst_gamma x v G2)) as domSubEq.
       pose proof (subst_gamma_dom G2 x v) as domGSubEq.
       eapply wf_env_typ. assumption. rewrite domSubEq. 
       rewrite <- domGSubEq.  simpl_env in H2. fsetdec.
       pose proof (wf_exp_weakening G1 G2 v wf_v) as wf_v_G1G2.
       assert (dom (G2 ++ G1) = dom (subst_gamma x v G2 ++ G1)).
       rewrite domSubEq. rewrite <- domGSubEq. auto.
       simpl_env in H1. simpl_env in H4.
       eapply subst_wf_typ.  apply H1. apply H4. apply wf_v.
       SCase "wf_env_typ_A".
       rewrite_env (([(X, binding_typ (subst_typ x v T))]) ++ (subst_gamma x v G2 ++ G1)).
       assert (X `notin` A0) as X_notin_A0.
         simpl_env in H2.
         assert (AtomSet.F.Subset A0 (dom (G2 ++ ([(x, binding_typ t_x)]) ++ G1))) as A0_subset.
           apply (wf_env_G_A_containment _ _ H2).
        simpl_env in A0_subset. simpl_env in H3. 
        unfold not. intro.
        unfold AtomSet.F.Subset in A0_subset.
        pose proof (A0_subset X H5) as for_contra. contradiction.
       destruct (aqual_union_identity_1 X A0 x k A H1 X_notin_A0 x_notin_A) as [A1 [A0_eq_aqxA1 x_notin_A1]].
       subst A0.
       assert (x <> X) as X_neq_x.
         unfold not; intro.
         subst x. 
         simpl_env in H3. fsetdec.
       destruct (aqual_union_identity_2 X x k  _ A H1 X_neq_x) as [A2 [A_eq_XA2 X_notin_A2]].
       assert (X `notin` A1) as X_notin_A1.
         unfold not; intro.
         pose proof (aqual_identity_0 X x k A1 H5) as for_contra. contradiction.
       assert (A1 = A2) as A1_eq_A2.
         eapply aqual_identity_3. apply X_notin_A1. assumption. apply x_notin_A1. apply x_notin_A. apply H1. apply (sym_equal A_eq_XA2).
       subst A2.
       pose proof (IHG2 A1 H2 x_notin_A1) as wf_TL.
       pose proof (dom_concat _ G1 G2) as domEq.
       pose proof (dom_concat _ G1 (subst_gamma x v G2)) as domSubEq.
       pose proof (subst_gamma_dom G2 x v) as domGSubEq.
       subst A.
       eapply wf_env_typ_A. assumption.
       rewrite domSubEq. rewrite <- domGSubEq. 
         simpl_env in H3. fsetdec.
       eapply subst_wf_typ. 
         simpl_env in H2. apply H2.
         simpl_env in H4. apply H4.
         apply wf_v.
       SCase "wf_env_kind".
       pose proof (IHG2 A H1 x_notin_A).
       rewrite_env ([(X, binding_kind K)] ++ subst_gamma x v G2 ++ G1).
       eapply wf_env_kind. assumption.
       pose proof (dom_concat _ G1 G2) as domEq.
       pose proof (dom_concat _ G1 (subst_gamma x v G2)) as domSubEq.
       pose proof (subst_gamma_dom G2 x v) as domGSubEq.
       rewrite domSubEq. rewrite <- domGSubEq. 
       simpl_env in H3. fsetdec.
Qed.
 

Lemma Ax_empty_A_empty : forall A x k, (aqual x k A) = {} -> A = {}.
Admitted.

Lemma Gamma_subst_binding_typ : forall G x y t v, 
  (binds x (binding_typ t) G) -> (y <> x) -> (binds x (binding_typ (subst_typ y v t)) (subst_gamma y v G)).
Admitted.

Lemma weakening : forall G1 G2 A1 A2 A e t, 
  wf_env G1 A1 -> splitA A1 A2 A -> wf_env (G2 ++ G1) A -> 
  typing G1 A1 e t -> typing (G2 ++ G1) A e t.
Admitted.

Lemma subst_wf_typ_ext : forall G t x v, 
  wf_typ t G -> x `notin` dom G -> (subst_typ x v t) = t.
Admitted.

Lemma subst_wf_exp_ext : forall G e x v, 
  wf_exp e G -> x `notin` dom G -> (subst_exp x v e) = e.
Admitted.

Lemma typing_wf_typ : forall G A e t fx, 
  typing G A e (t, fx) -> wf_typ t G.
Admitted.

Lemma typing_kinded_typ : forall G A e t fx, 
  typing G A e (t, fx) -> exists k, kinding G t k.
Admitted.

Lemma kinding_wf_typ : forall G t k, 
  kinding G t k -> wf_typ t G. 
Admitted.

Lemma typing_effect_ok : forall G A e t fx, 
  typing G A e (t, fx) -> effect_ok fx G.
Admitted.

Lemma effect_ok_weakening : forall G1 G2 fx A, 
  effect_ok fx G1 -> wf_env (G2 ++ G1) A -> effect_ok fx (G2 ++ G1).
Admitted.

Lemma effect_ok_subst : forall G1 x tx G2 fx1 fx2 A fx v, 
  wf_env (G2 ++ [(x, binding_typ tx)] ++ G1) A ->
  effect_ok fx1 (G2 ++ [(x, binding_typ tx)] ++ G1) ->
  effect_ok fx2 G1 ->
  splitA fx1 fx2 fx -> 
  effect_ok fx (subst_gamma x v G2 ++ G1).
Admitted.

Lemma typing_wf_exp : forall G A e t fx, 
  typing G A e (t,fx) -> wf_exp e G.
Admitted.
  
Lemma binds_wf : forall G A x t, 
  wf_env G A -> (binds x (binding_typ t) G) -> wf_typ t G.
intros.
generalize dependent A.
induction G.
(* Case "G_nil". *)
inversion H0.
intros.
(* Case "G_cons". *)
inversion H.
(* Case "wf_env_typ". *)
subst.
rewrite_env ([(X, binding_typ T)] ++ G) in H0.
destruct (binds_concat_inv binding x (binding_typ t) G _ H0) as [ [xNotX0 G_binds_x] | x_is_X0].
(* G_binds_x *)
pose proof (IHG G_binds_x A H3) as wf_t_G.
rewrite_env ([(X, binding_typ T)] ++ G).
eapply wf_typ_weakening; auto.
(* x_is_X0 *)
inversion x_is_X0.
destruct (x == X).
  (* subcase x=X *)
  assert (t = T).
  injection H2; auto.
  subst.
  inversion H.
     (* SubCase : wf_env_typ *)
     rewrite_env ([(X, binding_typ T)] ++ G).
     eapply wf_typ_weakening. assumption.
     rewrite_env ([(X, binding_typ T)] ++ G).
     eapply wf_typ_weakening. assumption.
  (* subcase x <> X *)
  discriminate.
(* Case "wf_env_typ_A". *)
subst.
rewrite_env ([(X, binding_typ T)] ++ G) in H0.
destruct (binds_concat_inv binding x (binding_typ t) G _ H0) as [ [xNotX0 G_binds_x] | x_is_X0].
(* G_binds_x *)
pose proof (IHG G_binds_x A0 H3) as wf_t_G.
rewrite_env ([(X, binding_typ T)] ++ G).
eapply wf_typ_weakening; auto.
(* x_is_X0 *)
inversion x_is_X0.
destruct (x == X).
  (* subcase x=X *)
  assert (t = T).
  injection H2; auto.
  subst.
  inversion H.
     (* SubCase : wf_env_typ *)
     rewrite_env ([(X, binding_typ T)] ++ G).
     eapply wf_typ_weakening. assumption.
     rewrite_env ([(X, binding_typ T)] ++ G).
     eapply wf_typ_weakening. assumption.
  (* subcase x <> X *)
  discriminate.
(* Case "wf_env_kind *)
subst.
rewrite_env ([(X, binding_kind K)] ++ G) in H0.
destruct (binds_concat_inv binding x (binding_typ t) G _ H0) as [ [xNotX G_binds_x] | x_is_X].
pose proof (IHG G_binds_x A H3) as wf_t_G.
rewrite_env ([(X, binding_kind K)] ++ G).
eapply wf_typ_weakening; auto.
inversion x_is_X. destruct (x == X); discriminate.
Qed.

Lemma binds_disjoint_1 : forall G1 x t G2 A, 
  wf_env (G2 ++ ([(x, binding_typ t)]) ++ G1) A -> x `notin` dom G1.
Admitted.

Lemma binds_disjoint_2 : forall G1 x t G2 A, 
  wf_env (G2 ++ ([(x, binding_typ t)]) ++ G1) A -> x `notin` dom G2.
Admitted.

Lemma aqual_commutes : forall A x k_x y k_y, 
  aqual x k_x (aqual y k_y A) = aqual y k_y (aqual x k_x A).
Admitted.

Lemma aqual_split_commute : forall A A0 x k a, 
  splitA (singleton a) A (aqual x k A0) ->  x <> a -> 
  exists A1, A = aqual x k A1 /\ (x `notin` A1).
Admitted.

Lemma split_aqual_dom_containment : forall A0 x k A A', 
  splitA A0 (aqual x k A) (aqual x k A') -> x `notin` A -> AtomSet.F.Subset A A'.
Admitted.

Lemma wf_env_containment : forall G A A', 
  wf_env G A -> AtomSet.F.Subset A' A -> wf_env G A'.
Admitted.

Lemma wf_env_aqual : forall G A x t k, 
  wf_env G A -> binds x (binding_typ t) G -> wf_env G (aqual x k A).
Admitted.

Lemma splitA_subset : forall A A', 
  AtomSet.F.Subset A A' -> exists A0, splitA A A0 A'.
Admitted.

Lemma split_trans_1 : forall A1 A2 A2_1 A2_2 A, 
  splitA A1 A2 A -> splitA A2_1 A2_2 A2 -> exists B, splitA A1 A2_1 B.
Admitted.

Lemma split_trans_2 : forall A1 A2 A2_1 A2_2 A, 
  splitA A1 A2 A -> splitA A2_1 A2_2 A2 -> exists B, splitA A1 A2_2 B.
Admitted.

Lemma binds_kind_subst : forall G a x v k, 
  binds a (binding_kind k) G -> binds a (binding_kind k) (subst_gamma x v G).
Admitted.

Lemma inter_distributes : forall A B C, 
  A `inter` B `union` C = {} -> (A `inter` C = {}) /\ (A `inter` B = {}).
Admitted.

Lemma inter_symmetric : forall A B C, 
  A `inter` B  = C -> B `inter` A = C.
Admitted.

Lemma inter_distributes_inv : forall A B C, 
  (A `inter` B = {}) -> (A `inter` C = {}) -> (A `inter` (B `union` C) = {}).
Admitted.

Lemma subst_ptype : forall x v t A fx, 
  (subst_typ x v (ptype t A fx)) = (ptype (subst_typ x v t) A fx).
Admitted.

Lemma empty_union : forall A B, 
  AtomSet.F.is_empty (A `union` B) = true -> (A = {}) /\ (B = {}).
Admitted.

Lemma non_empty_union : forall A, 
 AtomSet.F.is_empty A = false -> exists x, x `in` A.
Admitted.

Lemma non_empty_union_2 : forall A, 
  A <> {} -> exists x, x `in` A.
Admitted.

Lemma non_empty_union_inv : forall A x, 
  x `in` A -> AtomSet.F.is_empty A = false.
Admitted.

Lemma exclude_middle_empty_sets : forall (A:atoms), (A={}) \/ (A <> {}).
Admitted.

Lemma tequiv_preserves_kinding : forall G A t t' k, 
  kinding G t k -> tequiv G A t t' -> kinding G t' k.
Admitted.

Lemma kinding_uniqueness : forall G t k k', 
  kinding G t k -> kinding G t k' -> k = k'.
Admitted.

Lemma inversion_sing : forall G A e t a fx, 
  value e -> typing G A e (typ_sing t a, fx) -> 
  ((exists e', e = exp_new e') /\ fx <> {}).
Admitted.

Lemma inversion_empty_effect : forall G v t, 
  value v -> typing G {} v (t, {}) -> kinding G t kind_U.
Proof with inversion H.
intros.
assert (typing G {} v (t, {})) as T; auto.
remember (t, {}) as t_fx.
remember ({}) as fx.
generalize dependent t.
induction H0.
Case "type_X"...
Case "type_XA"...
Case "type_NEW"...
 intros. subst A. injection Heqt_fx; intros; subst a_fx.
  inversion H3. assert (a `in` {}). rewrite <- H6. fsetdec. fsetdec.
Case "type_ABS".
 intros; injection Heqt_fx; intros; subst A. subst fx. 
   rewrite H4 in T.
   destruct (typing_kinded_typ _ _ _ _ _ T) as [k_t K].
     unfold ptype in H4.
     case_eq (AtomSet.F.is_empty ({} `union` {})).
      intro empty_true. 
        rewrite empty_true in H4.
        subst t. inversion K. subst k_t; auto.
      intro empty_false.
        destruct (non_empty_union  _ empty_false) as [X X_in_nul]. fsetdec.
Case "type_TAB".
 intros; injection Heqt_fx; intros; subst A. subst fx. 
   rewrite H4 in T.
   destruct (typing_kinded_typ _ _ _ _ _ T) as [k_t K].
     unfold ptype in H4.
     case_eq (AtomSet.F.is_empty ({} `union` {})).
      intro empty_true. 
        rewrite empty_true in H4.
        subst t. inversion K. subst k_t; auto.
      intro empty_false.
        destruct (non_empty_union  _ empty_false) as [X X_in_nul]. fsetdec.
Case "type_TAB_FX".
 intros. injection Heqt_fx. intros eq_fx eq_t; subst A. subst fx.
   rewrite eq_t in T.
   destruct (typing_kinded_typ _ _ _ _ _ T) as [k_t K].
     unfold ptype in eq_t.
     case_eq (AtomSet.F.is_empty ({} `union` {})).
      intro empty_true. 
        rewrite empty_true in eq_t.
        subst t. inversion K. subst k_t; auto.
      intro empty_false.
        destruct (non_empty_union  _ empty_false) as [X X_in_nul]. fsetdec.
Case "type_APP"...
Case "type_QAPP"...
Case "type_TAP_NOFX"...
Case "type_TAP_FX"...
Case "type_QTAP_NOFX"...
Case "type_QTAP_FX"...
Case "type_BOT"...
Case "type_DROP".
  intros. injection Heqt_fx; intros; subst A; subst t0. subst fx.
  destruct (inversion_sing _ _ _ _ _ _ H H0) as [[e' e_eq_newe'] fx_neq_nul]. 
    destruct fx_neq_nul. auto.
Case "type_AFN".
  intros. injection Heqt_fx; intros; subst A; subst t0. subst fx.
  destruct (split_nil_nil A1 A2 H0).
  destruct (split_nil_nil fx1 fx2 H2). subst fx1. subst A1.
  eapply IHtyping; eauto.
Case "type_B"...
Focus 4.
Case "type_CONV".
  intros; injection Heqt_fx; intros; subst A; subst t0; subst fx.
  eapply tequiv_preserves_kinding; eauto using H1.
Case "type_B1". (* handled by well-formedness of signature : TODO *)
Admitted.
  
Lemma inversion_nonempty_effect : forall G A e t fx,
  (value e) ->
  typing G A e (t, fx) -> fx <> {} -> 
  (kinding G t kind_A \/ typing G A e (t, {})).
Proof with inversion H. 
intros.
assert (typing G A e (t, fx)) as T; auto.
remember (t, fx) as t_fx.
generalize dependent t.
generalize dependent fx.
induction H0.
Case "type_X"...
Case "type_XA"...
Case "type_NEW".
  intros.
  injection Heqt_fx; intros; subst t0.
  left.
  eapply kinding_N.
  eapply kinding_AFN. auto.
  eapply kinding_A; auto.
Case "type_ABS".
  intros.
  injection Heqt_fx; intros; subst t; subst fx0.
  left.
  unfold ptype.
  case_eq (AtomSet.F.is_empty (A `union` fx)).
  intro u_empty.
    destruct (empty_union _ _ u_empty) as [_ fx_empty]. 
    contradiction.
  intro.
   unfold ptype in T.
     rewrite H4 in T.
   destruct (typing_kinded_typ _ _ _ _ _ T) as [kk t_kind_k].
   inversion t_kind_k. subst kk; auto.
Case "type_TAB".
  intros.
  injection Heqt_fx. intros eq_fx eq_t. subst t. subst fx0.
  left.
  unfold ptype.
  case_eq (AtomSet.F.is_empty (A `union` fx)).
  intro u_empty.
    destruct (empty_union _ _ u_empty) as [_ fx_empty]; contradiction.
  intro.
    unfold ptype in T.
      rewrite H4 in T.
    destruct (typing_kinded_typ _ _ _ _ _ T) as [kk t_kind_k].
    inversion t_kind_k; subst kk; auto.
Case "type_TAB_FX".
  intros.
  injection Heqt_fx; intros; subst t. subst fx.
  left.
  unfold ptype.
  case_eq (AtomSet.F.is_empty (A `union` fx0)).
  intro u_empty.
    destruct (empty_union _ _ u_empty) as [_ fx_empty]; contradiction.
  intro.
    unfold ptype in T.
      rewrite H5 in T.
    destruct (typing_kinded_typ _ _ _ _ _ T) as [kk t_kind_k].
    inversion t_kind_k; subst kk; auto.
Case "type_APP"...
Case "type_QAPP"...
Case "type_TAP_NOFX"...
Case "type_TAP_FX"...
Case "type_QTAP_NOFX"...
Case "type_QTAP_FX"...
Case "type_BOT"...
Case "type_DROP".
  intros. injection Heqt_fx; intros; subst t0; subst fx0.
  destruct (typing_kinded_typ _ _ _ _ _ H0) as [kk t_kind_k].
  inversion t_kind_k.
  left; auto.
Case "type_AFN".
  intros; injection Heqt_fx; intros; subst t0; subst fx0.
  (* Hmm ... not sure how to proceed here without excluded middle *)
  destruct (exclude_middle_empty_sets fx1) as [fx1_empty | fx1_nonempty].
  subst fx1; right.
    eapply type_AFN. apply H0. apply H1. apply (split_nil_A_1 {}).
    unfold effect_ok. fsetdec.
  destruct (IHtyping H H1 fx1 fx1_nonempty t (refl_equal (t, fx1))) as [KK | TT].
    left; auto.
    right.
      eapply type_AFN. apply H0. apply TT. apply (split_nil_A_1 {}).
      unfold effect_ok. fsetdec.
Case "type_B"...
Focus 4.
Case "type_CONV".
  intros; injection Heqt_fx; intros; subst t0; subst fx0.
  destruct (IHtyping H H0 fx H2 t (refl_equal (t,fx))) as [ KK | TT ].
  left.
    eapply (tequiv_preserves_kinding G (affines G)). apply KK. apply H1.
  right.
    eapply type_CONV. apply TT. apply H1.
Case "type_B1". (* remaining cases handled by well-formedness of types in a Signature *)
(* TODO *)
Admitted.

Lemma ptype_union_equality : forall t A1 fx1 A2 fx2, 
  (AtomSet.F.is_empty (A1 `union` fx1) = AtomSet.F.is_empty (A2 `union` fx2)) -> 
   ptype t A1 fx1 = ptype t A2 fx2.
Admitted.

Lemma split_extension_R : forall A B C x, 
  splitA A B C -> x `notin` C -> splitA A (B `union` singleton x) (C `union` singleton x).
Admitted.

Lemma subst_preserves_typing : forall G1 G2 A A' x tx k e t v fx_e fx_v fx_ev,
  wf_env G1 {} -> kinding G1 tx k -> (x `notin` A) -> (A' = (aqual x k A) \/ A' = A) -> (* wf_env (G2 ++ G1) A ->  *)
  wf_env (G2 ++ [(x, binding_typ tx)] ++ G1) A' ->
  typing (G2 ++ [(x, binding_typ tx)] ++ G1) A' e (t, fx_e) ->
  value v -> typing G1 {} v (tx, fx_v) -> splitA fx_v fx_e fx_ev ->
  typing ((subst_gamma x v G2) ++ G1) A (subst_exp x v e) ((subst_typ x v t), fx_ev).
Proof.
intros until 1. rename H into wf_G1.
intros Kind_txk x_notin_A A_disjunct WF_G1xG2Ax T Val_v T_v split_fx.
assert (typing (G2 ++ [(x, binding_typ tx)] ++ G1) A' e (t, fx_e)) as Te; auto.
remember (G2 ++ [(x, binding_typ tx)] ++ G1) as G.
(* remember (aqual x k A) as Ax. *)
remember (t, fx_e) as t_fxe.
assert (wf_env G A') as WF_GA'; auto.
generalize dependent Te. 
generalize dependent HeqG.
(* generalize dependent HeqAx. *)
generalize dependent WF_GA'.
generalize dependent A.
generalize dependent t.
generalize dependent fx_v.
generalize dependent fx_e.
generalize dependent fx_ev.
generalize dependent G1.
generalize dependent G2.
destruct T using typing_ind2 with
  (P0 := 
    fun (G:gamma) (t:typ) (k:kind) (K:kinding G t k) =>
      forall G1 G2, (G = (G2 ++ [(x, binding_typ tx)] ++ G1)) -> kinding ((subst_gamma x v G2) ++ G1) (subst_typ x v t) k).
Case "typing_X".
intros G2 G1 wf_G1 Kind_txk fx_ev fx_e fx_v T_v split_fx t0 t_fxe A x_notin_A disjunct_A' wf_GEmpty G_eq_G2xG1 T_x0.
injection t_fxe; intros. subst t0. subst fx_e.
assert (A = {}) as A_Empty.
  destruct disjunct_A' as [AxEmpty | A_Empty];
     [ apply (Ax_empty_A_empty A x k (sym_equal AxEmpty)) | auto ].
rewrite G_eq_G2xG1 in b.
destruct (binds_concat_inv binding x0 (binding_typ t) _ G2 b) as [ [x0_Notin_G2 binds_x0_xG1] | binds_x0_G2].
destruct (binds_concat_inv binding x0 (binding_typ t) G1 _ binds_x0_xG1) as [ [ x0_Notin_x binds_x0_G1 ] | binds_x0_x].
(* x0 in G1 *)
simpl.
destruct (x == x0) as [x_eq_x0 | x_new_x0].
subst. destruct x0_Notin_x. simpl. fsetdec.
rewrite A_Empty. 
   eapply type_AFN.
   eapply split_nil_A_1. (* first premise of T-AFN: {},{} *)
   eapply type_X. (* second premise : IH *)
      assert ((subst_typ x v t) = t) as subt_eq_t.
         assert (wf_typ t G1) as wf_t_G1.
           eapply binds_wf. apply wf_G1. apply binds_x0_G1.
         eapply subst_wf_typ_ext. apply wf_t_G1. eapply binds_disjoint_1.
         rewrite G_eq_G2xG1 in WF_G1xG2Ax. apply WF_G1xG2Ax.
      rewrite subt_eq_t.
      assert (dom G2 = dom (subst_gamma x v G2)) as dom_G2.
         eapply (subst_gamma_dom G2 x v).
      rewrite dom_G2 in x0_Notin_G2.
         eapply binds_tail; auto. 
      apply IHT. assumption.
   eapply (split_commutes _ _ _ split_fx). (* third premise of T-AFN *)
   assert (effect_ok fx_v G1) as ok_G1. (* fourth premise of T-AFN *)
     eapply typing_effect_ok; auto. apply T_v.
   assert (fx_v = fx_ev) as fx_eq.
     eapply split_nil_eq_2; auto.
     assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2G1.
        eapply subst_gamma_wf. 
           assert (aqual x kind_U A = {}) as ax_U.
              rewrite A_Empty; simpl; auto.
           rewrite <- ax_U in wf_GEmpty.
           rewrite G_eq_G2xG1 in wf_GEmpty. apply wf_GEmpty.
        eapply typing_wf_exp. apply T_v. 
        apply x_notin_A.
     eapply effect_ok_weakening. subst; auto. apply wf_sG2G1.
(* binds_x0_x ... x0 in [(x, t)] *)
inversion binds_x0_x.
destruct (x0 == x) as [x0_eq_x | x0_neq_x].
rewrite x0_eq_x. simpl.
destruct (x==x).
assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2_G1.
   eapply subst_gamma_wf. 
      assert (aqual x kind_U A = {}) as ax_U.
         rewrite A_Empty; simpl; auto.
      rewrite <- ax_U in wf_GEmpty.
      rewrite G_eq_G2xG1 in wf_GEmpty.  apply wf_GEmpty. 
   eapply typing_wf_exp. apply T_v. 
   apply x_notin_A.
pose proof (weakening G1 (subst_gamma x v G2) {} A A v (tx, fx_v) wf_G1 (split_nil_A_1 A) wf_sG2_G1 T_v) as T_v'.
assert (tx = t) as tx_eq_t.
  injection H0. auto. 
rewrite <- tx_eq_t.
assert (subst_typ x v tx = tx) as stx_tx.
  eapply subst_wf_typ_ext.
  eapply typing_wf_typ. apply T_v.
  eapply binds_disjoint_1. rewrite G_eq_G2xG1 in WF_G1xG2Ax. apply WF_G1xG2Ax.
assert (fx_v = fx_ev) as eq_fx.
  eapply split_nil_eq_2. assumption.
rewrite stx_tx. rewrite <- eq_fx. assumption.
destruct n. auto. 
discriminate.
(* x0 in G2 *)
simpl. destruct (x == x0) as [x_eq_x0 | x_neq_x0]. 
(* Subcase x = x0 (impos) *)
assert (x0 `notin` dom G2) as x0_notin_G2.
  eapply binds_disjoint_2.
  rewrite G_eq_G2xG1 in WF_G1xG2Ax. subst. apply WF_G1xG2Ax.
assert (x0 `in` dom G2) as x0_in_G2. eapply binds_In. apply binds_x0_G2. fsetdec.
(* Subcase x <> x0 *)
pose proof (Gamma_subst_binding_typ G2 x0 x t v binds_x0_G2 x_neq_x0) as binds_subx0_G2.
pose proof (binds_head binding x0 _ G1 _ binds_subx0_G2) as binds_subx0_G1G2.
rewrite A_Empty.
   eapply type_AFN.
   eapply split_nil_A_1. (* first premise of T-AFN: {},{} *)
   eapply type_X. (* second premise : IH *)
     assumption.
     eapply IHT. assumption.
   eapply (split_commutes _ _ _ split_fx). (* third premise of T-AFN *)
   assert (effect_ok fx_v G1) as ok_G1. (* fourth premise of T-AFN *)
     eapply typing_effect_ok; auto. apply T_v.
   assert (fx_v = fx_ev) as fx_eq.
     eapply split_nil_eq_2; auto.
     assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2G1.
        eapply subst_gamma_wf. 
           assert (aqual x kind_U A = {}) as ax_U.
              rewrite A_Empty; simpl; auto.
           rewrite <- ax_U in wf_GEmpty.
           rewrite G_eq_G2xG1 in wf_GEmpty. apply wf_GEmpty.
        eapply typing_wf_exp. apply T_v.
        apply x_notin_A.
     eapply effect_ok_weakening. subst; auto. apply wf_sG2G1.
Case "typing_XA".
intros G2 G1 wf_G1 Kind_txk fx_ev fx_e fx_v T_v split_fx t0 t_fxe A x_notin_A disjunct_A' wf_Gx0 G_eq_G2xG1 T_x0. (*Ax0_eq*)
injection t_fxe; intros. subst t0. subst fx_e.
rewrite G_eq_G2xG1 in b.
destruct (binds_concat_inv binding x0 (binding_typ t) _ G2 b) as [ [x0_Notin_G2 binds_x0_xG1] | binds_x0_G2].
destruct (binds_concat_inv binding x0 (binding_typ t) G1 _ binds_x0_xG1) as [ [ x0_Notin_x binds_x0_G1 ] | binds_x0_x].
(* x0 in G1 *)
simpl.
destruct (x == x0) as [x_eq_x0 | x_neq_x0].
subst. destruct x0_Notin_x. simpl. fsetdec.
assert (A = singleton x0) as A_is_x0.
   destruct disjunct_A' as [x0_is_Ax | A_is_x0].
     destruct (aqual_cases A x k) as [L | R]. 
       rewrite L in x0_is_Ax; auto.
       rewrite R in x0_is_Ax. 
       assert (x `in` singleton x0) as x_in_x0.
         rewrite x0_is_Ax. fsetdec.
       assert (x = x0) as contr. fsetdec. 
       contradiction.
     auto.  
rewrite A_is_x0. 
   eapply type_AFN.
   eapply split_nil_A_2. (* first premise of T-AFN: {},{} *)
   eapply type_XA.
   assert ((subst_typ x v t) = t) as subt_eq_t.
     assert (wf_typ t G1) as wf_t_G1.
      eapply binds_wf. apply wf_G1. apply binds_x0_G1.
   eapply subst_wf_typ_ext. apply wf_t_G1. eapply binds_disjoint_1.
     rewrite G_eq_G2xG1 in WF_G1xG2Ax. apply WF_G1xG2Ax.
   rewrite subt_eq_t.
   assert (dom G2 = dom (subst_gamma x v G2)) as dom_G2.
     eapply (subst_gamma_dom G2 x v).
   rewrite dom_G2 in x0_Notin_G2.
   eapply binds_tail; auto. 
   apply IHT. assumption.
   eapply (split_commutes _ _ _ split_fx). (* third premise of T-AFN *)
   assert (effect_ok fx_v G1) as ok_G1. (* fourth premise of T-AFN *)
     eapply typing_effect_ok; auto. apply T_v.
   assert (fx_v = fx_ev) as fx_eq.
     eapply split_nil_eq_2; auto.
     assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2G1.
        eapply subst_gamma_wf.
          assert (aqual x kind_U A = A) as ax_U.
              simpl; auto.
          rewrite <- A_is_x0 in WF_G1xG2Ax. rewrite <- ax_U in WF_G1xG2Ax. 
          rewrite G_eq_G2xG1 in WF_G1xG2Ax. apply WF_G1xG2Ax.
        eapply typing_wf_exp. apply T_v.
        apply x_notin_A.
     eapply effect_ok_weakening. subst; auto. apply wf_sG2G1.
(* binds_x0_x ... x0 in [(x, t)] *)
inversion binds_x0_x.
destruct (x0 == x) as [x0_eq_x | x0_neq_x].
rewrite x0_eq_x. simpl.
destruct (x==x).
assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2_G1.
    rewrite G_eq_G2xG1 in WF_G1xG2Ax. 
    destruct disjunct_A' as [L | R].
       eapply subst_gamma_wf.
         rewrite L in WF_G1xG2Ax. apply WF_G1xG2Ax.
        eapply typing_wf_exp. apply T_v.
        apply x_notin_A.
       eapply subst_gamma_wf.
         rewrite R in WF_G1xG2Ax.
           assert (aqual x kind_U A = A) as ax_U.
             simpl; auto.
           rewrite <- ax_U in WF_G1xG2Ax. apply WF_G1xG2Ax.
         eapply typing_wf_exp; apply T_v.
         apply x_notin_A.
pose proof (weakening G1 (subst_gamma x v G2) {} A A v (tx, fx_v) wf_G1 (split_nil_A_1 A) wf_sG2_G1 T_v) as T_v'.
assert (tx = t) as tx_eq_t.
  injection H0. auto. 
rewrite <- tx_eq_t.
assert (subst_typ x v tx = tx) as stx_tx.
  eapply subst_wf_typ_ext.
  eapply typing_wf_typ. apply T_v.
  eapply binds_disjoint_1. rewrite G_eq_G2xG1 in WF_G1xG2Ax. apply WF_G1xG2Ax.
assert (fx_v = fx_ev) as eq_fx. 
  eapply split_nil_eq_2. assumption.
rewrite stx_tx. rewrite <- eq_fx. assumption.
destruct n. auto. 
discriminate.
(* x0 in G2 *)
simpl. destruct (x == x0) as [x_eq_x0 | x_neq_x0]. 
(* Subcase x = x0 (impos) *)
assert (x0 `notin` dom G2) as x0_notin_G2.
  eapply binds_disjoint_2.
  rewrite G_eq_G2xG1 in WF_G1xG2Ax. subst. apply WF_G1xG2Ax.
assert (x0 `in` dom G2) as x0_in_G2. eapply binds_In. apply binds_x0_G2. fsetdec.
(* Subcase x <> x0 *)
pose proof (Gamma_subst_binding_typ G2 x0 x t v binds_x0_G2 x_neq_x0) as binds_subx0_G2.
pose proof (binds_head binding x0 _ G1 _ binds_subx0_G2) as binds_subx0_G1G2.
assert (A = singleton x0) as A_is_x0.
   destruct disjunct_A' as [x0_is_Ax | A_is_x0].
     destruct (aqual_cases A x k) as [L | R]. 
       rewrite L in x0_is_Ax; auto.
       rewrite R in x0_is_Ax. 
       assert (x `in` singleton x0) as x_in_x0.
         rewrite x0_is_Ax. fsetdec.
       assert (x = x0) as contr. fsetdec. 
       contradiction.
     auto. 
rewrite A_is_x0.
   eapply type_AFN.
   eapply split_nil_A_2. (* first premise of T-AFN: {},{} *)
   eapply type_XA. (* second premise : IH *)
     assumption.
     eapply IHT. assumption.
   eapply (split_commutes _ _ _ split_fx). (* third premise of T-AFN *)
   assert (effect_ok fx_v G1) as ok_G1. (* fourth premise of T-AFN *)
     eapply typing_effect_ok; auto. apply T_v.
   assert (fx_v = fx_ev) as fx_eq.
     eapply split_nil_eq_2; auto.
     assert (wf_env ((subst_gamma x v G2) ++ G1) A) as wf_sG2G1.
        eapply subst_gamma_wf. 
         assert (aqual x kind_U A = A) as ax_U.
              simpl; auto.
         rewrite <- A_is_x0 in WF_G1xG2Ax. rewrite <- ax_U in WF_G1xG2Ax. 
         rewrite G_eq_G2xG1 in WF_G1xG2Ax. apply WF_G1xG2Ax.
        eapply typing_wf_exp. apply T_v.
        apply x_notin_A.
     eapply effect_ok_weakening. subst; auto. apply wf_sG2G1.
Case "typing_NEW".
intros G2 G1 wf_G1 Kind_txk fx_newe_v fx_newe fx_v T_v split_fx t0 t_fxe A0 x_notin_A0 disjunct_A' wf_GA G_eq T_newe.
injection t_fxe; intros. subst t0. subst a_fx. 
destruct (split_trans_2 _ _ _ _ _ split_fx s) as [fx_ev split_fxv_fxe].
simpl.
eapply type_NEW.
  eapply IHT; eauto using T_v, split_fxv_fxe. (* first premise *)
  eapply IHT0; eauto using G_eq. (* second premise *)
  assert (binds a (binding_kind kind_N) (G2 ++ G1)) as binds_G2G1_a.
    assert (x <> a) as x_neq_a.
      unfold not; intro x_eq_a.
      subst.
      pose proof (binds_mid_eq binding a _ _ G1 G2 b (wf_is_ok _ _ WF_G1xG2Ax)) as kind_N_eq_t.
      discriminate.
    eapply binds_remove_mid. rewrite G_eq in b. apply b. auto.
  assert ((dom G2) = dom (subst_gamma x v G2)) as dom_eq.
    eapply (subst_gamma_dom G2 x v).
  destruct (binds_concat_inv binding a (binding_kind kind_N) G1 G2 binds_G2G1_a) as [ [ a_Notin_G2 binds_a_G1 ] | binds_a_G2].
    (* case binds_a_G1 *)
    eapply binds_tail. assumption. rewrite dom_eq in a_Notin_G2. assumption.
    (* case binds_a_G2 *)
    assert (binds a (binding_kind kind_N) (subst_gamma x v G2)) as binds_a_sG2.
      eapply binds_kind_subst; auto.
    eapply binds_head; auto.
  (* final premise ... disjoint effects *)
  inversion s.
  inversion split_fx.  
  inversion split_fxv_fxe.
  rewrite <- H2.
  assert ((fx_v `union` (singleton a) `union` fx) = 
          ((singleton a) `union` (fx_v `union` fx))) as eq. fsetdec.
  rewrite eq.
  eapply split_A.
  rename H into inter_a_fx_nil.
  assert ((singleton a) `inter` fx_v = {}) as inter_a_fxv_nil.
    rewrite <- H2 in H3.
       destruct (inter_distributes fx_v (singleton a) fx H3) as [_ inter_fxv_a_nil].
       apply inter_symmetric. assumption.
  eapply inter_distributes_inv; auto.
Case "type_ABS".
intros G2 G1 wf_G1 Kind_txk fx_lamv fx_lam fx_v T_v split_fx t0 eq_tfx A_minus_x x_notin_Amx disjunct_A' wf_GA G_eq T_lam.
injection eq_tfx; intros. subst t0. subst fx.
simpl.
assert ((subst_typ x v (ptype (typ_fun t_x t_e) A fx_lam)) = (ptype (subst_typ x v (typ_fun t_x t_e)) A fx_lam)) as subst_ptype_eq.
  eapply subst_ptype; auto.
rewrite subst_ptype_eq. simpl.
destruct (exclude_middle_empty_sets fx_v) as [fx_v_empty | fx_v_nonempty].
  (* SubCase 1: fx_v_empty *)
   subst fx_v.
   assert (fx_lam = fx_lamv) as fx_lam_eq_lamv.
     eapply split_nil_eq_1; eauto using split_fx.
   assert (A = A_minus_x) as A_eq_Amx.
     assert (kinding G1 tx kind_U) as kinding_U.
     eapply inversion_empty_effect; eauto using T_v.
     assert (k = kind_U) as k_is_U.
     eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
     destruct disjunct_A' as [ LL | RR ];
       [ subst k; unfold aqual in LL; simpl in LL; assumption | assumption ].
   rewrite A_eq_Amx; rewrite fx_lam_eq_lamv.
   pick fresh y excluding (L `union` (singleton x) `union` dom G) and apply type_ABS.
   rename t_x into t_y. rename k0 into k_y.
    cutrewrite((open_exp 0 y (subst_exp x v e)) = (subst_exp x v (open_exp 0 y e))).
    cutrewrite((open_typ 0 y (subst_typ x v t_e)) = (subst_typ x v (open_typ 0 y t_e))).
    assert (y `notin` L) as y_notin_L. fsetdec.
    cut (wf_env ([(y, binding_typ t_y)] ++ G) (aqual y k_y A)). intro wf_yG.
    cut (x `notin` aqual y k_y A_minus_x). intro x_notin_aqAmx.
    assert ([(y, binding_typ t_y)] ++ G = ([(y, binding_typ t_y)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as yG_eq.
      rewrite G_eq. simpl_env. reflexivity.
    assert ((aqual y k_y A = aqual x k (aqual y k_y A_minus_x)) \/ (aqual y k_y A = aqual y k_y A_minus_x)) as dis.
      destruct disjunct_A' as [LL | RR].
         left; rewrite LL; eauto using aqual_commutes.
         right; rewrite RR; auto.
    pose proof (H y y_notin_L wf_yG (([(y, binding_typ t_y)] ++ G2)) G1 wf_G1 Kind_txk fx_lamv fx_lam {} T_v split_fx _ (refl_equal ((open_typ 0 y t_e), fx_lam))
                  (aqual y k_y A_minus_x) x_notin_aqAmx dis wf_yG yG_eq (t y y_notin_L)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : x_notin_aqAmx *)
    unfold not. intro x_in_aqAmx.
    destruct (aqual_identity x y k_y A_minus_x x_in_aqAmx) as [x_eq_y | x_in_Amx].
      subst x. fsetdec.
      contradiction.
    (* cut 2 : wf_yG *)
    pose proof (kinding_wf_typ G t_y k_y k1) as wf_t_y.
    destruct (aqual_cases A y k_y) as [ aq_eq_A | aq_eq_yA ].
      rewrite aq_eq_A.
      apply wf_env_typ. assumption. fsetdec. assumption.
      rewrite aq_eq_yA.
      apply wf_env_typ_A. assumption. fsetdec. assumption.
    (* cut 3 : (subst_typ x) (open_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_typ_open_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v). assumption.
    (* cut 4 : (subst_exp x) (open_exp y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v); assumption.
(* Second premise of type_ABS *)
    eapply IHT; auto.
(*** SubCase 2 : fx_v_nonempty ***)
    destruct (inversion_nonempty_effect _ _ _ _ _ Val_v T_v fx_v_nonempty) as [ tx_kind_A | T_v_fxempty ].
   (** SubSubCase 2a. tx_kind_A **)
     SCase "tx_kind_A".
      assert (k = kind_A) as k_is_A.
        eapply kinding_uniqueness; eauto using tx_kind_A, Kind_txk.
      assert (x `in` A) as x_in_A. (* this splits into two cases depending on disjunct_A' *)
        subst k; unfold aqual in A_eq; simpl in A_eq; fsetdec.
      destruct (non_empty_union_2 _ fx_v_nonempty) as [fx_elt fx_elt_in_v].
      assert (fx_elt `in` fx_lamv) as v_in_lamv.
        inversion split_fx; fsetdec.
      assert (AtomSet.F.is_empty (A `union` fx_lam) = false) as nonempty_L.
        eapply non_empty_union_inv; [eauto || fsetdec].
      assert (AtomSet.F.is_empty (A_minus_x `union` fx_lamv) = false) as nonempty_R.
        eapply non_empty_union_inv; [eauto || fsetdec].
      rewrite <- nonempty_R in nonempty_L.
      assert (ptype (typ_fun (subst_typ x v t_x) (subst_typ x v t_e)) A fx_lam = 
              ptype (typ_fun (subst_typ x v t_x) (subst_typ x v t_e)) A_minus_x fx_lamv) as rw.
        eapply ptype_union_equality; eauto using nonempty_L.
      rewrite rw. (* now starts the reconstruction of the derivation *)
      pick fresh y excluding (L `union` (singleton x) `union` dom G) and apply type_ABS.
      rename t_x into t_y. rename k0 into k_y.
      cutrewrite((open_exp 0 y (subst_exp x v e)) = (subst_exp x v (open_exp 0 y e))).
      cutrewrite((open_typ 0 y (subst_typ x v t_e)) = (subst_typ x v (open_typ 0 y t_e))).
      assert (y `notin` L) as y_notin_L. fsetdec.
      cut (wf_env ([(y, binding_typ t_y)] ++ G) (aqual y k_y A)). intro wf_yG.
      cut (aqual y k_y A = aqual x k (aqual y k_y A_minus_x)). intro aq_eq.
      cut (x `notin` aqual y k_y A_minus_x). intro x_notin_aqAmx.
      assert ([(y, binding_typ t_y)] ++ G = ([(y, binding_typ t_y)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as yG_eq.
      rewrite G_eq. simpl_env. reflexivity.
      pose proof (H y y_notin_L wf_yG (([(y, binding_typ t_y)] ++ G2)) G1 wf_G1 Kind_txk fx_lamv fx_lam fx_v T_v split_fx _ (refl_equal ((open_typ 0 y t_e), fx_lam))
        (aqual y k_y A_minus_x) x_notin_aqAmx wf_yG aq_eq yG_eq (t y y_notin_L)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
      (* cut 1 : x_notin_aqAmx *)
      unfold not. intro x_in_aqAmx.
      destruct (aqual_identity x y k_y A_minus_x x_in_aqAmx) as [x_eq_y | x_in_Amx].
      subst x. fsetdec.
      contradiction.
      (* cut 2 : aq_eq *)
      subst A. eapply aqual_commutes; auto.
               (* cut 3 : wf_yG *)
      pose proof (kinding_wf_typ G t_y k_y k1) as wf_t_y.
      destruct (aqual_cases A y k_y) as [ aq_eq_A | aq_eq_yA ].
      rewrite aq_eq_A.
      apply wf_env_typ. assumption. fsetdec. assumption.
      rewrite aq_eq_yA.
      apply wf_env_typ_A. assumption. fsetdec. assumption.
                                               (* cut 4 : (subst_typ x) (open_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_typ_open_typ_commute.
      fsetdec.
      destruct (typing_wf_exp G1 {} v tx fx_v T_v). assumption.
                                                  (* cut 5 : (subst_exp x) (open_exp y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_commute.
      fsetdec.
      destruct (typing_wf_exp G1 {} v tx fx_v T_v); assumption.
      (* Second premise of type_ABS *)
      eapply IHT; auto.
      (*** Subcase 2b ***)
      SCase "T_v_fxempty".
      assert (typing (subst_gamma x v G2 ++ G1) A_minus_x
                     (exp_lam (subst_typ x v t_x) (subst_exp x v e))
                        (ptype (typ_fun (subst_typ x v t_x) (subst_typ x v t_e)) A fx_lam,
                         fx_lam)) as premise_TAFN.
     assert (A = A_minus_x) as A_eq_Amx.
       assert (kinding G1 tx kind_U) as kinding_U.
       eapply inversion_empty_effect; eauto using T_v.
     assert (k = kind_U) as k_is_U.
       eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
       subst k; unfold aqual in A_eq; simpl in A_eq; assumption.
     rewrite A_eq_Amx.
     pick fresh y excluding (L `union` (singleton x) `union` dom G) and apply type_ABS.
     rename t_x into t_y. rename k0 into k_y.
     cutrewrite((open_exp 0 y (subst_exp x v e)) = (subst_exp x v (open_exp 0 y e))).
    cutrewrite((open_typ 0 y (subst_typ x v t_e)) = (subst_typ x v (open_typ 0 y t_e))).
    assert (y `notin` L) as y_notin_L. fsetdec.
    cut (wf_env ([(y, binding_typ t_y)] ++ G) (aqual y k_y A)). intro wf_yG.
    cut (aqual y k_y A = aqual x k (aqual y k_y A_minus_x)). intro aq_eq.
    cut (x `notin` aqual y k_y A_minus_x). intro x_notin_aqAmx.
    assert ([(y, binding_typ t_y)] ++ G = ([(y, binding_typ t_y)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as yG_eq.
      rewrite G_eq. simpl_env. reflexivity.
      pose proof (H y y_notin_L wf_yG (([(y, binding_typ t_y)] ++ G2)) G1 wf_G1 Kind_txk fx_lam fx_lam {} T_v_fxempty 
                  (split_nil_A_1 fx_lam) _ (refl_equal ((open_typ 0 y t_e), fx_lam))
                  (aqual y k_y A_minus_x) x_notin_aqAmx wf_yG aq_eq yG_eq (t y y_notin_L)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : x_notin_aqAmx *)
    unfold not. intro x_in_aqAmx.
    destruct (aqual_identity x y k_y A_minus_x x_in_aqAmx) as [x_eq_y | x_in_Amx].
      subst x. fsetdec.
      contradiction.
    (* cut 2 : aq_eq *)
    subst A. eapply aqual_commutes; auto.
    (* cut 3 : wf_yG *)
    pose proof (kinding_wf_typ G t_y k_y k1) as wf_t_y.
    destruct (aqual_cases A y k_y) as [ aq_eq_A | aq_eq_yA ].
      rewrite aq_eq_A.
      apply wf_env_typ. assumption. fsetdec. assumption.
      rewrite aq_eq_yA.
      apply wf_env_typ_A. assumption. fsetdec. assumption.
    (* cut 4 : (subst_typ x) (open_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_typ_open_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v_fxempty). assumption.
    (* cut 5 : (subst_exp x) (open_exp y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v_fxempty); assumption.
(* Second premise of type_ABS *)
    eapply IHT; auto.
   (* Finally ... conclude with T-AFN *)
   eapply type_AFN.
      apply (split_nil_A_2 A_minus_x).
      apply premise_TAFN.
      eapply (split_commutes _ _ _ split_fx).
      subst G.
      eapply effect_ok_subst.
        apply wf_GA. 
        eapply typing_effect_ok. 
          apply T_lam.
        eapply typing_effect_ok.
          apply T_v. apply (split_commutes _ _ _ split_fx).
Case "typing_TAB_NOFX".
intros G2 G1 wf_G1 Kind_txk fx_tabv fx_tab fx_v T_v split_fx t0 eq_tfx A_minus_x x_notin_Amx wf_GA A_eq G_eq T_tab.
injection eq_tfx; intros. subst t0. subst fx.
simpl.
assert ((subst_typ x v (ptype (typ_univ k0 NO_FX t_e) A fx_tab)) = (ptype (subst_typ x v (typ_univ k0 NO_FX t_e)) A fx_tab)) as subst_ptype_eq.
  eapply subst_ptype; auto.
rewrite subst_ptype_eq. simpl.
destruct (exclude_middle_empty_sets fx_v) as [fx_v_empty | fx_v_nonempty].
  (* SubCase 1: fx_v_empty *)
   subst fx_v.
   assert (fx_tab = fx_tabv) as fx_tab_eq_tabv.
     eapply split_nil_eq_1; eauto using split_fx.
   assert (A = A_minus_x) as A_eq_Amx.
     assert (kinding G1 tx kind_U) as kinding_U.
     eapply inversion_empty_effect; eauto using T_v.
     assert (k = kind_U) as k_is_U.
     eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
     subst k; unfold aqual in A_eq; simpl in A_eq; assumption.
   rewrite A_eq_Amx; rewrite fx_tab_eq_tabv.
   pick fresh b excluding (L `union` (singleton x) `union` dom G) and apply type_TAB.
   intro b_notin_fxtabv.
   rename k0 into k_b.
    cutrewrite((open_exp_typ 0 b (subst_exp x v e)) = (subst_exp x v (open_exp_typ 0 b e))).
    cutrewrite((open_typ_typ 0 b (subst_typ x v t_e)) = (subst_typ x v (open_typ_typ 0 b t_e))).
    assert (b `notin` L) as b_notin_L. fsetdec.
    cut (wf_env ([(b, binding_kind k_b)] ++ G) A). intro wf_bG.
    cut (b `notin` fx_tab). intro b_notin_fxtab.  
(*     cut (aqual y k_y A = aqual x k (aqual y k_y A_minus_x)). intro aq_eq. *)
(*     cut (x `notin` aqual y k_y A_minus_x). intro x_notin_aqAmx. *)
    assert ([(b, binding_kind k_b)] ++ G = ([(b, binding_kind k_b)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
      rewrite G_eq. simpl_env. reflexivity.
    pose proof (H b b_notin_L b_notin_fxtab wf_bG (([(b, binding_kind k_b)] ++ G2)) G1 wf_G1 Kind_txk fx_tabv fx_tab {} T_v split_fx _ 
                (refl_equal ((open_typ_typ 0 b t_e), fx_tab))
                  A_minus_x x_notin_Amx wf_bG A_eq bG_eq (t b b_notin_L b_notin_fxtab)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : b notin fx_tab *)
       subst fx_tabv; auto.
    (* cut 2 : wf_bG *)
    eapply wf_env_kind; ( auto || fsetdec ).
    (* cut 3 : (subst_typ x) (open_typ_typ y) commute (when x <> y) *) 
     apply sym_equal.
      eapply subst_typ_open_typ_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v). assumption.
    (* cut 4 : (subst_exp x) (open_exp_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v); assumption.
(* Second premise of type_ABS *)
     assumption.
(*** SubCase 2 : fx_v_nonempty ***)
    destruct (inversion_nonempty_effect _ _ _ _ _ Val_v T_v fx_v_nonempty) as [ tx_kind_A | T_v_fxempty ].
   (** SubSubCase 2a. tx_kind_A **)
     SCase "tx_kind_A".
      assert (k = kind_A) as k_is_A.
        eapply kinding_uniqueness; eauto using tx_kind_A, Kind_txk.
      assert (x `in` A) as x_in_A.
        subst k; unfold aqual in A_eq; simpl in A_eq; fsetdec.
      destruct (non_empty_union_2 _ fx_v_nonempty) as [fx_elt fx_elt_in_v].
      assert (fx_elt `in` fx_tabv) as v_in_tabv.
        inversion split_fx; fsetdec.
      assert (AtomSet.F.is_empty (A `union` fx_tab) = false) as nonempty_L.
        eapply non_empty_union_inv; [eauto || fsetdec].
      assert (AtomSet.F.is_empty (A_minus_x `union` fx_tabv) = false) as nonempty_R.
        eapply non_empty_union_inv; [eauto || fsetdec].
      rewrite <- nonempty_R in nonempty_L.
      assert ((ptype (typ_univ k0 NO_FX (subst_typ x v t_e)) A fx_tab) =
              (ptype (typ_univ k0 NO_FX (subst_typ x v t_e)) A_minus_x fx_tabv)) as rw.
        eapply ptype_union_equality; eauto using nonempty_L.
      rewrite rw. (* now starts the reconstruction of the derivation *)
   pick fresh b excluding (L `union` (singleton x) `union` dom G) and apply type_TAB.
   intro b_notin_fxtabv.
   rename k0 into k_b.
    cutrewrite((open_exp_typ 0 b (subst_exp x v e)) = (subst_exp x v (open_exp_typ 0 b e))).
    cutrewrite((open_typ_typ 0 b (subst_typ x v t_e)) = (subst_typ x v (open_typ_typ 0 b t_e))).
    assert (b `notin` L) as b_notin_L. fsetdec.
    cut (wf_env ([(b, binding_kind k_b)] ++ G) A). intro wf_bG.
    cut (b `notin` fx_tab). intro b_notin_fxtab.  
(*     cut (aqual y k_y A = aqual x k (aqual y k_y A_minus_x)). intro aq_eq. *)
(*     cut (x `notin` aqual y k_y A_minus_x). intro x_notin_aqAmx. *)
    assert ([(b, binding_kind k_b)] ++ G = ([(b, binding_kind k_b)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
      rewrite G_eq. simpl_env. reflexivity.
    pose proof (H b b_notin_L b_notin_fxtab wf_bG (([(b, binding_kind k_b)] ++ G2)) G1 wf_G1 Kind_txk fx_tabv fx_tab fx_v T_v split_fx _ 
                (refl_equal ((open_typ_typ 0 b t_e), fx_tab))
                  A_minus_x x_notin_Amx wf_bG A_eq bG_eq (t b b_notin_L b_notin_fxtab)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : b notin fx_tab *)
       inversion split_fx. subst fx_tabv; fsetdec.
    (* cut 2 : wf_bG *)
    eapply wf_env_kind; ( auto || fsetdec ).
    (* cut 3 : (subst_typ x) (open_typ_typ y) commute (when x <> y) *) 
     apply sym_equal.
      eapply subst_typ_open_typ_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx fx_v T_v). assumption.
    (* cut 4 : (subst_exp x) (open_exp_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx fx_v T_v); assumption.
(* Second premise of type_ABS *)
     assumption.
      (*** Subcase 2b ***)
      SCase "T_v_fxempty".
      assert (typing (subst_gamma x v G2 ++ G1) A_minus_x (exp_tab k0 (subst_exp x v e))
                     (ptype (typ_univ k0 NO_FX (subst_typ x v t_e)) A fx_tab, fx_tab)) as premise_TAFN.
      assert (A = A_minus_x) as A_eq_Amx.
         assert (kinding G1 tx kind_U) as kinding_U.
            eapply inversion_empty_effect; eauto using T_v.
         assert (k = kind_U) as k_is_U.
            eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
            subst k; unfold aqual in A_eq; simpl in A_eq; assumption.
      rewrite A_eq_Amx.
      pick fresh b excluding (L `union` (singleton x) `union` dom G) and apply type_TAB.
      intro b_notin_fxtabv.
      rename k0 into k_b.
      cutrewrite((open_exp_typ 0 b (subst_exp x v e)) = (subst_exp x v (open_exp_typ 0 b e))).
      cutrewrite((open_typ_typ 0 b (subst_typ x v t_e)) = (subst_typ x v (open_typ_typ 0 b t_e))).
      assert (b `notin` L) as b_notin_L. fsetdec.
      cut (wf_env ([(b, binding_kind k_b)] ++ G) A). intro wf_bG.
      cut (b `notin` fx_tab). intro b_notin_fxtab.  
      assert ([(b, binding_kind k_b)] ++ G = ([(b, binding_kind k_b)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
         rewrite G_eq. simpl_env. reflexivity.
      pose proof (H b b_notin_L b_notin_fxtab wf_bG (([(b, binding_kind k_b)] ++ G2)) G1 wf_G1 Kind_txk fx_tab fx_tab {} T_v_fxempty 
                 (split_nil_A_1 fx_tab) _ (refl_equal ((open_typ_typ 0 b t_e), fx_tab))
                    A_minus_x x_notin_Amx wf_bG A_eq bG_eq (t b b_notin_L b_notin_fxtab)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : b notin fx_tab *)
       auto.
    (* cut 2 : wf_bG *)
    eapply wf_env_kind; ( auto || fsetdec ).
    (* cut 3 : (subst_typ x) (open_typ_typ y) commute (when x <> y) *) 
     apply sym_equal.
      eapply subst_typ_open_typ_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v_fxempty). assumption.
    (* cut 4 : (subst_exp x) (open_exp_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v_fxempty); assumption.
(* Second premise of type_ABS *)
     assumption.
   (* Finally ... conclude with T-AFN *)
   eapply type_AFN.
      apply (split_nil_A_2 A_minus_x).
      apply premise_TAFN.
      eapply (split_commutes _ _ _ split_fx).
      subst G.
      eapply effect_ok_subst.
        apply wf_GA. 
        eapply typing_effect_ok. 
          apply T_tab.
        eapply typing_effect_ok.
          apply T_v. apply (split_commutes _ _ _ split_fx).
Case "typing_TAB_FX".
intros G2 G1 wf_G1 Kind_txk fx_tabv fx_tab fx_v T_v split_fx t0 eq_tfx A_minus_x x_notin_Amx wf_GA A_eq G_eq T_tab.
injection eq_tfx; intros. subst t0. subst fx.
simpl.
assert ((subst_typ x v (ptype (typ_univ k0 FX t_e) A fx_tab)) = (ptype (subst_typ x v (typ_univ k0 FX t_e)) A fx_tab)) as subst_ptype_eq.
  eapply subst_ptype; auto.
rewrite subst_ptype_eq. simpl.
destruct (exclude_middle_empty_sets fx_v) as [fx_v_empty | fx_v_nonempty].
  (* SubCase 1: fx_v_empty *)
   subst fx_v.
   assert (fx_tab = fx_tabv) as fx_tab_eq_tabv.
     eapply split_nil_eq_1; eauto using split_fx.
   assert (A = A_minus_x) as A_eq_Amx.
     assert (kinding G1 tx kind_U) as kinding_U.
     eapply inversion_empty_effect; eauto using T_v.
     assert (k = kind_U) as k_is_U.
     eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
     subst k; unfold aqual in A_eq; simpl in A_eq; assumption.
   rewrite A_eq_Amx; rewrite fx_tab_eq_tabv.
   subst fx_tabv.
   pick fresh b excluding (L `union` (singleton x) `union` dom G `union` fx_tab) and apply type_TAB_FX.
    (* first premise: fx subset L *)
     fsetdec.
    (* second premise : e typing *)
    assert (b `notin` L) as b_notin_L. fsetdec.
    pose proof (t b b_notin_L) as T_e.
    rename k0 into k_b.
    cutrewrite((open_exp_typ 0 b (subst_exp x v e)) = (subst_exp x v (open_exp_typ 0 b e))).
    cutrewrite((open_typ_typ 0 b (subst_typ x v t_e)) = (subst_typ x v (open_typ_typ 0 b t_e))).
    cut (wf_env ([(b, binding_kind k_b)] ++ G) A). intro wf_bG.
    cut (b `notin` fx_tab). intro b_notin_fxtab.  
    assert ([(b, binding_kind k_b)] ++ G = ([(b, binding_kind k_b)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
      rewrite G_eq. simpl_env. reflexivity.
    pose proof (H b b_notin_L wf_bG (([(b, binding_kind k_b)] ++ G2)) G1 wf_G1 Kind_txk _ _ {} T_v 
                (split_nil_A_1 (fx_tab `union` singleton b))  _ (refl_equal ((open_typ_typ 0 b t_e), (fx_tab `union` singleton b)))
                  A_minus_x x_notin_Amx wf_bG A_eq bG_eq T_e) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : b notin fx_tab *)
       fsetdec.
    (* cut 2 : wf_bG *)
    eapply wf_env_kind; ( auto || fsetdec ).
    (* cut 3 : (subst_typ x) (open_typ_typ y) commute (when x <> y) *) 
     apply sym_equal.
      eapply subst_typ_open_typ_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v). assumption.
    (* cut 4 : (subst_exp x) (open_exp_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v); assumption.
(* Second premise of type_TAB_FX *)
     assumption.
(*** SubCase 2 : fx_v_nonempty ***)
    destruct (inversion_nonempty_effect _ _ _ _ _ Val_v T_v fx_v_nonempty) as [ tx_kind_A | T_v_fxempty ].
   (** SubSubCase 2a. tx_kind_A **)
     SCase "tx_kind_A".
      assert (k = kind_A) as k_is_A.
        eapply kinding_uniqueness; eauto using tx_kind_A, Kind_txk.
      assert (x `in` A) as x_in_A.
        subst k; unfold aqual in A_eq; simpl in A_eq; fsetdec.
      destruct (non_empty_union_2 _ fx_v_nonempty) as [fx_elt fx_elt_in_v].
      assert (fx_elt `in` fx_tabv) as v_in_tabv.
        inversion split_fx; fsetdec.
      assert (AtomSet.F.is_empty (A `union` fx_tab) = false) as nonempty_L.
        eapply non_empty_union_inv; [eauto || fsetdec].
      assert (AtomSet.F.is_empty (A_minus_x `union` fx_tabv) = false) as nonempty_R.
        eapply non_empty_union_inv; [eauto || fsetdec].
      rewrite <- nonempty_R in nonempty_L.
      assert ((ptype (typ_univ k0 FX (subst_typ x v t_e)) A fx_tab) =
              (ptype (typ_univ k0 FX (subst_typ x v t_e)) A_minus_x fx_tabv)) as rw.
        eapply ptype_union_equality; eauto using nonempty_L.
      rewrite rw. (* now starts the reconstruction of the derivation *)
   pick fresh b excluding (L `union` (singleton x) `union` dom G `union` fx_tabv) and apply type_TAB_FX.
    (* first premise : subset fx_tabv *)
       fsetdec.
    (* second premise : e typing *)
    rename k0 into k_b.
    cut (splitA fx_v (fx_tab `union` singleton b) (fx_tabv `union` singleton b)). intro splitter.
    cutrewrite((open_exp_typ 0 b (subst_exp x v e)) = (subst_exp x v (open_exp_typ 0 b e))).
    cutrewrite((open_typ_typ 0 b (subst_typ x v t_e)) = (subst_typ x v (open_typ_typ 0 b t_e))).
    assert (b `notin` L) as b_notin_L. fsetdec.
    cut (wf_env ([(b, binding_kind k_b)] ++ G) A). intro wf_bG.
    cut (b `notin` fx_tab). intro b_notin_fxtab.  
(*     cut (aqual y k_y A = aqual x k (aqual y k_y A_minus_x)). intro aq_eq. *)
(*     cut (x `notin` aqual y k_y A_minus_x). intro x_notin_aqAmx. *)
    assert ([(b, binding_kind k_b)] ++ G = ([(b, binding_kind k_b)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
      rewrite G_eq. simpl_env. reflexivity.
    pose proof (H b b_notin_L wf_bG (([(b, binding_kind k_b)] ++ G2)) G1 wf_G1 Kind_txk _ (fx_tab `union` singleton b) _ T_v
               splitter _ (refl_equal ((open_typ_typ 0 b t_e), (fx_tab `union` singleton b)))
                  A_minus_x x_notin_Amx wf_bG A_eq bG_eq (t b b_notin_L)) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : b notin fx_tab *)
       inversion split_fx. subst fx_tabv; fsetdec.
    (* cut 2 : wf_bG *)
    eapply wf_env_kind; ( auto || fsetdec ).
    (* cut 3 : (subst_typ x) (open_typ_typ y) commute (when x <> y) *) 
     apply sym_equal.
      eapply subst_typ_open_typ_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx fx_v T_v). assumption.
    (* cut 4 : (subst_exp x) (open_exp_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx fx_v T_v); assumption.
    (* cut 5 : splitter *)
      eapply split_extension_R; (eauto || fsetdec).
(* Second premise of type_TAB_FX *)
     assumption.
      (*** Subcase 2b ***)
      SCase "T_v_fxempty".
      assert (typing (subst_gamma x v G2 ++ G1) A_minus_x (exp_tab k0 (subst_exp x v e))
                     (ptype (typ_univ k0 FX (subst_typ x v t_e)) A fx_tab, fx_tab)) as premise_TAFN.
      assert (A = A_minus_x) as A_eq_Amx.
         assert (kinding G1 tx kind_U) as kinding_U.
            eapply inversion_empty_effect; eauto using T_v.
         assert (k = kind_U) as k_is_U.
            eapply kinding_uniqueness; eauto using kinding_U, Kind_txk.
            subst k; unfold aqual in A_eq; simpl in A_eq; assumption.
      rewrite A_eq_Amx.
   pick fresh b excluding (L `union` (singleton x) `union` dom G `union` fx_tab) and apply type_TAB_FX.
    (* first premise: fx subset L *)
     fsetdec.
    (* second premise : e typing *)
    assert (b `notin` L) as b_notin_L. fsetdec.
    pose proof (t b b_notin_L) as T_e.
    rename k0 into k_b.
    cutrewrite((open_exp_typ 0 b (subst_exp x v e)) = (subst_exp x v (open_exp_typ 0 b e))).
    cutrewrite((open_typ_typ 0 b (subst_typ x v t_e)) = (subst_typ x v (open_typ_typ 0 b t_e))).
    cut (wf_env ([(b, binding_kind k_b)] ++ G) A). intro wf_bG.
    cut (b `notin` fx_tab). intro b_notin_fxtab.  
    assert ([(b, binding_kind k_b)] ++ G = ([(b, binding_kind k_b)] ++ G2) ++ [(x, binding_typ tx)] ++ G1) as bG_eq.
      rewrite G_eq. simpl_env. reflexivity.
    pose proof (H b b_notin_L wf_bG (([(b, binding_kind k_b)] ++ G2)) G1 wf_G1 Kind_txk _ _ {} T_v_fxempty
                (split_nil_A_1 (fx_tab `union` singleton b))  _ (refl_equal ((open_typ_typ 0 b t_e), (fx_tab `union` singleton b)))
                  A_minus_x x_notin_Amx wf_bG A_eq bG_eq T_e) as from_IH.
      simpl in from_IH. 
      simpl_env in from_IH.
      apply from_IH.
    (* cut 1 : b notin fx_tab *)
       fsetdec.
    (* cut 2 : wf_bG *)
    eapply wf_env_kind; ( auto || fsetdec ).
    (* cut 3 : (subst_typ x) (open_typ_typ y) commute (when x <> y) *) 
     apply sym_equal.
      eapply subst_typ_open_typ_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v_fxempty). assumption.
    (* cut 4 : (subst_exp x) (open_exp_typ y) commute (when x <> y) *) 
      apply sym_equal.
      eapply subst_exp_open_exp_typ_commute.
         fsetdec.
         destruct (typing_wf_exp G1 {} v tx {} T_v_fxempty); assumption.
(* Second premise of type_TAB_FX *)
     assumption.
   (* Finally ... conclude with T-AFN *)
   eapply type_AFN.
      apply (split_nil_A_2 A_minus_x).
      apply premise_TAFN.
      eapply (split_commutes _ _ _ split_fx).
      subst G.
      eapply effect_ok_subst.
        apply wf_GA. 
        eapply typing_effect_ok. 
          apply T_tab.
        eapply typing_effect_ok.
          apply T_v. apply (split_commutes _ _ _ split_fx).
Case "type_APP".
intros G2 G1 wf_G1 Kind_txk fx_app_v fx_app fx_v T_v split_fx t0 eq_tfx A_minus_x x_notin_Amx wf_GA A_eq G_eq T_app.
injection eq_tfx; intros; subst t0; subst fx.
